//07 11 2003///////////////////////////////////////////////////////////////////
//
//  OpenGL interface unit for Delphi
//  (c) Martin Waldegger
//
///////////////////////////////////////////////////////////////////////////////
//
//  This code is distributed in the hope that it will be useful but
//  WITHOUT ANY WARRANTY. ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
//  DISCLAMED. This includes but is not limited to warranties of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
///////////////////////////////////////////////////////////////////////////////
//
//  This is the standard OpenGL implementation unit for BaseGraph projects. It
//  contains the generic OpenGL 1.5 header and is partly based on the OpenGL
//  unit included with Borland Delphi 6.0, the OpenGL12 unit of the JEDI
//  project (www.delphi-jedi.org) and extgl.h by Lev Povalahev.
//
///////////////////////////////////////////////////////////////////////////////
//
//  Define one or more of the following compiler switches to support a certain
//  OpenGL implementation or additional extensions (if none is given, this
//  interface unit will support generic OpenGL implementations up to version 1.1)
//  - this will ensure, that no unnecessary code is generated
//
//    GL_12 : support generic OpenGL implementations up to version 1.2
//    defining GL_12 automatically adds support for the following extensions:
//      EXT_texture_3d, EXT_bgra, EXT_packed_pixels, EXT_rescale_normal,
//      EXT_separate_specular_color, SGIS_texture_edge_clamp, SGIS_texture_lod,
//      EXT_draw_range_elements
//
//    GL_13 : support generic OpenGL implementations up to version 1.3
//    defining GL_13 automatically adds support for the following extensions:
//      ARB_texture_compression, ARB_texture_cube_map, ARB_multisample,
//      ARB_multitexture, ARB_texture_env_add, ARB_texture_env_combine,
//      ARB_texture_env_dot3, ARB_texture_border_clamp, ARB_transpose_matrix
//
//    GL_14 : support generic OpenGL implementations up to version 1.4
//    defining GL_14 automatically adds support for the following extensions:
//      SGIS_generate_mipmap, NV_blend_square, ARB_depth_texture, ARB_shadow,
//      EXT_fog_coord, EXT_multi_draw_arrays, ARB_point_parameters,
//      EXT_secondary_color, EXT_blend_func_separate, EXT_stencil_wrap,
//      ARB_texture_env_crossbar, EXT_texture_lod_bias,
//      ARB_texture_mirrored_repeat, ARB_window_pos
//
//    GL_15 : support generic OpenGL implementations up to version 1.5
//    defining GL_15 automatically adds support for the following extensions:
//      ARB_vertex_buffer_object, ARB_occlusion_query, EXT_shadow_funcs
//
//    GL_PROGRAM : support vertex and fragment programs (ARB)
//    defining GL_PROGRAM automatically adds support for the following extensions:
//      ARB_vertex_program, ARB_fragment_program
//
//    GL_SHADER : support high level shading language (ARB)
//    defining GL_SHADER automatically adds support for the following extensions:
//      ARB_shader_objects, ARB_vertex_shader, ARB_fragment_shader
//
//
//  If you do not want to use GLU, this unit can provide it's own subset of
//  GLU functions, which will suffice for BaseGraph projects. Define NO_GLU to
//  enable this functionality
//
{$INCLUDE includes.inc}
unit OpenGL;

///////////////////////////////////////////////////////////////////////////////
//
//  This is an interface unit for the use of OpenGL from within Delphi and
//  contains the translations of gl.h, glu.h as well as some support functions.
//
//  Vendor specific extensions are not supported, if you want those, you have to
//  check for them yourself or use another OpenGL implementation for Delphi
//  (for example dglOpenGL at http://www.delphigl.com).
//
///////////////////////////////////////////////////////////////////////////////
//
// function InitOpenGL: Boolean;
//   Needed to load the OpenGL DLLs and all addresses of the standard functions.
//   In case OpenGL is already initialized this function does nothing. No error
//   is raised, if something goes wrong, but you need to inspect the Result in
//   order to know if all went okay.
//   Result: True if successful or already loaded, False otherwise.
//
// function InitOpenGLFromLibrary(GL_Name, GLU_Name: String): Boolean;
//   Same as InitOpenGL, but you can specify specific DLLs. Useful if you want to
//   use different DLLs then those of Windows. This function closes eventually
//   loaded DLLs before it tries to open the newly given.
//   Result: True if successful, False otherwise.
//
// procedure CloseOpenGL;
//   Unloads the OpenGL DLLs and sets all function addresses to nil, including
//   GLU. You can load and unload the DLLs as often as you like.
//
// function  CreateRenderingContext(DC: HDC; PixelFormat: Integer; var PFC: TPixelFormatDescriptor): HGLRC;
//   Sets up a new rendering context depending of the given pixel format
//   given parameters:
//     DC          - the device context for which the rc is to be created
//     PixelFormat - 0=choose from PFC, other value=try to set it
//     PFC         - the pixel format descriptor to use
//   Result: the newly created context or 0 if setup failed
//
// procedure ActivateRenderingContext(DC: HDC; RC: HGLRC);
//   Makes RC in DC 'current' (wglMakeCurrent(..)) and loads all extension
//   addresses and flags if necessary.
//
// after creating and activating a context, you have to free it again with
// wglMakeCurrent(<DC>, 0);
// wglDeleteContext(<RC>);
// ReleaseDC(<WND>,<DC>);
//
// procedure ReadImplementationProperties;
//   Determines version properties of the OpenGL DLL
//   Again, a valid rendering context must be active.
//
// procedure ShowError(MsgString: String);
//   Displays a message box with an error message
//
///////////////////////////////////////////////////////////////////////////
//
// This translation is based on different sources:
//
// - extgl.h by Lev Povalahev
// - OpenGL12.pas by Mike Lischke
// - Borland OpenGL.pas (from Delphi 6)
// - Microsoft OpenGL header files
// - SGI OpenGL header files
// - dglOpenGL from http://www.delphigl.com
//
// The support functions of this unit only generate one active rendering context.
// This is due to the fact, that certain old video cards won't work correctly,
// if multiple rendering contexts are defined. Since the BaseGraph library works
// with one rendering context anyway, this is no real limitation.
//
///////////////////////////////////////////////////////////////////////////

interface

uses Windows;

type
  UINT        = Cardinal;
  DWORD       = UINT;

  GLenum      = UINT;
  PGLenum     = ^GLenum;

  GLboolean   = BYTEBOOL;
  PGLboolean  = ^GLboolean;

  GLbitfield  = UINT;
  PGLbitfield = ^GLbitfield;

  GLbyte      = ShortInt;
  PGLbyte     = ^GLbyte;

  GLshort     = SmallInt;
  PGLshort    = ^GLshort;

  GLint       = Integer;
  PGLint      = ^GLint;

  GLsizei     = Integer;
  PGLsizei    = ^GLsizei;

  GLubyte     = Byte;
  PGLubyte    = ^GLubyte;

  GLushort    = Word;
  PGLushort   = ^GLushort;

  GLuint      = UINT;
  PGLuint     = ^GLuint;

  GLfloat     = Single;
  PGLfloat    = ^GLfloat;

  GLclampf    = Single;
  PGLclampf   = ^GLclampf;

  GLdouble    = Double;
  PGLdouble   = ^GLdouble;

  GLclampd    = Double;
  PGLclampd   = ^GLclampd;

  PPointer    = ^Pointer;


  TVector3f = array[0..2] of GLfloat;
  TVector3d = array[0..2] of GLdouble;

  TVector4f = array[0..3] of GLfloat;
  TVector4d = array[0..3] of GLdouble;

  TMatrix4f = array[0..3, 0..3] of GLfloat;
  TMatrix4d = array[0..3, 0..3] of GLdouble;

  TVector4p = array[0..3] of Pointer;
  TVector4i = array[0..3] of GLint;

  PWGLSwap  = Pointer;
  DevMode   = TDeviceModeA;

{$IFDEF GL_SHADER}
  // GL_ARB_SHADER_OBJECTS
  PGLHandleARB = ^GLHandleARB;
  GLHandleARB  = Cardinal;
  PPGLCharARB  = ^PGLCharARB;
  PGLCharARB   = ^GLCharARB;
  GLCharARB    = Char;
{$ENDIF}

var
  // supported OpenGL version
  GL_10,
  GL_11,
  GL_12,
  GL_13,
  GL_14,
  GL_15,

  // OpenGL extensions not defined in the OpenGL standard implementation
  GL_PROGRAM,         // GL_ARB_vertex_program (+ optional GL_ARB_fragment_program)
  GL_SHADER,          // GL_ARB_shader_objects

  // supported GLU version
  GLU_11,
  GLU_12,
  GLU_13,
  GLU_14 : Boolean;


const
  // ********** GL generic constants **********

  // errors
  GL_NO_ERROR                                       = 0;
  GL_INVALID_ENUM                                   = $0500;
  GL_INVALID_VALUE                                  = $0501;
  GL_INVALID_OPERATION                              = $0502;
  GL_STACK_OVERFLOW                                 = $0503;
  GL_STACK_UNDERFLOW                                = $0504;
  GL_OUT_OF_MEMORY                                  = $0505;

  // attribute bits
  GL_CURRENT_BIT                                    = $00000001;
  GL_POINT_BIT                                      = $00000002;
  GL_LINE_BIT                                       = $00000004;
  GL_POLYGON_BIT                                    = $00000008;
  GL_POLYGON_STIPPLE_BIT                            = $00000010;
  GL_PIXEL_MODE_BIT                                 = $00000020;
  GL_LIGHTING_BIT                                   = $00000040;
  GL_FOG_BIT                                        = $00000080;
  GL_DEPTH_BUFFER_BIT                               = $00000100;
  GL_ACCUM_BUFFER_BIT                               = $00000200;
  GL_STENCIL_BUFFER_BIT                             = $00000400;
  GL_VIEWPORT_BIT                                   = $00000800;
  GL_TRANSFORM_BIT                                  = $00001000;
  GL_ENABLE_BIT                                     = $00002000;
  GL_COLOR_BUFFER_BIT                               = $00004000;
  GL_HINT_BIT                                       = $00008000;
  GL_EVAL_BIT                                       = $00010000;
  GL_LIST_BIT                                       = $00020000;
  GL_TEXTURE_BIT                                    = $00040000;
  GL_SCISSOR_BIT                                    = $00080000;
  GL_ALL_ATTRIB_BITS                                = $000FFFFF;

  // client attribute bits
  GL_CLIENT_PIXEL_STORE_BIT                         = $00000001;
  GL_CLIENT_VERTEX_ARRAY_BIT                        = $00000002;
  GL_CLIENT_ALL_ATTRIB_BITS                         = $FFFFFFFF;

  // boolean values
  GL_FALSE                                          = 0;
  GL_TRUE                                           = 1;

  // primitives
  GL_POINTS                                         = $0000;
  GL_LINES                                          = $0001;
  GL_LINE_LOOP                                      = $0002;
  GL_LINE_STRIP                                     = $0003;
  GL_TRIANGLES                                      = $0004;
  GL_TRIANGLE_STRIP                                 = $0005;
  GL_TRIANGLE_FAN                                   = $0006;
  GL_QUADS                                          = $0007;
  GL_QUAD_STRIP                                     = $0008;
  GL_POLYGON                                        = $0009;

  // blending
  GL_ZERO                                           = 0;
  GL_ONE                                            = 1;
  GL_SRC_COLOR                                      = $0300;
  GL_ONE_MINUS_SRC_COLOR                            = $0301;
  GL_SRC_ALPHA                                      = $0302;
  GL_ONE_MINUS_SRC_ALPHA                            = $0303;
  GL_DST_ALPHA                                      = $0304;
  GL_ONE_MINUS_DST_ALPHA                            = $0305;
  GL_DST_COLOR                                      = $0306;
  GL_ONE_MINUS_DST_COLOR                            = $0307;
  GL_SRC_ALPHA_SATURATE                             = $0308;
  GL_BLEND_DST                                      = $0BE0;
  GL_BLEND_SRC                                      = $0BE1;
  GL_BLEND                                          = $0BE2;

  // buffers
  GL_NONE                                           = 0;
  GL_FRONT_LEFT                                     = $0400;
  GL_FRONT_RIGHT                                    = $0401;
  GL_BACK_LEFT                                      = $0402;
  GL_BACK_RIGHT                                     = $0403;
  GL_FRONT                                          = $0404;
  GL_BACK                                           = $0405;
  GL_LEFT                                           = $0406;
  GL_RIGHT                                          = $0407;
  GL_FRONT_AND_BACK                                 = $0408;
  GL_AUX0                                           = $0409;
  GL_AUX1                                           = $040A;
  GL_AUX2                                           = $040B;
  GL_AUX3                                           = $040C;
  GL_AUX_BUFFERS                                    = $0C00;
  GL_DRAW_BUFFER                                    = $0C01;
  GL_READ_BUFFER                                    = $0C02;
  GL_DOUBLEBUFFER                                   = $0C32;
  GL_STEREO                                         = $0C33;

  // depth buffer
  GL_DEPTH_RANGE                                    = $0B70;
  GL_DEPTH_TEST                                     = $0B71;
  GL_DEPTH_WRITEMASK                                = $0B72;
  GL_DEPTH_CLEAR_VALUE                              = $0B73;
  GL_DEPTH_FUNC                                     = $0B74;
  GL_NEVER                                          = $0200;
  GL_LESS                                           = $0201;
  GL_EQUAL                                          = $0202;
  GL_LEQUAL                                         = $0203;
  GL_GREATER                                        = $0204;
  GL_NOTEQUAL                                       = $0205;
  GL_GEQUAL                                         = $0206;
  GL_ALWAYS                                         = $0207;

  // accumulation buffer
  GL_ACCUM                                          = $0100;
  GL_LOAD                                           = $0101;
  GL_RETURN                                         = $0102;
  GL_MULT                                           = $0103;
  GL_ADD                                            = $0104;
  GL_ACCUM_CLEAR_VALUE                              = $0B80;

  // feedback buffer
  GL_FEEDBACK_BUFFER_POINTER                        = $0DF0;
  GL_FEEDBACK_BUFFER_SIZE                           = $0DF1;
  GL_FEEDBACK_BUFFER_TYPE                           = $0DF2;

  // feedback types
  GL_2D                                             = $0600;
  GL_3D                                             = $0601;
  GL_3D_COLOR                                       = $0602;
  GL_3D_COLOR_TEXTURE                               = $0603;
  GL_4D_COLOR_TEXTURE                               = $0604;

  // feedback tokens
  GL_PASS_THROUGH_TOKEN                             = $0700;
  GL_POINT_TOKEN                                    = $0701;
  GL_LINE_TOKEN                                     = $0702;
  GL_POLYGON_TOKEN                                  = $0703;
  GL_BITMAP_TOKEN                                   = $0704;
  GL_DRAW_PIXEL_TOKEN                               = $0705;
  GL_COPY_PIXEL_TOKEN                               = $0706;
  GL_LINE_RESET_TOKEN                               = $0707;

  // fog
  GL_EXP                                            = $0800;
  GL_EXP2                                           = $0801;
  GL_FOG                                            = $0B60;
  GL_FOG_INDEX                                      = $0B61;
  GL_FOG_DENSITY                                    = $0B62;
  GL_FOG_START                                      = $0B63;
  GL_FOG_END                                        = $0B64;
  GL_FOG_MODE                                       = $0B65;
  GL_FOG_COLOR                                      = $0B66;

  // pixel mode, transfer
  GL_PIXEL_MAP_I_TO_I                               = $0C70;
  GL_PIXEL_MAP_S_TO_S                               = $0C71;
  GL_PIXEL_MAP_I_TO_R                               = $0C72;
  GL_PIXEL_MAP_I_TO_G                               = $0C73;
  GL_PIXEL_MAP_I_TO_B                               = $0C74;
  GL_PIXEL_MAP_I_TO_A                               = $0C75;
  GL_PIXEL_MAP_R_TO_R                               = $0C76;
  GL_PIXEL_MAP_G_TO_G                               = $0C77;
  GL_PIXEL_MAP_B_TO_B                               = $0C78;
  GL_PIXEL_MAP_A_TO_A                               = $0C79;

  // vertex arrays
  GL_VERTEX_ARRAY_POINTER                           = $808E;
  GL_NORMAL_ARRAY_POINTER                           = $808F;
  GL_COLOR_ARRAY_POINTER                            = $8090;
  GL_INDEX_ARRAY_POINTER                            = $8091;
  GL_TEXTURE_COORD_ARRAY_POINTER                    = $8092;
  GL_EDGE_FLAG_ARRAY_POINTER                        = $8093;

  // stenciling
  GL_STENCIL_TEST                                   = $0B90;
  GL_STENCIL_CLEAR_VALUE                            = $0B91;
  GL_STENCIL_FUNC                                   = $0B92;
  GL_STENCIL_VALUE_MASK                             = $0B93;
  GL_STENCIL_FAIL                                   = $0B94;
  GL_STENCIL_PASS_DEPTH_FAIL                        = $0B95;
  GL_STENCIL_PASS_DEPTH_PASS                        = $0B96;
  GL_STENCIL_REF                                    = $0B97;
  GL_STENCIL_WRITEMASK                              = $0B98;
  GL_KEEP                                           = $1E00;
  GL_REPLACE                                        = $1E01;
  GL_INCR                                           = $1E02;
  GL_DECR                                           = $1E03;

  // color material
  GL_COLOR_MATERIAL_FACE                            = $0B55;
  GL_COLOR_MATERIAL_PARAMETER                       = $0B56;
  GL_COLOR_MATERIAL                                 = $0B57;

  // points
  GL_POINT_SMOOTH                                   = $0B10;
  GL_POINT_SIZE                                     = $0B11;
  GL_POINT_SIZE_RANGE                               = $0B12;
  GL_POINT_SIZE_GRANULARITY                         = $0B13;

  // lines
  GL_LINE_SMOOTH                                    = $0B20;
  GL_LINE_WIDTH                                     = $0B21;
  GL_LINE_WIDTH_RANGE                               = $0B22;
  GL_LINE_WIDTH_GRANULARITY                         = $0B23;
  GL_LINE_STIPPLE                                   = $0B24;
  GL_LINE_STIPPLE_PATTERN                           = $0B25;
  GL_LINE_STIPPLE_REPEAT                            = $0B26;

  // polygons
  GL_POLYGON_MODE                                   = $0B40;
  GL_POLYGON_SMOOTH                                 = $0B41;
  GL_POLYGON_STIPPLE                                = $0B42;
  GL_EDGE_FLAG                                      = $0B43;
  GL_CULL_FACE                                      = $0B44;
  GL_CULL_FACE_MODE                                 = $0B45;
  GL_FRONT_FACE                                     = $0B46;
  GL_CW                                             = $0900;
  GL_CCW                                            = $0901;
  GL_POINT                                          = $1B00;
  GL_LINE                                           = $1B01;
  GL_FILL                                           = $1B02;

  // display lists
  GL_LIST_MODE                                      = $0B30;
  GL_LIST_BASE                                      = $0B32;
  GL_LIST_INDEX                                     = $0B33;
  GL_COMPILE                                        = $1300;
  GL_COMPILE_AND_EXECUTE                            = $1301;

  // lighting
  GL_LIGHTING                                       = $0B50;
  GL_LIGHT_MODEL_LOCAL_VIEWER                       = $0B51;
  GL_LIGHT_MODEL_TWO_SIDE                           = $0B52;
  GL_LIGHT_MODEL_AMBIENT                            = $0B53;
  GL_SHADE_MODEL                                    = $0B54;
  GL_NORMALIZE                                      = $0BA1;
  GL_AMBIENT                                        = $1200;
  GL_DIFFUSE                                        = $1201;
  GL_SPECULAR                                       = $1202;
  GL_POSITION                                       = $1203;
  GL_SPOT_DIRECTION                                 = $1204;
  GL_SPOT_EXPONENT                                  = $1205;
  GL_SPOT_CUTOFF                                    = $1206;
  GL_CONSTANT_ATTENUATION                           = $1207;
  GL_LINEAR_ATTENUATION                             = $1208;
  GL_QUADRATIC_ATTENUATION                          = $1209;
  GL_EMISSION                                       = $1600;
  GL_SHININESS                                      = $1601;
  GL_AMBIENT_AND_DIFFUSE                            = $1602;
  GL_COLOR_INDEXES                                  = $1603;
  GL_FLAT                                           = $1D00;
  GL_SMOOTH                                         = $1D01;
  GL_LIGHT0                                         = $4000;
  GL_LIGHT1                                         = $4001;
  GL_LIGHT2                                         = $4002;
  GL_LIGHT3                                         = $4003;
  GL_LIGHT4                                         = $4004;
  GL_LIGHT5                                         = $4005;
  GL_LIGHT6                                         = $4006;
  GL_LIGHT7                                         = $4007;

  // matrix modes
  GL_MATRIX_MODE                                    = $0BA0;
  GL_MODELVIEW                                      = $1700;
  GL_PROJECTION                                     = $1701;
  GL_TEXTURE                                        = $1702;

  // gets
  GL_CURRENT_COLOR                                  = $0B00;
  GL_CURRENT_INDEX                                  = $0B01;
  GL_CURRENT_NORMAL                                 = $0B02;
  GL_CURRENT_TEXTURE_COORDS                         = $0B03;
  GL_CURRENT_RASTER_COLOR                           = $0B04;
  GL_CURRENT_RASTER_INDEX                           = $0B05;
  GL_CURRENT_RASTER_TEXTURE_COORDS                  = $0B06;
  GL_CURRENT_RASTER_POSITION                        = $0B07;
  GL_CURRENT_RASTER_POSITION_VALID                  = $0B08;
  GL_CURRENT_RASTER_DISTANCE                        = $0B09;
  GL_MAX_LIST_NESTING                               = $0B31;
  GL_VIEWPORT                                       = $0BA2;
  GL_MODELVIEW_STACK_DEPTH                          = $0BA3;
  GL_PROJECTION_STACK_DEPTH                         = $0BA4;
  GL_TEXTURE_STACK_DEPTH                            = $0BA5;
  GL_MODELVIEW_MATRIX                               = $0BA6;
  GL_PROJECTION_MATRIX                              = $0BA7;
  GL_TEXTURE_MATRIX                                 = $0BA8;
  GL_ATTRIB_STACK_DEPTH                             = $0BB0;
  GL_CLIENT_ATTRIB_STACK_DEPTH                      = $0BB1;

  // alpha testing
  GL_ALPHA_TEST                                     = $0BC0;
  GL_ALPHA_TEST_FUNC                                = $0BC1;
  GL_ALPHA_TEST_REF                                 = $0BC2;

  GL_LOGIC_OP_MODE                                  = $0BF0;
  GL_INDEX_LOGIC_OP                                 = $0BF1;
  GL_LOGIC_OP                                       = $0BF1;
  GL_COLOR_LOGIC_OP                                 = $0BF2;
  GL_SCISSOR_BOX                                    = $0C10;
  GL_SCISSOR_TEST                                   = $0C11;
  GL_INDEX_CLEAR_VALUE                              = $0C20;
  GL_INDEX_WRITEMASK                                = $0C21;
  GL_COLOR_CLEAR_VALUE                              = $0C22;
  GL_COLOR_WRITEMASK                                = $0C23;
  GL_INDEX_MODE                                     = $0C30;
  GL_RGBA_MODE                                      = $0C31;
  GL_RENDER_MODE                                    = $0C40;
  GL_PERSPECTIVE_CORRECTION_HINT                    = $0C50;
  GL_POINT_SMOOTH_HINT                              = $0C51;
  GL_LINE_SMOOTH_HINT                               = $0C52;
  GL_POLYGON_SMOOTH_HINT                            = $0C53;
  GL_FOG_HINT                                       = $0C54;
  GL_TEXTURE_GEN_S                                  = $0C60;
  GL_TEXTURE_GEN_T                                  = $0C61;
  GL_TEXTURE_GEN_R                                  = $0C62;
  GL_TEXTURE_GEN_Q                                  = $0C63;
  GL_PIXEL_MAP_I_TO_I_SIZE                          = $0CB0;
  GL_PIXEL_MAP_S_TO_S_SIZE                          = $0CB1;
  GL_PIXEL_MAP_I_TO_R_SIZE                          = $0CB2;
  GL_PIXEL_MAP_I_TO_G_SIZE                          = $0CB3;
  GL_PIXEL_MAP_I_TO_B_SIZE                          = $0CB4;
  GL_PIXEL_MAP_I_TO_A_SIZE                          = $0CB5;
  GL_PIXEL_MAP_R_TO_R_SIZE                          = $0CB6;
  GL_PIXEL_MAP_G_TO_G_SIZE                          = $0CB7;
  GL_PIXEL_MAP_B_TO_B_SIZE                          = $0CB8;
  GL_PIXEL_MAP_A_TO_A_SIZE                          = $0CB9;
  GL_UNPACK_SWAP_BYTES                              = $0CF0;
  GL_UNPACK_LSB_FIRST                               = $0CF1;
  GL_UNPACK_ROW_LENGTH                              = $0CF2;
  GL_UNPACK_SKIP_ROWS                               = $0CF3;
  GL_UNPACK_SKIP_PIXELS                             = $0CF4;
  GL_UNPACK_ALIGNMENT                               = $0CF5;
  GL_PACK_SWAP_BYTES                                = $0D00;
  GL_PACK_LSB_FIRST                                 = $0D01;
  GL_PACK_ROW_LENGTH                                = $0D02;
  GL_PACK_SKIP_ROWS                                 = $0D03;
  GL_PACK_SKIP_PIXELS                               = $0D04;
  GL_PACK_ALIGNMENT                                 = $0D05;
  GL_MAP_COLOR                                      = $0D10;
  GL_MAP_STENCIL                                    = $0D11;
  GL_INDEX_SHIFT                                    = $0D12;
  GL_INDEX_OFFSET                                   = $0D13;
  GL_RED_SCALE                                      = $0D14;
  GL_RED_BIAS                                       = $0D15;
  GL_ZOOM_X                                         = $0D16;
  GL_ZOOM_Y                                         = $0D17;
  GL_GREEN_SCALE                                    = $0D18;
  GL_GREEN_BIAS                                     = $0D19;
  GL_BLUE_SCALE                                     = $0D1A;
  GL_BLUE_BIAS                                      = $0D1B;
  GL_ALPHA_SCALE                                    = $0D1C;
  GL_ALPHA_BIAS                                     = $0D1D;
  GL_DEPTH_SCALE                                    = $0D1E;
  GL_DEPTH_BIAS                                     = $0D1F;
  GL_MAX_EVAL_ORDER                                 = $0D30;
  GL_MAX_LIGHTS                                     = $0D31;
  GL_MAX_CLIP_PLANES                                = $0D32;
  GL_MAX_TEXTURE_SIZE                               = $0D33;
  GL_MAX_PIXEL_MAP_TABLE                            = $0D34;
  GL_MAX_ATTRIB_STACK_DEPTH                         = $0D35;
  GL_MAX_MODELVIEW_STACK_DEPTH                      = $0D36;
  GL_MAX_NAME_STACK_DEPTH                           = $0D37;
  GL_MAX_PROJECTION_STACK_DEPTH                     = $0D38;
  GL_MAX_TEXTURE_STACK_DEPTH                        = $0D39;
  GL_MAX_VIEWPORT_DIMS                              = $0D3A;
  GL_MAX_CLIENT_ATTRIB_STACK_DEPTH                  = $0D3B;
  GL_SUBPIXEL_BITS                                  = $0D50;
  GL_INDEX_BITS                                     = $0D51;
  GL_RED_BITS                                       = $0D52;
  GL_GREEN_BITS                                     = $0D53;
  GL_BLUE_BITS                                      = $0D54;
  GL_ALPHA_BITS                                     = $0D55;
  GL_DEPTH_BITS                                     = $0D56;
  GL_STENCIL_BITS                                   = $0D57;
  GL_ACCUM_RED_BITS                                 = $0D58;
  GL_ACCUM_GREEN_BITS                               = $0D59;
  GL_ACCUM_BLUE_BITS                                = $0D5A;
  GL_ACCUM_ALPHA_BITS                               = $0D5B;
  GL_NAME_STACK_DEPTH                               = $0D70;
  GL_AUTO_NORMAL                                    = $0D80;
  GL_MAP1_COLOR_4                                   = $0D90;
  GL_MAP1_INDEX                                     = $0D91;
  GL_MAP1_NORMAL                                    = $0D92;
  GL_MAP1_TEXTURE_COORD_1                           = $0D93;
  GL_MAP1_TEXTURE_COORD_2                           = $0D94;
  GL_MAP1_TEXTURE_COORD_3                           = $0D95;
  GL_MAP1_TEXTURE_COORD_4                           = $0D96;
  GL_MAP1_VERTEX_3                                  = $0D97;
  GL_MAP1_VERTEX_4                                  = $0D98;
  GL_MAP2_COLOR_4                                   = $0DB0;
  GL_MAP2_INDEX                                     = $0DB1;
  GL_MAP2_NORMAL                                    = $0DB2;
  GL_MAP2_TEXTURE_COORD_1                           = $0DB3;
  GL_MAP2_TEXTURE_COORD_2                           = $0DB4;
  GL_MAP2_TEXTURE_COORD_3                           = $0DB5;
  GL_MAP2_TEXTURE_COORD_4                           = $0DB6;
  GL_MAP2_VERTEX_3                                  = $0DB7;
  GL_MAP2_VERTEX_4                                  = $0DB8;
  GL_MAP1_GRID_DOMAIN                               = $0DD0;
  GL_MAP1_GRID_SEGMENTS                             = $0DD1;
  GL_MAP2_GRID_DOMAIN                               = $0DD2;
  GL_MAP2_GRID_SEGMENTS                             = $0DD3;
  GL_TEXTURE_1D                                     = $0DE0;
  GL_TEXTURE_2D                                     = $0DE1;
  GL_SELECTION_BUFFER_POINTER                       = $0DF3;
  GL_SELECTION_BUFFER_SIZE                          = $0DF4;
  GL_POLYGON_OFFSET_UNITS                           = $2A00;
  GL_POLYGON_OFFSET_POINT                           = $2A01;
  GL_POLYGON_OFFSET_LINE                            = $2A02;
  GL_POLYGON_OFFSET_FILL                            = $8037;
  GL_POLYGON_OFFSET_FACTOR                          = $8038;
  GL_TEXTURE_BINDING_1D                             = $8068;
  GL_TEXTURE_BINDING_2D                             = $8069;
  GL_VERTEX_ARRAY                                   = $8074;
  GL_NORMAL_ARRAY                                   = $8075;
  GL_COLOR_ARRAY                                    = $8076;
  GL_INDEX_ARRAY                                    = $8077;
  GL_TEXTURE_COORD_ARRAY                            = $8078;
  GL_EDGE_FLAG_ARRAY                                = $8079;
  GL_VERTEX_ARRAY_SIZE                              = $807A;
  GL_VERTEX_ARRAY_TYPE                              = $807B;
  GL_VERTEX_ARRAY_STRIDE                            = $807C;
  GL_NORMAL_ARRAY_TYPE                              = $807E;
  GL_NORMAL_ARRAY_STRIDE                            = $807F;
  GL_COLOR_ARRAY_SIZE                               = $8081;
  GL_COLOR_ARRAY_TYPE                               = $8082;
  GL_COLOR_ARRAY_STRIDE                             = $8083;
  GL_INDEX_ARRAY_TYPE                               = $8085;
  GL_INDEX_ARRAY_STRIDE                             = $8086;
  GL_TEXTURE_COORD_ARRAY_SIZE                       = $8088;
  GL_TEXTURE_COORD_ARRAY_TYPE                       = $8089;
  GL_TEXTURE_COORD_ARRAY_STRIDE                     = $808A;
  GL_EDGE_FLAG_ARRAY_STRIDE                         = $808C;

  // evaluators
  GL_COEFF                                          = $0A00;
  GL_ORDER                                          = $0A01;
  GL_DOMAIN                                         = $0A02;

  // texture mapping
  GL_TEXTURE_WIDTH                                  = $1000;
  GL_TEXTURE_HEIGHT                                 = $1001;
  GL_TEXTURE_INTERNAL_FORMAT                        = $1003;
  GL_TEXTURE_COMPONENTS                             = $1003;
  GL_TEXTURE_BORDER_COLOR                           = $1004;
  GL_TEXTURE_BORDER                                 = $1005;
  GL_TEXTURE_RED_SIZE                               = $805C;
  GL_TEXTURE_GREEN_SIZE                             = $805D;
  GL_TEXTURE_BLUE_SIZE                              = $805E;
  GL_TEXTURE_ALPHA_SIZE                             = $805F;
  GL_TEXTURE_LUMINANCE_SIZE                         = $8060;
  GL_TEXTURE_INTENSITY_SIZE                         = $8061;
  GL_TEXTURE_PRIORITY                               = $8066;
  GL_TEXTURE_RESIDENT                               = $8067;
  GL_S                                              = $2000;
  GL_T                                              = $2001;
  GL_R                                              = $2002;
  GL_Q                                              = $2003;
  GL_MODULATE                                       = $2100;
  GL_DECAL                                          = $2101;
  GL_TEXTURE_ENV_MODE                               = $2200;
  GL_TEXTURE_ENV_COLOR                              = $2201;
  GL_TEXTURE_ENV                                    = $2300;
  GL_EYE_LINEAR                                     = $2400;
  GL_OBJECT_LINEAR                                  = $2401;
  GL_SPHERE_MAP                                     = $2402;
  GL_TEXTURE_GEN_MODE                               = $2500;
  GL_OBJECT_PLANE                                   = $2501;
  GL_EYE_PLANE                                      = $2502;
  GL_NEAREST                                        = $2600;
  GL_LINEAR                                         = $2601;
  GL_NEAREST_MIPMAP_NEAREST                         = $2700;
  GL_LINEAR_MIPMAP_NEAREST                          = $2701;
  GL_NEAREST_MIPMAP_LINEAR                          = $2702;
  GL_LINEAR_MIPMAP_LINEAR                           = $2703;
  GL_TEXTURE_MAG_FILTER                             = $2800;
  GL_TEXTURE_MIN_FILTER                             = $2801;
  GL_TEXTURE_WRAP_S                                 = $2802;
  GL_TEXTURE_WRAP_T                                 = $2803;
  GL_PROXY_TEXTURE_1D                               = $8063;
  GL_PROXY_TEXTURE_2D                               = $8064;
  GL_CLAMP                                          = $2900;
  GL_REPEAT                                         = $2901;

  // hints
  GL_DONT_CARE                                      = $1100;
  GL_FASTEST                                        = $1101;
  GL_NICEST                                         = $1102;

  // data types
  GL_BYTE                                           = $1400;
  GL_UNSIGNED_BYTE                                  = $1401;
  GL_SHORT                                          = $1402;
  GL_UNSIGNED_SHORT                                 = $1403;
  GL_INT                                            = $1404;
  GL_UNSIGNED_INT                                   = $1405;
  GL_FLOAT                                          = $1406;
  GL_2_BYTES                                        = $1407;
  GL_3_BYTES                                        = $1408;
  GL_4_BYTES                                        = $1409;
  GL_DOUBLE                                         = $140A;
  GL_DOUBLE_EXT                                     = $140A;

  // logic operations
  GL_CLEAR                                          = $1500;
  GL_AND                                            = $1501;
  GL_AND_REVERSE                                    = $1502;
  GL_COPY                                           = $1503;
  GL_AND_INVERTED                                   = $1504;
  GL_NOOP                                           = $1505;
  GL_XOR                                            = $1506;
  GL_OR                                             = $1507;
  GL_NOR                                            = $1508;
  GL_EQUIV                                          = $1509;
  GL_INVERT                                         = $150A;
  GL_OR_REVERSE                                     = $150B;
  GL_COPY_INVERTED                                  = $150C;
  GL_OR_INVERTED                                    = $150D;
  GL_NAND                                           = $150E;
  GL_SET                                            = $150F;

  // PixelCopyType
  GL_COLOR                                          = $1800;
  GL_DEPTH                                          = $1801;
  GL_STENCIL                                        = $1802;

  // pixel formats
  GL_COLOR_INDEX                                    = $1900;
  GL_STENCIL_INDEX                                  = $1901;
  GL_DEPTH_COMPONENT                                = $1902;
  GL_RED                                            = $1903;
  GL_GREEN                                          = $1904;
  GL_BLUE                                           = $1905;
  GL_ALPHA                                          = $1906;
  GL_RGB                                            = $1907;
  GL_RGBA                                           = $1908;
  GL_LUMINANCE                                      = $1909;
  GL_LUMINANCE_ALPHA                                = $190A;

  // pixel type
  GL_BITMAP                                         = $1A00;

  // rendering modes
  GL_RENDER                                         = $1C00;
  GL_FEEDBACK                                       = $1C01;
  GL_SELECT                                         = $1C02;

  // implementation strings
  GL_VENDOR                                         = $1F00;
  GL_RENDERER                                       = $1F01;
  GL_VERSION                                        = $1F02;
  GL_EXTENSIONS                                     = $1F03;

  // pixel formats
  GL_R3_G3_B2                                       = $2A10;
  GL_ALPHA4                                         = $803B;
  GL_ALPHA8                                         = $803C;
  GL_ALPHA12                                        = $803D;
  GL_ALPHA16                                        = $803E;
  GL_LUMINANCE4                                     = $803F;
  GL_LUMINANCE8                                     = $8040;
  GL_LUMINANCE12                                    = $8041;
  GL_LUMINANCE16                                    = $8042;
  GL_LUMINANCE4_ALPHA4                              = $8043;
  GL_LUMINANCE6_ALPHA2                              = $8044;
  GL_LUMINANCE8_ALPHA8                              = $8045;
  GL_LUMINANCE12_ALPHA4                             = $8046;
  GL_LUMINANCE12_ALPHA12                            = $8047;
  GL_LUMINANCE16_ALPHA16                            = $8048;
  GL_INTENSITY                                      = $8049;
  GL_INTENSITY4                                     = $804A;
  GL_INTENSITY8                                     = $804B;
  GL_INTENSITY12                                    = $804C;
  GL_INTENSITY16                                    = $804D;
  GL_RGB4                                           = $804F;
  GL_RGB5                                           = $8050;
  GL_RGB8                                           = $8051;
  GL_RGB10                                          = $8052;
  GL_RGB12                                          = $8053;
  GL_RGB16                                          = $8054;
  GL_RGBA2                                          = $8055;
  GL_RGBA4                                          = $8056;
  GL_RGB5_A1                                        = $8057;
  GL_RGBA8                                          = $8058;
  GL_RGB10_A2                                       = $8059;
  GL_RGBA12                                         = $805A;
  GL_RGBA16                                         = $805B;

  // interleaved arrays formats
  GL_V2F                                            = $2A20;
  GL_V3F                                            = $2A21;
  GL_C4UB_V2F                                       = $2A22;
  GL_C4UB_V3F                                       = $2A23;
  GL_C3F_V3F                                        = $2A24;
  GL_N3F_V3F                                        = $2A25;
  GL_C4F_N3F_V3F                                    = $2A26;
  GL_T2F_V3F                                        = $2A27;
  GL_T4F_V4F                                        = $2A28;
  GL_T2F_C4UB_V3F                                   = $2A29;
  GL_T2F_C3F_V3F                                    = $2A2A;
  GL_T2F_N3F_V3F                                    = $2A2B;
  GL_T2F_C4F_N3F_V3F                                = $2A2C;
  GL_T4F_C4F_N3F_V4F                                = $2A2D;

  // clip planes
  GL_CLIP_PLANE0                                    = $3000;
  GL_CLIP_PLANE1                                    = $3001;
  GL_CLIP_PLANE2                                    = $3002;
  GL_CLIP_PLANE3                                    = $3003;
  GL_CLIP_PLANE4                                    = $3004;
  GL_CLIP_PLANE5                                    = $3005;

  // miscellaneous
  GL_DITHER                                         = $0BD0;

{$IFDEF GL_12}
  // ********** GL 1.2 constants **********
  GL_RESCALE_NORMAL                                 = $803A; // GL 1.2
  GL_CLAMP_TO_EDGE                                  = $812F; // GL 1.2
  GL_MAX_ELEMENTS_VERTICES                          = $80E8; // GL 1.2
  GL_MAX_ELEMENTS_INDICES                           = $80E9; // GL 1.2
  GL_BGR                                            = $80E0; // GL 1.2
  GL_BGRA                                           = $80E1; // GL 1.2
  GL_UNSIGNED_BYTE_3_3_2                            = $8032; // GL 1.2
  GL_UNSIGNED_BYTE_2_3_3_REV                        = $8362; // GL 1.2
  GL_UNSIGNED_SHORT_5_6_5                           = $8363; // GL 1.2
  GL_UNSIGNED_SHORT_5_6_5_REV                       = $8364; // GL 1.2
  GL_UNSIGNED_SHORT_4_4_4_4                         = $8033; // GL 1.2
  GL_UNSIGNED_SHORT_4_4_4_4_REV                     = $8365; // GL 1.2
  GL_UNSIGNED_SHORT_5_5_5_1                         = $8034; // GL 1.2
  GL_UNSIGNED_SHORT_1_5_5_5_REV                     = $8366; // GL 1.2
  GL_UNSIGNED_INT_8_8_8_8                           = $8035; // GL 1.2
  GL_UNSIGNED_INT_8_8_8_8_REV                       = $8367; // GL 1.2
  GL_UNSIGNED_INT_10_10_10_2                        = $8036; // GL 1.2
  GL_UNSIGNED_INT_2_10_10_10_REV                    = $8368; // GL 1.2
  GL_LIGHT_MODEL_COLOR_CONTROL                      = $81F8; // GL 1.2
  GL_SINGLE_COLOR                                   = $81F9; // GL 1.2
  GL_SEPARATE_SPECULAR_COLOR                        = $81FA; // GL 1.2
  GL_TEXTURE_MIN_LOD                                = $813A; // GL 1.2
  GL_TEXTURE_MAX_LOD                                = $813B; // GL 1.2
  GL_TEXTURE_BASE_LEVEL                             = $813C; // GL 1.2
  GL_TEXTURE_MAX_LEVEL                              = $813D; // GL 1.2
  GL_SMOOTH_POINT_SIZE_RANGE                        = $0B12; // GL 1.2
  GL_SMOOTH_POINT_SIZE_GRANULARITY                  = $0B13; // GL 1.2
  GL_SMOOTH_LINE_WIDTH_RANGE                        = $0B22; // GL 1.2
  GL_SMOOTH_LINE_WIDTH_GRANULARITY                  = $0B23; // GL 1.2
  GL_ALIASED_POINT_SIZE_RANGE                       = $846D; // GL 1.2
  GL_ALIASED_LINE_WIDTH_RANGE                       = $846E; // GL 1.2
  GL_PACK_SKIP_IMAGES                               = $806B; // GL 1.2
  GL_PACK_IMAGE_HEIGHT                              = $806C; // GL 1.2
  GL_UNPACK_SKIP_IMAGES                             = $806D; // GL 1.2
  GL_UNPACK_IMAGE_HEIGHT                            = $806E; // GL 1.2
  GL_TEXTURE_3D                                     = $806F; // GL 1.2
  GL_PROXY_TEXTURE_3D                               = $8070; // GL 1.2
  GL_TEXTURE_DEPTH                                  = $8071; // GL 1.2
  GL_TEXTURE_WRAP_R                                 = $8072; // GL 1.2
  GL_MAX_3D_TEXTURE_SIZE                            = $8073; // GL 1.2
  GL_TEXTURE_BINDING_3D                             = $806A; // GL 1.2
{$ENDIF}

{$IFDEF GL_13}
  // ********** GL 1.3 constants **********
  GL_TEXTURE0                                       = $84C0;  // GL 1.3
  GL_TEXTURE1                                       = $84C1;  // GL 1.3
  GL_TEXTURE2                                       = $84C2;  // GL 1.3
  GL_TEXTURE3                                       = $84C3;  // GL 1.3
  GL_TEXTURE4                                       = $84C4;  // GL 1.3
  GL_TEXTURE5                                       = $84C5;  // GL 1.3
  GL_TEXTURE6                                       = $84C6;  // GL 1.3
  GL_TEXTURE7                                       = $84C7;  // GL 1.3
  GL_TEXTURE8                                       = $84C8;  // GL 1.3
  GL_TEXTURE9                                       = $84C9;  // GL 1.3
  GL_TEXTURE10                                      = $84CA;  // GL 1.3
  GL_TEXTURE11                                      = $84CB;  // GL 1.3
  GL_TEXTURE12                                      = $84CC;  // GL 1.3
  GL_TEXTURE13                                      = $84CD;  // GL 1.3
  GL_TEXTURE14                                      = $84CE;  // GL 1.3
  GL_TEXTURE15                                      = $84CF;  // GL 1.3
  GL_TEXTURE16                                      = $84D0;  // GL 1.3
  GL_TEXTURE17                                      = $84D1;  // GL 1.3
  GL_TEXTURE18                                      = $84D2;  // GL 1.3
  GL_TEXTURE19                                      = $84D3;  // GL 1.3
  GL_TEXTURE20                                      = $84D4;  // GL 1.3
  GL_TEXTURE21                                      = $84D5;  // GL 1.3
  GL_TEXTURE22                                      = $84D6;  // GL 1.3
  GL_TEXTURE23                                      = $84D7;  // GL 1.3
  GL_TEXTURE24                                      = $84D8;  // GL 1.3
  GL_TEXTURE25                                      = $84D9;  // GL 1.3
  GL_TEXTURE26                                      = $84DA;  // GL 1.3
  GL_TEXTURE27                                      = $84DB;  // GL 1.3
  GL_TEXTURE28                                      = $84DC;  // GL 1.3
  GL_TEXTURE29                                      = $84DD;  // GL 1.3
  GL_TEXTURE30                                      = $84DE;  // GL 1.3
  GL_TEXTURE31                                      = $84DF;  // GL 1.3
  GL_ACTIVE_TEXTURE                                 = $84E0;  // GL 1.3
  GL_CLIENT_ACTIVE_TEXTURE                          = $84E1;  // GL 1.3
  GL_MAX_TEXTURE_UNITS                              = $84E2;  // GL 1.3

  GL_NORMAL_MAP                                     = $8511;  // GL 1.3
  GL_REFLECTION_MAP                                 = $8512;  // GL 1.3
  GL_TEXTURE_CUBE_MAP                               = $8513;  // GL 1.3
  GL_TEXTURE_BINDING_CUBE_MAP                       = $8514;  // GL 1.3
  GL_TEXTURE_CUBE_MAP_POSITIVE_X                    = $8515;  // GL 1.3
  GL_TEXTURE_CUBE_MAP_NEGATIVE_X                    = $8516;  // GL 1.3
  GL_TEXTURE_CUBE_MAP_POSITIVE_Y                    = $8517;  // GL 1.3
  GL_TEXTURE_CUBE_MAP_NEGATIVE_Y                    = $8518;  // GL 1.3
  GL_TEXTURE_CUBE_MAP_POSITIVE_Z                    = $8519;  // GL 1.3
  GL_TEXTURE_CUBE_MAP_NEGATIVE_Z                    = $851A;  // GL 1.3
  GL_PROXY_TEXTURE_CUBE_MAP                         = $851B;  // GL 1.3
  GL_MAX_CUBE_MAP_TEXTURE_SIZE                      = $851C;  // GL 1.3

  GL_COMPRESSED_ALPHA                               = $84E9;  // GL 1.3
  GL_COMPRESSED_LUMINANCE                           = $84EA;  // GL 1.3
  GL_COMPRESSED_LUMINANCE_ALPHA                     = $84EB;  // GL 1.3
  GL_COMPRESSED_INTENSITY                           = $84EC;  // GL 1.3
  GL_COMPRESSED_RGB                                 = $84ED;  // GL 1.3
  GL_COMPRESSED_RGBA                                = $84EE;  // GL 1.3
  GL_TEXTURE_COMPRESSION_HINT                       = $84EF;  // GL 1.3
  GL_TEXTURE_COMPRESSED_IMAGE_SIZE                  = $86A0;  // GL 1.3
  GL_TEXTURE_COMPRESSED                             = $86A1;  // GL 1.3
  GL_NUM_COMPRESSED_TEXTURE_FORMATS                 = $86A2;  // GL 1.3
  GL_COMPRESSED_TEXTURE_FORMATS                     = $86A3;  // GL 1.3

  GL_MULTISAMPLE                                    = $809D;  // GL 1.3
  GL_SAMPLE_ALPHA_TO_COVERAGE                       = $809E;  // GL 1.3
  GL_SAMPLE_ALPHA_TO_ONE                            = $809F;  // GL 1.3
  GL_SAMPLE_COVERAGE                                = $80A0;  // GL 1.3
  GL_SAMPLE_BUFFERS                                 = $80A8;  // GL 1.3
  GL_SAMPLES                                        = $80A9;  // GL 1.3
  GL_SAMPLE_COVERAGE_VALUE                          = $80AA;  // GL 1.3
  GL_SAMPLE_COVERAGE_INVERT                         = $80AB;  // GL 1.3
  GL_MULTISAMPLE_BIT                                = $20000000;  // GL 1.3

  GL_TRANSPOSE_MODELVIEW_MATRIX                     = $84E3;  // GL 1.3
  GL_TRANSPOSE_PROJECTION_MATRIX                    = $84E4;  // GL 1.3
  GL_TRANSPOSE_TEXTURE_MATRIX                       = $84E5;  // GL 1.3
  GL_TRANSPOSE_COLOR_MATRIX                         = $84E6;  // GL 1.3

  GL_COMBINE                                        = $8570;  // GL 1.3
  GL_COMBINE_RGB                                    = $8571;  // GL 1.3
  GL_COMBINE_ALPHA                                  = $8572;  // GL 1.3
  GL_SRC0_RGB                                       = $8580;  // GL 1.3
  GL_SRC1_RGB                                       = $8581;  // GL 1.3
  GL_SRC2_RGB                                       = $8582;  // GL 1.3
  GL_SRC0_ALPHA                                     = $8588;  // GL 1.3
  GL_SRC1_ALPHA                                     = $8589;  // GL 1.3
  GL_SRC2_ALPHA                                     = $858A;  // GL 1.3
  GL_OPERAND0_RGB                                   = $8590;  // GL 1.3
  GL_OPERAND1_RGB                                   = $8591;  // GL 1.3
  GL_OPERAND2_RGB                                   = $8592;  // GL 1.3
  GL_OPERAND0_ALPHA                                 = $8598;  // GL 1.3
  GL_OPERAND1_ALPHA                                 = $8599;  // GL 1.3
  GL_OPERAND2_ALPHA                                 = $859A;  // GL 1.3
  GL_RGB_SCALE                                      = $8573;  // GL 1.3
  GL_ADD_SIGNED                                     = $8574;  // GL 1.3
  GL_INTERPOLATE                                    = $8575;  // GL 1.3
  GL_SUBTRACT                                       = $84E7;  // GL 1.3
  GL_CONSTANT                                       = $8576;  // GL 1.3
  GL_PRIMARY_COLOR                                  = $8577;  // GL 1.3
  GL_PREVIOUS                                       = $8578;  // GL 1.3
  GL_DOT3_RGB                                       = $86AE;  // GL 1.3
  GL_DOT3_RGBA                                      = $86AF;  // GL 1.3
  GL_CLAMP_TO_BORDER                                = $812D;  // GL 1.3

// deprecated
  GL_SOURCE0_RGB = GL_SRC0_RGB;
  GL_SOURCE1_RGB = GL_SRC1_RGB;
  GL_SOURCE2_RGB = GL_SRC2_RGB;
  GL_SOURCE0_ALPHA = GL_SRC0_ALPHA;
  GL_SOURCE1_ALPHA = GL_SRC1_ALPHA;
  GL_SOURCE2_ALPHA = GL_SRC2_ALPHA;
{$ENDIF}

{$IFDEF GL_14}
  // ********** GL 1.4 constants **********
  GL_GENERATE_MIPMAP                                = $8191;   // GL 1.4
  GL_GENERATE_MIPMAP_HINT                           = $8192;   // GL 1.4
  GL_DEPTH_COMPONENT16                              = $81A5;   // GL 1.4
  GL_DEPTH_COMPONENT24                              = $81A6;   // GL 1.4
  GL_DEPTH_COMPONENT32                              = $81A7;   // GL 1.4
  GL_TEXTURE_DEPTH_SIZE                             = $884A;   // GL 1.4
  GL_DEPTH_TEXTURE_MODE                             = $884B;   // GL 1.4
  GL_TEXTURE_COMPARE_MODE                           = $884C;   // GL 1.4
  GL_TEXTURE_COMPARE_FUNC                           = $884D;   // GL 1.4
  GL_COMPARE_R_TO_TEXTURE                           = $884E;   // GL 1.4
  GL_FOG_COORD_SOURCE                               = $8450;   // GL 1.4
  GL_FOG_COORD                                      = $8451;   // GL 1.4
  GL_FRAGMENT_DEPTH                                 = $8452;   // GL 1.4
  GL_CURRENT_FOG_COORD                              = $8453;   // GL 1.4
  GL_FOG_COORD_ARRAY_TYPE                           = $8454;   // GL 1.4
  GL_FOG_COORD_ARRAY_STRIDE                         = $8455;   // GL 1.4
  GL_FOG_COORD_ARRAY_POINTER                        = $8456;   // GL 1.4
  GL_FOG_COORD_ARRAY                                = $8457;   // GL 1.4
  GL_POINT_SIZE_MIN                                 = $8126;   // GL 1.4
  GL_POINT_SIZE_MAX                                 = $8127;   // GL 1.4
  GL_POINT_FADE_THRESHOLD_SIZE                      = $8128;   // GL 1.4
  GL_POINT_DISTANCE_ATTENUATION                     = $8129;   // GL 1.4
  GL_COLOR_SUM                                      = $8458;   // GL 1.4
  GL_CURRENT_SECONDARY_COLOR                        = $8459;   // GL 1.4
  GL_SECONDARY_COLOR_ARRAY_SIZE                     = $845A;   // GL 1.4
  GL_SECONDARY_COLOR_ARRAY_TYPE                     = $845B;   // GL 1.4
  GL_SECONDARY_COLOR_ARRAY_STRIDE                   = $845C;   // GL 1.4
  GL_SECONDARY_COLOR_ARRAY_POINTER                  = $845D;   // GL 1.4
  GL_SECONDARY_COLOR_ARRAY                          = $845E;   // GL 1.4
  GL_BLEND_DST_RGB                                  = $80C8;   // GL 1.4
  GL_BLEND_SRC_RGB                                  = $80C9;   // GL 1.4
  GL_BLEND_DST_ALPHA                                = $80CA;   // GL 1.4
  GL_BLEND_SRC_ALPHA                                = $80CB;   // GL 1.4
  GL_INCR_WRAP                                      = $8507;   // GL 1.4
  GL_DECR_WRAP                                      = $8508;   // GL 1.4
  GL_TEXTURE_FILTER_CONTROL                         = $8500;   // GL 1.4
  GL_TEXTURE_LOD_BIAS                               = $8501;   // GL 1.4
  GL_MAX_TEXTURE_LOD_BIAS                           = $84FD;   // GL 1.4
  GL_MIRRORED_REPEAT                                = $8370;   // GL 1.4

  // deprecated
  GL_FOG_COORDINATE_SOURCE = GL_FOG_COORD_SOURCE;
  GL_FOG_COORDINATE = GL_FOG_COORD;
  GL_CURRENT_FOG_COORDINATE = GL_CURRENT_FOG_COORD;
  GL_FOG_COORDINATE_ARRAY_TYPE = GL_FOG_COORD_ARRAY_TYPE;
  GL_FOG_COORDINATE_ARRAY_STRIDE = GL_FOG_COORD_ARRAY_STRIDE;
  GL_FOG_COORDINATE_ARRAY_POINTER = GL_FOG_COORD_ARRAY_POINTER;
  GL_FOG_COORDINATE_ARRAY = GL_FOG_COORD_ARRAY;
{$ENDIF}

{$IFDEF GL_15}
  // ********** 1.5 constants **********
  GL_ARRAY_BUFFER                                   = $8892;   // GL 1.5
  GL_ELEMENT_ARRAY_BUFFER                           = $8893;   // GL 1.5
  GL_ARRAY_BUFFER_BINDING                           = $8894;   // GL 1.5
  GL_ELEMENT_ARRAY_BUFFER_BINDING                   = $8895;   // GL 1.5
  GL_VERTEX_ARRAY_BUFFER_BINDING                    = $8896;   // GL 1.5
  GL_NORMAL_ARRAY_BUFFER_BINDING                    = $8897;   // GL 1.5
  GL_COLOR_ARRAY_BUFFER_BINDING                     = $8898;   // GL 1.5
  GL_INDEX_ARRAY_BUFFER_BINDING                     = $8899;   // GL 1.5
  GL_TEXTURE_COORD_ARRAY_BUFFER_BINDING             = $889A;   // GL 1.5
  GL_EDGE_FLAG_ARRAY_BUFFER_BINDING                 = $889B;   // GL 1.5
  GL_SECONDARY_COLOR_ARRAY_BUFFER_BINDING           = $889C;   // GL 1.5
  GL_FOG_COORD_ARRAY_BUFFER_BINDING                 = $889D;   // GL 1.5
  GL_WEIGHT_ARRAY_BUFFER_BINDING                    = $889E;   // GL 1.5
  GL_VERTEX_ATTRIB_ARRAY_BUFFER_BINDING             = $889F;   // GL 1.5
  GL_STREAM_DRAW                                    = $88E0;   // GL 1.5
  GL_STREAM_READ                                    = $88E1;   // GL 1.5
  GL_STREAM_COPY                                    = $88E2;   // GL 1.5
  GL_STATIC_DRAW                                    = $88E4;   // GL 1.5
  GL_STATIC_READ                                    = $88E5;   // GL 1.5
  GL_STATIC_COPY                                    = $88E6;   // GL 1.5
  GL_DYNAMIC_DRAW                                   = $88E8;   // GL 1.5
  GL_DYNAMIC_READ                                   = $88E9;   // GL 1.5
  GL_DYNAMIC_COPY                                   = $88EA;   // GL 1.5
  GL_READ_ONLY                                      = $88B8;   // GL 1.5
  GL_WRITE_ONLY                                     = $88B9;   // GL 1.5
  GL_READ_WRITE                                     = $88BA;   // GL 1.5
  GL_BUFFER_SIZE                                    = $8764;   // GL 1.5
  GL_BUFFER_USAGE                                   = $8765;   // GL 1.5
  GL_BUFFER_ACCESS                                  = $88BB;   // GL 1.5
  GL_BUFFER_MAPPED                                  = $88BC;   // GL 1.5
  GL_BUFFER_MAP_POINTER                             = $88BD;   // GL 1.5

  GL_SAMPLES_PASSED                                 = $8914;   // GL 1.5
  GL_QUERY_COUNTER_BITS                             = $8864;   // GL 1.5
  GL_CURRENT_QUERY                                  = $8865;   // GL 1.5
  GL_QUERY_RESULT                                   = $8866;   // GL 1.5
  GL_QUERY_RESULT_AVAILABLE                         = $8867;   // GL 1.5
{$ENDIF}

{$IFDEF GL_PROGRAM}
  // ********** GL_ARB_vertex_program, GL_ARB_fragment_program constants **********
  GL_VERTEX_PROGRAM_ARB                             = $8620;
  GL_VERTEX_PROGRAM_POINT_SIZE_ARB                  = $8642;
  GL_VERTEX_PROGRAM_TWO_SIDE_ARB                    = $8643;
  GL_COLOR_SUM_ARB                                  = $8458;
  GL_PROGRAM_FORMAT_ASCII_ARB                       = $8875;
  GL_VERTEX_ATTRIB_ARRAY_ENABLED_ARB                = $8622;
  GL_VERTEX_ATTRIB_ARRAY_SIZE_ARB                   = $8623;
  GL_VERTEX_ATTRIB_ARRAY_STRIDE_ARB                 = $8624;
  GL_VERTEX_ATTRIB_ARRAY_TYPE_ARB                   = $8625;
  GL_VERTEX_ATTRIB_ARRAY_NORMALIZED_ARB             = $886A;
  GL_CURRENT_VERTEX_ATTRIB_ARB                      = $8626;
  GL_VERTEX_ATTRIB_ARRAY_POINTER_ARB                = $8645;
  GL_PROGRAM_LENGTH_ARB                             = $8627;
  GL_PROGRAM_FORMAT_ARB                             = $8876;
  GL_PROGRAM_BINDING_ARB                            = $8677;
  GL_PROGRAM_INSTRUCTIONS_ARB                       = $88A0;
  GL_MAX_PROGRAM_INSTRUCTIONS_ARB                   = $88A1;
  GL_PROGRAM_NATIVE_INSTRUCTIONS_ARB                = $88A2;
  GL_MAX_PROGRAM_NATIVE_INSTRUCTIONS_ARB            = $88A3;
  GL_PROGRAM_TEMPORARIES_ARB                        = $88A4;
  GL_MAX_PROGRAM_TEMPORARIES_ARB                    = $88A5;
  GL_PROGRAM_NATIVE_TEMPORARIES_ARB                 = $88A6;
  GL_MAX_PROGRAM_NATIVE_TEMPORARIES_ARB             = $88A7;
  GL_PROGRAM_PARAMETERS_ARB                         = $88A8;
  GL_MAX_PROGRAM_PARAMETERS_ARB                     = $88A9;
  GL_PROGRAM_NATIVE_PARAMETERS_ARB                  = $88AA;
  GL_MAX_PROGRAM_NATIVE_PARAMETERS_ARB              = $88AB;
  GL_PROGRAM_ATTRIBS_ARB                            = $88AC;
  GL_MAX_PROGRAM_ATTRIBS_ARB                        = $88AD;
  GL_PROGRAM_NATIVE_ATTRIBS_ARB                     = $88AE;
  GL_MAX_PROGRAM_NATIVE_ATTRIBS_ARB                 = $88AF;
  GL_PROGRAM_ADDRESS_REGISTERS_ARB                  = $88B0;
  GL_MAX_PROGRAM_ADDRESS_REGISTERS_ARB              = $88B1;
  GL_PROGRAM_NATIVE_ADDRESS_REGISTERS_ARB           = $88B2;
  GL_MAX_PROGRAM_NATIVE_ADDRESS_REGISTERS_ARB       = $88B3;
  GL_MAX_PROGRAM_LOCAL_PARAMETERS_ARB               = $88B4;
  GL_MAX_PROGRAM_ENV_PARAMETERS_ARB                 = $88B5;
  GL_PROGRAM_UNDER_NATIVE_LIMITS_ARB                = $88B6;
  GL_PROGRAM_STRING_ARB                             = $8628;
  GL_PROGRAM_ERROR_POSITION_ARB                     = $864B;
  GL_CURRENT_MATRIX_ARB                             = $8641;
  GL_TRANSPOSE_CURRENT_MATRIX_ARB                   = $88B7;
  GL_CURRENT_MATRIX_STACK_DEPTH_ARB                 = $8640;
  GL_MAX_VERTEX_ATTRIBS_ARB                         = $8869;
  GL_MAX_PROGRAM_MATRICES_ARB                       = $862F;
  GL_MAX_PROGRAM_MATRIX_STACK_DEPTH_ARB             = $862E;
  GL_PROGRAM_ERROR_STRING_ARB                       = $8874;
  GL_MATRIX0_ARB                                    = $88C0;
  GL_MATRIX1_ARB                                    = $88C1;
  GL_MATRIX2_ARB                                    = $88C2;
  GL_MATRIX3_ARB                                    = $88C3;
  GL_MATRIX4_ARB                                    = $88C4;
  GL_MATRIX5_ARB                                    = $88C5;
  GL_MATRIX6_ARB                                    = $88C6;
  GL_MATRIX7_ARB                                    = $88C7;
  GL_MATRIX8_ARB                                    = $88C8;
  GL_MATRIX9_ARB                                    = $88C9;
  GL_MATRIX10_ARB                                   = $88CA;
  GL_MATRIX11_ARB                                   = $88CB;
  GL_MATRIX12_ARB                                   = $88CC;
  GL_MATRIX13_ARB                                   = $88CD;
  GL_MATRIX14_ARB                                   = $88CE;
  GL_MATRIX15_ARB                                   = $88CF;
  GL_MATRIX16_ARB                                   = $88D0;
  GL_MATRIX17_ARB                                   = $88D1;
  GL_MATRIX18_ARB                                   = $88D2;
  GL_MATRIX19_ARB                                   = $88D3;
  GL_MATRIX20_ARB                                   = $88D4;
  GL_MATRIX21_ARB                                   = $88D5;
  GL_MATRIX22_ARB                                   = $88D6;
  GL_MATRIX23_ARB                                   = $88D7;
  GL_MATRIX24_ARB                                   = $88D8;
  GL_MATRIX25_ARB                                   = $88D9;
  GL_MATRIX26_ARB                                   = $88DA;
  GL_MATRIX27_ARB                                   = $88DB;
  GL_MATRIX28_ARB                                   = $88DC;
  GL_MATRIX29_ARB                                   = $88DD;
  GL_MATRIX30_ARB                                   = $88DE;
  GL_MATRIX31_ARB                                   = $88DF;

  GL_FRAGMENT_PROGRAM_ARB                           = $8804;
  GL_PROGRAM_ALU_INSTRUCTIONS_ARB                   = $8805;
  GL_PROGRAM_TEX_INSTRUCTIONS_ARB                   = $8806;
  GL_PROGRAM_TEX_INDIRECTIONS_ARB                   = $8807;
  GL_PROGRAM_NATIVE_ALU_INSTRUCTIONS_ARB            = $8808;
  GL_PROGRAM_NATIVE_TEX_INSTRUCTIONS_ARB            = $8809;
  GL_PROGRAM_NATIVE_TEX_INDIRECTIONS_ARB            = $880A;
  GL_MAX_PROGRAM_ALU_INSTRUCTIONS_ARB               = $880B;
  GL_MAX_PROGRAM_TEX_INSTRUCTIONS_ARB               = $880C;
  GL_MAX_PROGRAM_TEX_INDIRECTIONS_ARB               = $880D;
  GL_MAX_PROGRAM_NATIVE_ALU_INSTRUCTIONS_ARB        = $880E;
  GL_MAX_PROGRAM_NATIVE_TEX_INSTRUCTIONS_ARB        = $880F;
  GL_MAX_PROGRAM_NATIVE_TEX_INDIRECTIONS_ARB        = $8810;
  GL_MAX_TEXTURE_COORDS_ARB                         = $8871;
  GL_MAX_TEXTURE_IMAGE_UNITS_ARB                    = $8872;
{$ENDIF}

{$IFDEF GL_SHADER}
  // ********** GL_ARB_shader_objects, GL_ARB_vertex_shader, GL_ARB_fragment_shader constants **********

   // GL_ARB_vertex_shader
  GL_VERTEX_SHADER_ARB                               = $8B31;
  GL_MAX_VERTEX_UNIFORM_COMPONENTS_ARB               = $8B4A;
  GL_MAX_VARYING_FLOATS_ARB                          = $8B4B;
  GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS_ARB              = $8B4C;
  GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS_ARB            = $8B4D;
  GL_OBJECT_ACTIVE_ATTRIBUTES_ARB                    = $8B89;
  GL_OBJECT_ACTIVE_ATTRIBUTE_MAX_LENGTH_ARB          = $8B8A;

  // GL_ARB_fragment_shader
  GL_FRAGMENT_SHADER_ARB                             = $8B30;

  // GL_ARB_shader_objects
  GL_PROGRAM_OBJECT_ARB                              = $8B40;

  GL_OBJECT_TYPE_ARB                                 = $8B4E;
  GL_OBJECT_SUBTYPE_ARB                              = $8B4F;
  GL_OBJECT_DELETE_STATUS_ARB                        = $8B80;
  GL_OBJECT_COMPILE_STATUS_ARB                       = $8B81;
  GL_OBJECT_LINK_STATUS_ARB                          = $8B82;
  GL_OBJECT_VALIDATE_STATUS_ARB                      = $8B83;
  GL_OBJECT_INFO_LOG_LENGTH_ARB                      = $8B84;
  GL_OBJECT_ATTACHED_OBJECTS_ARB                     = $8B85;
  GL_OBJECT_ACTIVE_UNIFORMS_ARB                      = $8B86;
  GL_OBJECT_ACTIVE_UNIFORM_MAX_LENGTH_ARB            = $8B87;
  GL_OBJECT_SHADER_SOURCE_LENGTH_ARB                 = $8B88;

  GL_SHADER_OBJECT_ARB                               = $8B48;

  GL_FLOAT_VEC2_ARB                                  = $8B50;
  GL_FLOAT_VEC3_ARB                                  = $8B51;
  GL_FLOAT_VEC4_ARB                                  = $8B52;
  GL_INT_VEC2_ARB                                    = $8B53;
  GL_INT_VEC3_ARB                                    = $8B54;
  GL_INT_VEC4_ARB                                    = $8B55;
  GL_BOOL_ARB                                        = $8B56;
  GL_BOOL_VEC2_ARB                                   = $8B57;
  GL_BOOL_VEC3_ARB                                   = $8B58;
  GL_BOOL_VEC4_ARB                                   = $8B59;
  GL_FLOAT_MAT2_ARB                                  = $8B5A;
  GL_FLOAT_MAT3_ARB                                  = $8B5B;
  GL_FLOAT_MAT4_ARB                                  = $8B5C;
{$ENDIF}

{$IFNDEF NO_GLU}
  // ********** GLU generic constants **********

  // Errors: (return value 0 = no error)
  GLU_INVALID_ENUM                                  = 100900;
  GLU_INVALID_VALUE                                 = 100901;
  GLU_OUT_OF_MEMORY                                 = 100902;
  GLU_INCOMPATIBLE_GL_VERSION                       = 100903;

  // StringName
  GLU_VERSION                                       = 100800;
  GLU_EXTENSIONS                                    = 100801;

  // Boolean
  GLU_TRUE                                          = GL_TRUE;
  GLU_FALSE                                         = GL_FALSE;

  // Quadric constants
  // QuadricNormal
  GLU_SMOOTH                                        = 100000;
  GLU_FLAT                                          = 100001;
  GLU_NONE                                          = 100002;

  // QuadricDrawStyle
  GLU_POINT                                         = 100010;
  GLU_LINE                                          = 100011;
  GLU_FILL                                          = 100012;
  GLU_SILHOUETTE                                    = 100013;

  // QuadricOrientation
  GLU_OUTSIDE                                       = 100020;
  GLU_INSIDE                                        = 100021;

  // Tesselation constants
  GLU_TESS_MAX_COORD                                = 1.0e150;

  // TessProperty
  GLU_TESS_WINDING_RULE                             = 100140;
  GLU_TESS_BOUNDARY_ONLY                            = 100141;
  GLU_TESS_TOLERANCE                                = 100142;

  // TessWinding
  GLU_TESS_WINDING_ODD                              = 100130;
  GLU_TESS_WINDING_NONZERO                          = 100131;
  GLU_TESS_WINDING_POSITIVE                         = 100132;
  GLU_TESS_WINDING_NEGATIVE                         = 100133;
  GLU_TESS_WINDING_ABS_GEQ_TWO                      = 100134;

  // TessCallback
  GLU_TESS_BEGIN                                    = 100100; // GLUTessBeginProc
  GLU_TESS_VERTEX                                   = 100101; // GLUTessVertexProc
  GLU_TESS_END                                      = 100102; // GLUTessEndProc
  GLU_TESS_ERROR                                    = 100103; // GLUTessErrorProc
  GLU_TESS_EDGE_FLAG                                = 100104; // GLUTessEdgeFlagProc
  GLU_TESS_COMBINE                                  = 100105; // GLUTessCombineProc
  GLU_TESS_BEGIN_DATA                               = 100106; // GLUTessBeginDataProc
  GLU_TESS_VERTEX_DATA                              = 100107; // GLUTessVertexDataProc
  GLU_TESS_END_DATA                                 = 100108; // GLUTessEndDataProc
  GLU_TESS_ERROR_DATA                               = 100109; // GLUTessErrorDataProc
  GLU_TESS_EDGE_FLAG_DATA                           = 100110; // GLUTessEdgeFlagDataProc
  GLU_TESS_COMBINE_DATA                             = 100111; // GLUTessCombineDataProc

  // TessError
  GLU_TESS_ERROR1                                   = 100151;
  GLU_TESS_ERROR2                                   = 100152;
  GLU_TESS_ERROR3                                   = 100153;
  GLU_TESS_ERROR4                                   = 100154;
  GLU_TESS_ERROR5                                   = 100155;
  GLU_TESS_ERROR6                                   = 100156;
  GLU_TESS_ERROR7                                   = 100157;
  GLU_TESS_ERROR8                                   = 100158;

  GLU_TESS_MISSING_BEGIN_POLYGON                    = GLU_TESS_ERROR1;
  GLU_TESS_MISSING_BEGIN_CONTOUR                    = GLU_TESS_ERROR2;
  GLU_TESS_MISSING_END_POLYGON                      = GLU_TESS_ERROR3;
  GLU_TESS_MISSING_END_CONTOUR                      = GLU_TESS_ERROR4;
  GLU_TESS_COORD_TOO_LARGE                          = GLU_TESS_ERROR5;
  GLU_TESS_NEED_COMBINE_CALLBACK                    = GLU_TESS_ERROR6;

  // NURBS constants

  // NurbsProperty
  GLU_AUTO_LOAD_MATRIX                              = 100200;
  GLU_CULLING                                       = 100201;
  GLU_SAMPLING_TOLERANCE                            = 100203;
  GLU_DISPLAY_MODE                                  = 100204;
  GLU_PARAMETRIC_TOLERANCE                          = 100202;
  GLU_SAMPLING_METHOD                               = 100205;
  GLU_U_STEP                                        = 100206;
  GLU_V_STEP                                        = 100207;

  // NurbsSampling
  GLU_PATH_LENGTH                                   = 100215;
  GLU_PARAMETRIC_ERROR                              = 100216;
  GLU_DOMAIN_DISTANCE                               = 100217;

  // NurbsTrim
  GLU_MAP1_TRIM_2                                   = 100210;
  GLU_MAP1_TRIM_3                                   = 100211;

  // NurbsDisplay
  GLU_OUTLINE_POLYGON                               = 100240;
  GLU_OUTLINE_PATCH                                 = 100241;

  // NurbsErrors
  GLU_NURBS_ERROR1                                  = 100251;
  GLU_NURBS_ERROR2                                  = 100252;
  GLU_NURBS_ERROR3                                  = 100253;
  GLU_NURBS_ERROR4                                  = 100254;
  GLU_NURBS_ERROR5                                  = 100255;
  GLU_NURBS_ERROR6                                  = 100256;
  GLU_NURBS_ERROR7                                  = 100257;
  GLU_NURBS_ERROR8                                  = 100258;
  GLU_NURBS_ERROR9                                  = 100259;
  GLU_NURBS_ERROR10                                 = 100260;
  GLU_NURBS_ERROR11                                 = 100261;
  GLU_NURBS_ERROR12                                 = 100262;
  GLU_NURBS_ERROR13                                 = 100263;
  GLU_NURBS_ERROR14                                 = 100264;
  GLU_NURBS_ERROR15                                 = 100265;
  GLU_NURBS_ERROR16                                 = 100266;
  GLU_NURBS_ERROR17                                 = 100267;
  GLU_NURBS_ERROR18                                 = 100268;
  GLU_NURBS_ERROR19                                 = 100269;
  GLU_NURBS_ERROR20                                 = 100270;
  GLU_NURBS_ERROR21                                 = 100271;
  GLU_NURBS_ERROR22                                 = 100272;
  GLU_NURBS_ERROR23                                 = 100273;
  GLU_NURBS_ERROR24                                 = 100274;
  GLU_NURBS_ERROR25                                 = 100275;
  GLU_NURBS_ERROR26                                 = 100276;
  GLU_NURBS_ERROR27                                 = 100277;
  GLU_NURBS_ERROR28                                 = 100278;
  GLU_NURBS_ERROR29                                 = 100279;
  GLU_NURBS_ERROR30                                 = 100280;
  GLU_NURBS_ERROR31                                 = 100281;
  GLU_NURBS_ERROR32                                 = 100282;
  GLU_NURBS_ERROR33                                 = 100283;
  GLU_NURBS_ERROR34                                 = 100284;
  GLU_NURBS_ERROR35                                 = 100285;
  GLU_NURBS_ERROR36                                 = 100286;
  GLU_NURBS_ERROR37                                 = 100287;

  // Contours types -- obsolete!
  GLU_CW                                            = 100120;
  GLU_CCW                                           = 100121;
  GLU_INTERIOR                                      = 100122;
  GLU_EXTERIOR                                      = 100123;
  GLU_UNKNOWN                                       = 100124;

  // Names without "TESS_" prefix
  GLU_BEGIN                                         = GLU_TESS_BEGIN;
  GLU_VERTEX                                        = GLU_TESS_VERTEX;
  GLU_END                                           = GLU_TESS_END;
  GLU_ERROR                                         = GLU_TESS_ERROR;
  GLU_EDGE_FLAG                                     = GLU_TESS_EDGE_FLAG;


type
  // GLU types
  GLUNurbs = record end;
  GLUQuadric = record end;
  GLUTesselator = record end;

  PGLUNurbs = ^GLUNurbs;
  PGLUQuadric = ^GLUQuadric;
  PGLUTesselator = ^GLUTesselator;

  // backwards compatibility
  GLUNurbsObj = GLUNurbs;
  GLUQuadricObj = GLUQuadric;
  GLUTesselatorObj = GLUTesselator;
  GLUTriangulatorObj = GLUTesselator;

  PGLUNurbsObj = PGLUNurbs;
  PGLUQuadricObj = PGLUQuadric;
  PGLUTesselatorObj = PGLUTesselator;
  PGLUTriangulatorObj = PGLUTesselator;

  // Callback function prototypes
  // GLUQuadricCallback
  GLUQuadricErrorProc = procedure(errorCode: GLEnum); stdcall;

  // GLUTessCallback
  GLUTessBeginProc = procedure(AType: GLEnum); stdcall;
  GLUTessEdgeFlagProc = procedure(Flag: GLboolean); stdcall;
  GLUTessVertexProc = procedure(VertexData: Pointer); stdcall;
  GLUTessEndProc = procedure; stdcall;
  GLUTessErrorProc = procedure(ErrNo: GLEnum); stdcall;
  GLUTessCombineProc = procedure(Coords: TVector3d; VertexData: TVector4p; Weight: TVector4f; OutData: PPointer); stdcall;
  GLUTessBeginDataProc = procedure(AType: GLEnum; UserData: Pointer); stdcall;
  GLUTessEdgeFlagDataProc = procedure(Flag: GLboolean; UserData: Pointer); stdcall;
  GLUTessVertexDataProc = procedure(VertexData: Pointer; UserData: Pointer); stdcall;
  GLUTessEndDataProc = procedure(UserData: Pointer); stdcall;
  GLUTessErrorDataProc = procedure(ErrNo: GLEnum; UserData: Pointer); stdcall;
  GLUTessCombineDataProc = procedure(Coords: TVector3d; VertexData: TVector4p; Weight: TVector4f; OutData: PPointer; UserData: Pointer); stdcall;

  // GLUNurbsCallback
  GLUNurbsErrorProc = procedure(ErrorCode: GLEnum); stdcall;
{$ENDIF}

var
  // GL functions and procedures
  glAccum: procedure(op: GLuint; value: GLfloat); stdcall;
  glAlphaFunc: procedure(func: GLEnum; ref: GLclampf); stdcall;
  glAreTexturesResident: function(n: GLsizei; Textures: PGLuint; residences: PGLboolean): GLboolean; stdcall;
  glArrayElement: procedure(i: GLint); stdcall;
  glBegin: procedure(mode: GLEnum); stdcall;
  glBindTexture: procedure(target: GLEnum; texture: GLuint); stdcall;
  glBitmap: procedure(width: GLsizei; height: GLsizei; xorig, yorig: GLfloat; xmove: GLfloat; ymove: GLfloat; bitmap: Pointer); stdcall;
  glBlendFunc: procedure(sfactor: GLEnum; dfactor: GLEnum); stdcall;
  glCallList: procedure(list: GLuint); stdcall;
  glCallLists: procedure(n: GLsizei; atype: GLEnum; lists: Pointer); stdcall;
  glClear: procedure(mask: GLbitfield); stdcall;
  glClearAccum: procedure(red, green, blue, alpha: GLfloat); stdcall;
  glClearColor: procedure(red, green, blue, alpha: GLclampf); stdcall;
  glClearDepth: procedure(depth: GLclampd); stdcall;
  glClearIndex: procedure(c: GLfloat); stdcall;
  glClearStencil: procedure(s: GLint ); stdcall;
  glClipPlane: procedure(plane: GLEnum; equation: PGLdouble); stdcall;
  glColor3b: procedure(red, green, blue: GLbyte); stdcall;
  glColor3bv: procedure(v: PGLbyte); stdcall;
  glColor3d: procedure(red, green, blue: GLdouble); stdcall;
  glColor3dv: procedure(v: PGLdouble); stdcall;
  glColor3f: procedure(red, green, blue: GLfloat); stdcall;
  glColor3fv: procedure(v: PGLfloat); stdcall;
  glColor3i: procedure(red, green, blue: GLint); stdcall;
  glColor3iv: procedure(v: PGLint); stdcall;
  glColor3s: procedure(red, green, blue: GLshort); stdcall;
  glColor3sv: procedure(v: PGLshort); stdcall;
  glColor3ub: procedure(red, green, blue: GLubyte); stdcall;
  glColor3ubv: procedure(v: PGLubyte); stdcall;
  glColor3ui: procedure(red, green, blue: GLuint); stdcall;
  glColor3uiv: procedure(v: PGLuint); stdcall;
  glColor3us: procedure(red, green, blue: GLushort); stdcall;
  glColor3usv: procedure(v: PGLushort); stdcall;
  glColor4b: procedure(red, green, blue, alpha: GLbyte); stdcall;
  glColor4bv: procedure(v: PGLbyte); stdcall;
  glColor4d: procedure(red, green, blue, alpha: GLdouble ); stdcall;
  glColor4dv: procedure(v: PGLdouble); stdcall;
  glColor4f: procedure(red, green, blue, alpha: GLfloat); stdcall;
  glColor4fv: procedure(v: PGLfloat); stdcall;
  glColor4i: procedure(red, green, blue, alpha: GLint); stdcall;
  glColor4iv: procedure(v: PGLint); stdcall;
  glColor4s: procedure(red, green, blue, alpha: GLshort); stdcall;
  glColor4sv: procedure(v: PGLshort); stdcall;
  glColor4ub: procedure(red, green, blue, alpha: GLubyte); stdcall;
  glColor4ubv: procedure(v: PGLubyte); stdcall;
  glColor4ui: procedure(red, green, blue, alpha: GLuint); stdcall;
  glColor4uiv: procedure(v: PGLuint); stdcall;
  glColor4us: procedure(red, green, blue, alpha: GLushort); stdcall;
  glColor4usv: procedure(v: PGLushort); stdcall;
  glColorMask: procedure(red, green, blue, alpha: GLboolean); stdcall;
  glColorMaterial: procedure(face: GLEnum; mode: GLEnum); stdcall;
  glColorPointer: procedure(size: GLint; atype: GLEnum; stride: GLsizei; data: pointer); stdcall;
  glCopyPixels: procedure(x, y: GLint; width, height: GLsizei; atype: GLEnum); stdcall;
  glCopyTexImage1D: procedure(target: GLEnum; level: GLint; internalFormat: GLEnum; x, y: GLint; width: GLsizei; border: GLint); stdcall;
  glCopyTexImage2D: procedure(target: GLEnum; level: GLint; internalFormat: GLEnum; x, y: GLint; width, height: GLsizei; border: GLint); stdcall;
  glCopyTexSubImage1D: procedure(target: GLEnum; level, xoffset, x, y: GLint; width: GLsizei); stdcall;
  glCopyTexSubImage2D: procedure(target: GLEnum; level, xoffset, yoffset, x, y: GLint; width, height: GLsizei); stdcall;
  glCullFace: procedure(mode: GLEnum); stdcall;
  glDeleteLists: procedure(list: GLuint; range: GLsizei); stdcall;
  glDeleteTextures: procedure(n: GLsizei; textures: PGLuint); stdcall;
  glDepthFunc: procedure(func: GLEnum); stdcall;
  glDepthMask: procedure(flag: GLboolean); stdcall;
  glDepthRange: procedure(zNear, zFar: GLclampd); stdcall;
  glDisable: procedure(cap: GLEnum); stdcall;
  glDisableClientState: procedure(aarray: GLEnum); stdcall;
  glDrawArrays: procedure(mode: GLEnum; first: GLint; count: GLsizei); stdcall;
  glDrawBuffer: procedure(mode: GLEnum); stdcall;
  glDrawElements: procedure(mode: GLEnum; count: GLsizei; atype: GLEnum; indices: Pointer); stdcall;
  glDrawPixels: procedure(width, height: GLsizei; format, atype: GLEnum; pixels: Pointer); stdcall;
  glEdgeFlag: procedure(flag: GLboolean); stdcall;
  glEdgeFlagPointer: procedure(stride: GLsizei; data: pointer); stdcall;
  glEdgeFlagv: procedure(flag: PGLboolean); stdcall;
  glEnable: procedure(cap: GLEnum); stdcall;
  glEnableClientState: procedure(aarray: GLEnum); stdcall;
  glEnd: procedure; stdcall;
  glEndList: procedure; stdcall;
  glEvalCoord1d: procedure(u: GLdouble); stdcall;
  glEvalCoord1dv: procedure(u: PGLdouble); stdcall;
  glEvalCoord1f: procedure(u: GLfloat); stdcall;
  glEvalCoord1fv: procedure(u: PGLfloat); stdcall;
  glEvalCoord2d: procedure(u: GLdouble; v: GLdouble); stdcall;
  glEvalCoord2dv: procedure(u: PGLdouble); stdcall;
  glEvalCoord2f: procedure(u, v: GLfloat); stdcall;
  glEvalCoord2fv: procedure(u: PGLfloat); stdcall;
  glEvalMesh1: procedure(mode: GLEnum; i1, i2: GLint); stdcall;
  glEvalMesh2: procedure(mode: GLEnum; i1, i2, j1, j2: GLint); stdcall;
  glEvalPoint1: procedure(i: GLint); stdcall;
  glEvalPoint2: procedure(i, j: GLint); stdcall;
  glFeedbackBuffer: procedure(size: GLsizei; atype: GLEnum; buffer: PGLfloat); stdcall;
  glFinish: procedure; stdcall;
  glFlush: procedure; stdcall;
  glFogf: procedure(pname: GLEnum; param: GLfloat); stdcall;
  glFogfv: procedure(pname: GLEnum; params: PGLfloat); stdcall;
  glFogi: procedure(pname: GLEnum; param: GLint); stdcall;
  glFogiv: procedure(pname: GLEnum; params: PGLint); stdcall;
  glFrontFace: procedure(mode: GLEnum); stdcall;
  glFrustum: procedure(left, right, bottom, top, zNear, zFar: GLdouble); stdcall;
  glGenLists: function(range: GLsizei): GLuint; stdcall;
  glGenTextures: procedure(n: GLsizei; textures: PGLuint); stdcall;
  glGetBooleanv: procedure(pname: GLEnum; params: PGLboolean); stdcall;
  glGetClipPlane: procedure(plane: GLEnum; equation: PGLdouble); stdcall;
  glGetDoublev: procedure(pname: GLEnum; params: PGLdouble); stdcall;
  glGetError: function: GLuint; stdcall;
  glGetFloatv: procedure(pname: GLEnum; params: PGLfloat); stdcall;
  glGetIntegerv: procedure(pname: GLEnum; params: PGLint); stdcall;
  glGetLightfv: procedure(light, pname: GLEnum; params: PGLfloat); stdcall;
  glGetLightiv: procedure(light, pname: GLEnum; params: PGLint); stdcall;
  glGetMapdv: procedure(target, query: GLEnum; v: PGLdouble); stdcall;
  glGetMapfv: procedure(target, query: GLEnum; v: PGLfloat); stdcall;
  glGetMapiv: procedure(target, query: GLEnum; v: PGLint); stdcall;
  glGetMaterialfv: procedure(face, pname: GLEnum; params: PGLfloat); stdcall;
  glGetMaterialiv: procedure(face, pname: GLEnum; params: PGLint); stdcall;
  glGetPixelMapfv: procedure(map: GLEnum; values: PGLfloat); stdcall;
  glGetPixelMapuiv: procedure(map: GLEnum; values: PGLuint); stdcall;
  glGetPixelMapusv: procedure(map: GLEnum; values: PGLushort); stdcall;
  glGetPointerv: procedure(pname: GLEnum; params: Pointer); stdcall;
  glGetPolygonStipple: procedure(mask: PGLubyte); stdcall;
  glGetString: function(name: GLEnum): PChar; stdcall;
  glGetTexEnvfv: procedure(target, pname: GLEnum; params: PGLfloat); stdcall;
  glGetTexEnviv: procedure(target, pname: GLEnum; params: PGLint); stdcall;
  glGetTexGendv: procedure(coord, pname: GLEnum; params: PGLdouble); stdcall;
  glGetTexGenfv: procedure(coord, pname: GLEnum; params: PGLfloat); stdcall;
  glGetTexGeniv: procedure(coord, pname: GLEnum; params: PGLint); stdcall;
  glGetTexImage: procedure(target: GLEnum; level: GLint; format, atype: GLEnum; pixels: Pointer); stdcall;
  glGetTexLevelParameterfv: procedure(target: GLEnum; level: GLint; pname: GLEnum; params: PGLfloat); stdcall;
  glGetTexLevelParameteriv: procedure(target: GLEnum; level: GLint; pname: GLEnum; params: PGLint); stdcall;
  glGetTexParameterfv: procedure(target, pname: GLEnum; params: PGLfloat); stdcall;
  glGetTexParameteriv: procedure(target, pname: GLEnum; params: PGLint); stdcall;
  glHint: procedure(target, mode: GLEnum); stdcall;
  glIndexMask: procedure(mask: GLuint); stdcall;
  glIndexPointer: procedure(atype: GLEnum; stride: GLsizei; data: pointer); stdcall;
  glIndexd: procedure(c: GLdouble); stdcall;
  glIndexdv: procedure(c: PGLdouble); stdcall;
  glIndexf: procedure(c: GLfloat); stdcall;
  glIndexfv: procedure(c: PGLfloat); stdcall;
  glIndexi: procedure(c: GLint); stdcall;
  glIndexiv: procedure(c: PGLint); stdcall;
  glIndexs: procedure(c: GLshort); stdcall;
  glIndexsv: procedure(c: PGLshort); stdcall;
  glIndexub: procedure(c: GLubyte); stdcall;
  glIndexubv: procedure(c: PGLubyte); stdcall;
  glInitNames: procedure; stdcall;
  glInterleavedArrays: procedure(format: GLEnum; stride: GLsizei; data: pointer); stdcall;
  glIsEnabled: function(cap: GLEnum): GLboolean; stdcall;
  glIsList: function(list: GLuint): GLboolean; stdcall;
  glIsTexture: function(texture: GLuint): GLboolean; stdcall;
  glLightModelf: procedure(pname: GLEnum; param: GLfloat); stdcall;
  glLightModelfv: procedure(pname: GLEnum; params: PGLfloat); stdcall;
  glLightModeli: procedure(pname: GLEnum; param: GLint); stdcall;
  glLightModeliv: procedure(pname: GLEnum; params: PGLint); stdcall;
  glLightf: procedure(light, pname: GLEnum; param: GLfloat); stdcall;
  glLightfv: procedure(light, pname: GLEnum; params: PGLfloat); stdcall;
  glLighti: procedure(light, pname: GLEnum; param: GLint); stdcall;
  glLightiv: procedure(light, pname: GLEnum; params: PGLint); stdcall;
  glLineStipple: procedure(factor: GLint; pattern: GLushort); stdcall;
  glLineWidth: procedure(width: GLfloat); stdcall;
  glListBase: procedure(base: GLuint); stdcall;
  glLoadIdentity: procedure; stdcall;
  glLoadMatrixd: procedure(m: PGLdouble); stdcall;
  glLoadMatrixf: procedure(m: PGLfloat); stdcall;
  glLoadName: procedure(name: GLuint); stdcall;
  glLogicOp: procedure(opcode: GLEnum); stdcall;
  glMap1d: procedure(target: GLEnum; u1, u2: GLdouble; stride, order: GLint; points: PGLdouble); stdcall;
  glMap1f: procedure(target: GLEnum; u1, u2: GLfloat; stride, order: GLint; points: PGLfloat);   stdcall;
  glMap2d: procedure(target: GLEnum; u1, u2: GLdouble; ustride, uorder: GLint; v1, v2: GLdouble; vstride,
    vorder: GLint; points: PGLdouble); stdcall;
  glMap2f: procedure(target: GLEnum; u1, u2: GLfloat; ustride, uorder: GLint; v1, v2: GLfloat; vstride,
    vorder: GLint; points: PGLfloat); stdcall;
  glMapGrid1d: procedure(un: GLint; u1, u2: GLdouble); stdcall;
  glMapGrid1f: procedure(un: GLint; u1, u2: GLfloat); stdcall;
  glMapGrid2d: procedure(un: GLint; u1, u2: GLdouble; vn: GLint; v1, v2: GLdouble); stdcall;
  glMapGrid2f: procedure(un: GLint; u1, u2: GLfloat; vn: GLint; v1, v2: GLfloat); stdcall;
  glMaterialf: procedure(face, pname: GLEnum; param: GLfloat); stdcall;
  glMaterialfv: procedure(face, pname: GLEnum; params: PGLfloat); stdcall;
  glMateriali: procedure(face, pname: GLEnum; param: GLint); stdcall;
  glMaterialiv: procedure(face, pname: GLEnum; params: PGLint); stdcall;
  glMatrixMode: procedure(mode: GLEnum); stdcall;
  glMultMatrixd: procedure(m: PGLdouble); stdcall;
  glMultMatrixf: procedure(m: PGLfloat); stdcall;
  glNewList: procedure(list: GLuint; mode: GLEnum); stdcall;
  glNormal3b: procedure(nx, ny, nz: GLbyte); stdcall;
  glNormal3bv: procedure(v: PGLbyte); stdcall;
  glNormal3d: procedure(nx, ny, nz: GLdouble); stdcall;
  glNormal3dv: procedure(v: PGLdouble); stdcall;
  glNormal3f: procedure(nx, ny, nz: GLfloat); stdcall;
  glNormal3fv: procedure(v: PGLfloat); stdcall;
  glNormal3i: procedure(nx, ny, nz: GLint); stdcall;
  glNormal3iv: procedure(v: PGLint); stdcall;
  glNormal3s: procedure(nx, ny, nz: GLshort); stdcall;
  glNormal3sv: procedure(v: PGLshort); stdcall;
  glNormalPointer: procedure(atype: GLEnum; stride: GLsizei; data: pointer); stdcall;
  glOrtho: procedure(left, right, bottom, top, zNear, zFar: GLdouble); stdcall;
  glPassThrough: procedure(token: GLfloat); stdcall;
  glPixelMapfv: procedure(map: GLEnum; mapsize: GLsizei; values: PGLfloat); stdcall;
  glPixelMapuiv: procedure(map: GLEnum; mapsize: GLsizei; values: PGLuint); stdcall;
  glPixelMapusv: procedure(map: GLEnum; mapsize: GLsizei; values: PGLushort); stdcall;
  glPixelStoref: procedure(pname: GLEnum; param: GLfloat); stdcall;
  glPixelStorei: procedure(pname: GLEnum; param: GLint); stdcall;
  glPixelTransferf: procedure(pname: GLEnum; param: GLfloat); stdcall;
  glPixelTransferi: procedure(pname: GLEnum; param: GLint); stdcall;
  glPixelZoom: procedure(xfactor, yfactor: GLfloat); stdcall;
  glPointSize: procedure(size: GLfloat); stdcall;
  glPolygonMode: procedure(face, mode: GLEnum); stdcall;
  glPolygonOffset: procedure(factor, units: GLfloat); stdcall;
  glPolygonStipple: procedure(mask: PGLubyte); stdcall;
  glPopAttrib: procedure; stdcall;
  glPopClientAttrib: procedure; stdcall;
  glPopMatrix: procedure; stdcall;
  glPopName: procedure; stdcall;
  glPrioritizeTextures: procedure(n: GLsizei; textures: PGLuint; priorities: PGLclampf); stdcall;
  glPushAttrib: procedure(mask: GLbitfield); stdcall;
  glPushClientAttrib: procedure(mask: GLbitfield); stdcall;
  glPushMatrix: procedure; stdcall;
  glPushName: procedure(name: GLuint); stdcall;
  glRasterPos2d: procedure(x, y: GLdouble); stdcall;
  glRasterPos2dv: procedure(v: PGLdouble); stdcall;
  glRasterPos2f: procedure(x, y: GLfloat); stdcall;
  glRasterPos2fv: procedure(v: PGLfloat); stdcall;
  glRasterPos2i: procedure(x, y: GLint); stdcall;
  glRasterPos2iv: procedure(v: PGLint); stdcall;
  glRasterPos2s: procedure(x, y: PGLshort); stdcall;
  glRasterPos2sv: procedure(v: PGLshort); stdcall;
  glRasterPos3d: procedure(x, y, z: GLdouble); stdcall;
  glRasterPos3dv: procedure(v: PGLdouble); stdcall;
  glRasterPos3f: procedure(x, y, z: GLfloat); stdcall;
  glRasterPos3fv: procedure(v: PGLfloat); stdcall;
  glRasterPos3i: procedure(x, y, z: GLint); stdcall;
  glRasterPos3iv: procedure(v: PGLint); stdcall;
  glRasterPos3s: procedure(x, y, z: GLshort); stdcall;
  glRasterPos3sv: procedure(v: PGLshort); stdcall;
  glRasterPos4d: procedure(x, y, z, w: GLdouble); stdcall;
  glRasterPos4dv: procedure(v: PGLdouble); stdcall;
  glRasterPos4f: procedure(x, y, z, w: GLfloat); stdcall;
  glRasterPos4fv: procedure(v: PGLfloat); stdcall;
  glRasterPos4i: procedure(x, y, z, w: GLint); stdcall;
  glRasterPos4iv: procedure(v: PGLint); stdcall;
  glRasterPos4s: procedure(x, y, z, w: GLshort); stdcall;
  glRasterPos4sv: procedure(v: PGLshort); stdcall;
  glReadBuffer: procedure(mode: GLEnum); stdcall;
  glReadPixels: procedure(x, y: GLint; width, height: GLsizei; format, atype: GLEnum; pixels: Pointer); stdcall;
  glRectd: procedure(x1, y1, x2, y2: GLdouble); stdcall;
  glRectdv: procedure(v1, v2: PGLdouble); stdcall;
  glRectf: procedure(x1, y1, x2, y2: GLfloat); stdcall;
  glRectfv: procedure(v1, v2: PGLfloat); stdcall;
  glRecti: procedure(x1, y1, x2, y2: GLint); stdcall;
  glRectiv: procedure(v1, v2: PGLint); stdcall;
  glRects: procedure(x1, y1, x2, y2: GLshort); stdcall;
  glRectsv: procedure(v1, v2: PGLshort); stdcall;
  glRenderMode: function(mode: GLEnum): GLint; stdcall;
  glRotated: procedure(angle, x, y, z: GLdouble); stdcall;
  glRotatef: procedure(angle, x, y, z: GLfloat); stdcall;
  glScaled: procedure(x, y, z: GLdouble); stdcall;
  glScalef: procedure(x, y, z: GLfloat); stdcall;
  glScissor: procedure(x, y: GLint; width, height: GLsizei); stdcall;
  glSelectBuffer: procedure(size: GLsizei; buffer: PGLuint); stdcall;
  glShadeModel: procedure(mode: GLEnum); stdcall;
  glStencilFunc: procedure(func: GLEnum; ref: GLint; mask: GLuint); stdcall;
  glStencilMask: procedure(mask: GLuint); stdcall;
  glStencilOp: procedure(fail, zfail, zpass: GLEnum); stdcall;
  glTexCoord1d: procedure(s: GLdouble); stdcall;
  glTexCoord1dv: procedure(v: PGLdouble); stdcall;
  glTexCoord1f: procedure(s: GLfloat); stdcall;
  glTexCoord1fv: procedure(v: PGLfloat); stdcall;
  glTexCoord1i: procedure(s: GLint); stdcall;
  glTexCoord1iv: procedure(v: PGLint); stdcall;
  glTexCoord1s: procedure(s: GLshort); stdcall;
  glTexCoord1sv: procedure(v: PGLshort); stdcall;
  glTexCoord2d: procedure(s, t: GLdouble); stdcall;
  glTexCoord2dv: procedure(v: PGLdouble); stdcall;
  glTexCoord2f: procedure(s, t: GLfloat); stdcall;
  glTexCoord2fv: procedure(v: PGLfloat); stdcall;
  glTexCoord2i: procedure(s, t: GLint); stdcall;
  glTexCoord2iv: procedure(v: PGLint); stdcall;
  glTexCoord2s: procedure(s, t: GLshort); stdcall;
  glTexCoord2sv: procedure(v: PGLshort); stdcall;
  glTexCoord3d: procedure(s, t, r: GLdouble); stdcall;
  glTexCoord3dv: procedure(v: PGLdouble); stdcall;
  glTexCoord3f: procedure(s, t, r: GLfloat); stdcall;
  glTexCoord3fv: procedure(v: PGLfloat); stdcall;
  glTexCoord3i: procedure(s, t, r: GLint); stdcall;
  glTexCoord3iv: procedure(v: PGLint); stdcall;
  glTexCoord3s: procedure(s, t, r: GLshort); stdcall;
  glTexCoord3sv: procedure(v: PGLshort); stdcall;
  glTexCoord4d: procedure(s, t, r, q: GLdouble); stdcall;
  glTexCoord4dv: procedure(v: PGLdouble); stdcall;
  glTexCoord4f: procedure(s, t, r, q: GLfloat); stdcall;
  glTexCoord4fv: procedure(v: PGLfloat); stdcall;
  glTexCoord4i: procedure(s, t, r, q: GLint); stdcall;
  glTexCoord4iv: procedure(v: PGLint); stdcall;
  glTexCoord4s: procedure(s, t, r, q: GLshort); stdcall;
  glTexCoord4sv: procedure(v: PGLshort); stdcall;
  glTexCoordPointer: procedure(size: GLint; atype: GLEnum; stride: GLsizei; data: pointer); stdcall;
  glTexEnvf: procedure(target, pname: GLEnum; param: GLfloat); stdcall;
  glTexEnvfv: procedure(target, pname: GLEnum; params: PGLfloat); stdcall;
  glTexEnvi: procedure(target, pname: GLEnum; param: GLint); stdcall;
  glTexEnviv: procedure(target, pname: GLEnum; params: PGLint); stdcall;
  glTexGend: procedure(coord, pname: GLEnum; param: GLdouble); stdcall;
  glTexGendv: procedure(coord, pname: GLEnum; params: PGLdouble); stdcall;
  glTexGenf: procedure(coord, pname: GLEnum; param: GLfloat); stdcall;
  glTexGenfv: procedure(coord, pname: GLEnum; params: PGLfloat); stdcall;
  glTexGeni: procedure(coord, pname: GLEnum; param: GLint); stdcall;
  glTexGeniv: procedure(coord, pname: GLEnum; params: PGLint); stdcall;
  glTexImage1D: procedure(target: GLEnum; level, internalformat: GLint; width: GLsizei; border: GLint; format,
    atype: GLEnum; pixels: Pointer); stdcall;
  glTexImage2D: procedure(target: GLEnum; level, internalformat: GLint; width, height: GLsizei; border: GLint;
    format, atype: GLEnum; Pixels:Pointer); stdcall;
  glTexParameterf: procedure(target, pname: GLEnum; param: GLfloat); stdcall;
  glTexParameterfv: procedure(target, pname: GLEnum; params: PGLfloat); stdcall;
  glTexParameteri: procedure(target, pname: GLEnum; param: GLint); stdcall;
  glTexParameteriv: procedure(target, pname: GLEnum; params: PGLint); stdcall;
  glTexSubImage1D: procedure(target: GLEnum; level, xoffset: GLint; width: GLsizei; format, atype: GLEnum;
    pixels: Pointer); stdcall;
  glTexSubImage2D: procedure(target: GLEnum; level, xoffset, yoffset: GLint; width, height: GLsizei; format,
    atype: GLEnum; pixels: Pointer); stdcall;
  glTranslated: procedure(x, y, z: GLdouble); stdcall;
  glTranslatef: procedure(x, y, z: GLfloat); stdcall;
  glVertex2d: procedure(x, y: GLdouble); stdcall;
  glVertex2dv: procedure(v: PGLdouble); stdcall;
  glVertex2f: procedure(x, y: GLfloat); stdcall;
  glVertex2fv: procedure(v: PGLfloat); stdcall;
  glVertex2i: procedure(x, y: GLint); stdcall;
  glVertex2iv: procedure(v: PGLint); stdcall;
  glVertex2s: procedure(x, y: GLshort); stdcall;
  glVertex2sv: procedure(v: PGLshort); stdcall;
  glVertex3d: procedure(x, y, z: GLdouble); stdcall;
  glVertex3dv: procedure(v: PGLdouble); stdcall;
  glVertex3f: procedure(x, y, z: GLfloat); stdcall;
  glVertex3fv: procedure(v: PGLfloat); stdcall;
  glVertex3i: procedure(x, y, z: GLint); stdcall;
  glVertex3iv: procedure(v: PGLint); stdcall;
  glVertex3s: procedure(x, y, z: GLshort); stdcall;
  glVertex3sv: procedure(v: PGLshort); stdcall;
  glVertex4d: procedure(x, y, z, w: GLdouble); stdcall;
  glVertex4dv: procedure(v: PGLdouble); stdcall;
  glVertex4f: procedure(x, y, z, w: GLfloat); stdcall;
  glVertex4fv: procedure(v: PGLfloat); stdcall;
  glVertex4i: procedure(x, y, z, w: GLint); stdcall;
  glVertex4iv: procedure(v: PGLint); stdcall;
  glVertex4s: procedure(x, y, z, w: GLshort); stdcall;
  glVertex4sv: procedure(v: PGLshort); stdcall;
  glVertexPointer: procedure(size: GLint; atype: GLEnum; stride: GLsizei; data: pointer); stdcall;
  glViewport: procedure(x, y: GLint; width, height: GLsizei); stdcall;

{$IFDEF GL_12}
  // GL 1.2
  glDrawRangeElements: procedure (mode: GLenum; start, Aend: GLuint; count: GLsizei; Atype: GLenum; indices: Pointer); stdcall;
  glTexImage3D: procedure (target: GLenum; level: GLint; internalFormat: GLint; width: GLsizei; height: GLsizei;
    depth: GLsizei; border: GLint; format: GLenum; Atype: GLenum; pixels: Pointer); stdcall;
  glTexSubImage3D: procedure (target: GLenum; level, xoffset, yoffset, zoffset: GLint;
    width, height, depth: GLsizei; format, Atype: GLenum; pixels: Pointer); stdcall;
  glCopyTexSubImage3D: procedure (target: GLenum; level: GLint; xoffset: GLint; yoffset: GLint; zoffset: GLint;
    x: GLint; y: GLint; width: GLsizei; height: GLsizei); stdcall;
{$ENDIF}

{$IFDEF GL_13}
  // GL 1.3
  glActiveTexture: procedure (texture: GLenum); stdcall;
  glClientActiveTexture: procedure (texture: GLenum); stdcall;
  glCompressedTexImage1D: procedure (target: GLenum; level: GLint; internalformat: GLenum; width: GLsizei; border: GLint;
    imageSize: GLsizei; data: Pointer); stdcall;
  glCompressedTexImage2D: procedure (target: GLenum; level: GLint; internalformat: GLenum; width: GLsizei; height: GLsizei;
    border: GLint; imageSize: GLsizei; data: Pointer); stdcall;
  glCompressedTexImage3D: procedure (target: GLenum; level: GLint; internalformat: GLenum; width: GLsizei; height: GLsizei;
    depth: GLsizei; border: GLint; imageSize: GLsizei; data: Pointer); stdcall;
  glCompressedTexSubImage1D: procedure (target: GLenum; level: GLint; xoffset: GLint; width: GLsizei; format: GLenum;
    imageSize: GLsizei; data: Pointer); stdcall;
  glCompressedTexSubImage2D: procedure (target: GLenum; level: GLint; xoffset: GLint; yoffset: GLint; width: GLsizei;
    height: GLsizei; format: GLenum; imageSize: GLsizei; data: Pointer); stdcall;
  glCompressedTexSubImage3D: procedure (target: GLenum; level: GLint; xoffset: GLint; yoffset: GLint; zoffset: GLint;
    width: GLsizei; height: GLsizei; depth: GLsizei; format: GLenum; imageSize: GLsizei; data: Pointer); stdcall;
  glGetCompressedTexImage: procedure (target: GLenum; lod: GLint; img: Pointer); stdcall;
  glMultiTexCoord1d: procedure (target: GLenum; s: GLdouble); stdcall;
  glMultiTexCoord1dv: procedure (target: GLenum; v: PGLdouble); stdcall;
  glMultiTexCoord1f: procedure (target: GLenum; s: GLfloat); stdcall;
  glMultiTexCoord1fv: procedure (target: GLenum; v: PGLfloat); stdcall;
  glMultiTexCoord1i: procedure (target: GLenum; s: GLint); stdcall;
  glMultiTexCoord1iv: procedure (target: GLenum; v: PGLint); stdcall;
  glMultiTexCoord1s: procedure (target: GLenum; s: GLshort); stdcall;
  glMultiTexCoord1sv: procedure (target: GLenum; v: PGLshort); stdcall;
  glMultiTexCoord2d: procedure (target: GLenum; s: GLdouble; t: GLdouble); stdcall;
  glMultiTexCoord2dv: procedure (target: GLenum; v: PGLdouble); stdcall;
  glMultiTexCoord2f: procedure (target: GLenum; s: GLfloat; t: GLfloat); stdcall;
  glMultiTexCoord2fv: procedure (target: GLenum; v: PGLfloat); stdcall;
  glMultiTexCoord2i: procedure (target: GLenum; s: GLint; t: GLint); stdcall;
  glMultiTexCoord2iv: procedure (target: GLenum; v: PGLint); stdcall;
  glMultiTexCoord2s: procedure (target: GLenum; s: GLshort; t: GLshort); stdcall;
  glMultiTexCoord2sv: procedure (target: GLenum; v: PGLshort); stdcall;
  glMultiTexCoord3d: procedure (target: GLenum; s: GLdouble; t: GLdouble; r: GLdouble); stdcall;
  glMultiTexCoord3dv: procedure (target: GLenum; v: PGLdouble); stdcall;
  glMultiTexCoord3f: procedure (target: GLenum; s: GLfloat; t: GLfloat; r: GLfloat); stdcall;
  glMultiTexCoord3fv: procedure (target: GLenum; v: PGLfloat); stdcall;
  glMultiTexCoord3i: procedure (target: GLenum; s: GLint; t: GLint; r: GLint); stdcall;
  glMultiTexCoord3iv: procedure (target: GLenum; v: PGLint); stdcall;
  glMultiTexCoord3s: procedure (target: GLenum; s: GLshort; t: GLshort; r: GLshort); stdcall;
  glMultiTexCoord3sv: procedure (target: GLenum; v: PGLshort); stdcall;
  glMultiTexCoord4d: procedure (target: GLenum; s: GLdouble; t: GLdouble; r: GLdouble; q: GLdouble); stdcall;
  glMultiTexCoord4dv: procedure (target: GLenum; v: PGLdouble); stdcall;
  glMultiTexCoord4f: procedure (target: GLenum; s: GLfloat; t: GLfloat; r: GLfloat; q: GLfloat); stdcall;
  glMultiTexCoord4fv: procedure (target: GLenum; v: PGLfloat); stdcall;
  glMultiTexCoord4i: procedure (target: GLenum; s: GLint; t: GLint; r: GLint; q: GLint); stdcall;
  glMultiTexCoord4iv: procedure (target: GLenum; v: PGLint); stdcall;
  glMultiTexCoord4s: procedure (target: GLenum; s: GLshort; t: GLshort; r: GLshort; q: GLshort); stdcall;
  glMultiTexCoord4sv: procedure (target: GLenum; v: PGLshort); stdcall;
  glLoadTransposeMatrixd: procedure (m: PGLdouble); stdcall;
  glLoadTransposeMatrixf: procedure (m: PGLfloat); stdcall;
  glMultTransposeMatrixd: procedure (m: PGLdouble); stdcall;
  glMultTransposeMatrixf: procedure (m: PGLfloat); stdcall;
  glSampleCoverage: procedure (value: GLclampf; invert: GLboolean); stdcall;
{$ENDIF}

{$IFDEF GL_14}
  // GL 1.4
  glFogCoordf: procedure (coord: GLfloat); stdcall;
  glFogCoordfv: procedure (coord: PGLfloat); stdcall;
  glFogCoordd: procedure (coord: GLdouble); stdcall;
  glFogCoorddv: procedure (coord: PGLdouble); stdcall;
  glFogCoordPointer: procedure (Atype: GLenum; stride: GLsizei; Apointer: Pointer); stdcall;
  glMultiDrawArrays: procedure (mode: GLenum; first: PGLint; count: PGLsizei; primcount: GLsizei); stdcall;
  glMultiDrawElements: procedure (mode: GLenum; count: PGLsizei; Atype: GLenum; indices: Pointer; primcount: GLsizei); stdcall;
  glPointParameterf: procedure (pname: GLenum; param: GLfloat); stdcall;
  glPointParameterfv: procedure (pname: GLenum; params: PGLfloat); stdcall;
  glSecondaryColor3b: procedure (red: GLbyte; green: GLbyte; blue: GLbyte); stdcall;
  glSecondaryColor3bv: procedure (v: PGLbyte); stdcall;
  glSecondaryColor3d: procedure (red: GLdouble; green: GLdouble; blue: GLdouble); stdcall;
  glSecondaryColor3dv: procedure (v: PGLdouble); stdcall;
  glSecondaryColor3f: procedure (red: GLfloat; green: GLfloat; blue: GLfloat); stdcall;
  glSecondaryColor3fv: procedure (v: PGLfloat); stdcall;
  glSecondaryColor3i: procedure (red: GLint; green: GLint; blue: GLint); stdcall;
  glSecondaryColor3iv: procedure (v: PGLint); stdcall;
  glSecondaryColor3s: procedure (red: GLshort; green: GLshort; blue: GLshort); stdcall;
  glSecondaryColor3sv: procedure (v: PGLshort); stdcall;
  glSecondaryColor3ub: procedure (red: GLubyte; green: GLubyte; blue: GLubyte); stdcall;
  glSecondaryColor3ubv: procedure (v: PGLubyte); stdcall;
  glSecondaryColor3ui: procedure (red: GLuint; green: GLuint; blue: GLuint); stdcall;
  glSecondaryColor3uiv: procedure (v: PGLuint); stdcall;
  glSecondaryColor3us: procedure (red: GLushort; green: GLushort; blue: GLushort); stdcall;
  glSecondaryColor3usv: procedure (v: PGLushort); stdcall;
  glSecondaryColorPointer: procedure (size: GLint; Atype: GLenum; stride: GLsizei; Apointer: Pointer); stdcall;
  glBlendFuncSeparate: procedure (sfactorRGB: GLenum; dfactorRGB: GLenum; sfactorAlpha: GLenum; dfactorAlpha: GLenum); stdcall;
  glWindowPos2d: procedure (x: GLdouble; y: GLdouble); stdcall;
  glWindowPos2f: procedure (x: GLfloat; y: GLfloat); stdcall;
  glWindowPos2i: procedure (x: GLint; y: GLint); stdcall;
  glWindowPos2s: procedure (x: GLshort; y: GLshort); stdcall;
  glWindowPos2dv: procedure (p: PGLdouble); stdcall;
  glWindowPos2fv: procedure (p: PGLfloat); stdcall;
  glWindowPos2iv: procedure (p: PGLint); stdcall;
  glWindowPos2sv: procedure (p: PGLshort); stdcall;
  glWindowPos3d: procedure (x: GLdouble; y: GLdouble; z: GLdouble); stdcall;
  glWindowPos3f: procedure (x: GLfloat; y: GLfloat; z: GLfloat); stdcall;
  glWindowPos3i: procedure (x: GLint; y: GLint; z: GLint); stdcall;
  glWindowPos3s: procedure (x: GLshort; y: GLshort; z: GLshort); stdcall;
  glWindowPos3dv: procedure (p: PGLdouble); stdcall;
  glWindowPos3fv: procedure (p: PGLfloat); stdcall;
  glWindowPos3iv: procedure (p: PGLint); stdcall;
  glWindowPos3sv: procedure (p: PGLshort); stdcall;
{$ENDIF}

{$IFDEF GL_15}
  glBindBuffer: procedure(target: GLenum; buffer: GLuint); stdcall;
  glDeleteBuffers: procedure(n: GLsizei; buffer: PGLuint); stdcall;
  glGenBuffers: procedure(n: GLsizei; buffer: PGLuint); stdcall;
  glIsBuffer: function(buffer: GLuint): GLboolean; stdcall;
  glBufferData: procedure(target: GLenum; size: GLsizei; data: Pointer; usage: GLenum); stdcall;
  glBufferSubData: procedure(target: GLenum; offset: GLsizei; size: GLsizei; data: Pointer); stdcall;
  glGetBufferSubData: procedure(target: GLenum; offset: GLsizei; size: GLsizei; data: Pointer); stdcall;
  glMapBuffer: function(target: GLenum; access: GLenum): Pointer; stdcall;
  glUnmapBuffer: function(target: GLenum): GLboolean; stdcall;
  glGetBufferParameteriv: procedure(target: GLenum; pname: GLenum; params: PGLint); stdcall;
  glGetBufferPointerv: procedure(target: GLenum; pname: GLenum; param: PPointer); stdcall;

  glGenQueries:procedure(n:GLsizei;ids:PGLuint);stdcall;
  glDeleteQueries:procedure(n:GLsizei;const ids:PGLuint);stdcall;
  glIsQuery:function(id:GLuint):boolean;stdcall;
  glBeginQuery:procedure(target:GLenum;id:GLuint);stdcall;
  glEndQuery:procedure(target:GLenum);stdcall;
  glGetQueryiv:procedure(target,pname:GLenum;params:PGLint);stdcall;
  glGetQueryObjectiv:procedure(id:GLuint;pname:GLenum;params:PGLint);stdcall;
  glGetQueryObjectuiv:procedure(id:GLuint;pname:GLenum;params:PGLuint);stdcall;
{$ENDIF}

{$IFDEF GL_PROGRAM}
  // GL_ARB_vertex_program, GL_ARB_fragment_program
  glVertexAttrib1sARB: procedure(index: GLuint; x: GLshort); stdcall;
  glVertexAttrib1fARB: procedure(index: GLuint; x: GLfloat); stdcall;
  glVertexAttrib1dARB: procedure(index: GLuint; x: GLdouble); stdcall;
  glVertexAttrib2sARB: procedure(index: GLuint; x: GLshort; y: GLshort); stdcall;
  glVertexAttrib2fARB: procedure(index: GLuint; x: GLfloat; y: GLfloat); stdcall;
  glVertexAttrib2dARB: procedure(index: GLuint; x: GLdouble; y: GLdouble); stdcall;
  glVertexAttrib3sARB: procedure(index: GLuint; x: GLshort; y: GLshort; z: GLshort); stdcall;
  glVertexAttrib3fARB: procedure(index: GLuint; x: GLfloat; y: GLfloat; z: GLfloat); stdcall;
  glVertexAttrib3dARB: procedure(index: GLuint; x: GLdouble; y: GLdouble; z: GLdouble); stdcall;
  glVertexAttrib4sARB: procedure(index: GLuint; x: GLshort; y: GLshort; z: GLshort; w: GLshort); stdcall;
  glVertexAttrib4fARB: procedure(index: GLuint; x: GLfloat; y: GLfloat; z: GLfloat; w: GLfloat); stdcall;
  glVertexAttrib4dARB: procedure(index: GLuint; x: GLdouble; y: GLdouble; z: GLdouble; w: GLdouble); stdcall;
  glVertexAttrib4NubARB: procedure(index: GLuint; x: GLubyte; y: GLubyte; z: GLubyte; w: GLubyte); stdcall;
  glVertexAttrib1svARB: procedure(index: GLuint; v: PGLshort); stdcall;
  glVertexAttrib1fvARB: procedure(index: GLuint; v: PGLfloat); stdcall;
  glVertexAttrib1dvARB: procedure(index: GLuint; v: PGLdouble); stdcall;
  glVertexAttrib2svARB: procedure(index: GLuint; v: PGLshort); stdcall;
  glVertexAttrib2fvARB: procedure(index: GLuint; v: PGLfloat); stdcall;
  glVertexAttrib2dvARB: procedure(index: GLuint; v: PGLdouble); stdcall;
  glVertexAttrib3svARB: procedure(index: GLuint; v: PGLshort); stdcall;
  glVertexAttrib3fvARB: procedure(index: GLuint; v: PGLfloat); stdcall;
  glVertexAttrib3dvARB: procedure(index: GLuint; v: PGLdouble); stdcall;
  glVertexAttrib4bvARB: procedure(index: GLuint; v: PGLbyte); stdcall;
  glVertexAttrib4svARB: procedure(index: GLuint; v: PGLshort); stdcall;
  glVertexAttrib4ivARB: procedure(index: GLuint; v: PGLint); stdcall;
  glVertexAttrib4ubvARB: procedure(index: GLuint; v: PGLubyte); stdcall;
  glVertexAttrib4usvARB: procedure(index: GLuint; v: PGLushort); stdcall;
  glVertexAttrib4uivARB: procedure(index: GLuint; v: PGLuint); stdcall;
  glVertexAttrib4fvARB: procedure(index: GLuint; v: PGLfloat); stdcall;
  glVertexAttrib4dvARB: procedure(index: GLuint; v: PGLdouble); stdcall;
  glVertexAttrib4NbvARB: procedure(index: GLuint; v: PGLbyte); stdcall;
  glVertexAttrib4NsvARB: procedure(index: GLuint; v: PGLshort); stdcall;
  glVertexAttrib4NivARB: procedure(index: GLuint; v: PGLint); stdcall;
  glVertexAttrib4NubvARB: procedure(index: GLuint; v: PGLubyte); stdcall;
  glVertexAttrib4NusvARB: procedure(index: GLuint; v: PGLushort); stdcall;
  glVertexAttrib4NuivARB: procedure(index: GLuint; v: PGLuint); stdcall;
  glVertexAttribPointerARB: procedure(index: GLuint; size: GLint; atype: GLenum; normalized: GLboolean; stride: GLsizei; pointer: pointer); stdcall;
  glEnableVertexAttribArrayARB: procedure(index: GLuint); stdcall;
  glDisableVertexAttribArrayARB: procedure(index: GLuint); stdcall;

  glProgramStringARB: procedure(target: GLenum; format: GLenum; len: GLsizei; astring: PChar); stdcall;
  glBindProgramARB: procedure(target: GLenum; aprogram: GLuint); stdcall;
  glDeleteProgramsARB: procedure(n: GLsizei; programs: PGLuint); stdcall;
  glGenProgramsARB: procedure(n: GLsizei; programs: PGLuint); stdcall;

  glProgramEnvParameter4fARB: procedure(target: GLenum; index: GLuint; x, y, z, w: GLfloat); stdcall;
  glProgramEnvParameter4dARB: procedure(target: GLenum; index: GLuint; x, y, z, w: GLdouble); stdcall;
  glProgramEnvParameter4fvARB: procedure(target: GLenum; index: GLuint; params: PGLfloat); stdcall;
  glProgramEnvParameter4dvARB: procedure(target: GLenum; index: GLuint; params: PGLdouble); stdcall;
  glProgramLocalParameter4fARB: procedure(target: GLenum; index: GLuint; x, y, z, w: GLfloat); stdcall;
  glProgramLocalParameter4dARB: procedure(target: GLenum; index: GLuint; x, y, z, w: GLdouble); stdcall;
  glProgramLocalParameter4fvARB: procedure(target: GLenum; index: GLuint; params: PGLfloat); stdcall;
  glProgramLocalParameter4dvARB: procedure(target: GLenum; index: GLuint; params: PGLdouble); stdcall;

  glGetProgramEnvParameterfvARB: procedure(target: GLenum; index: GLuint; params: PGLfloat); stdcall;
  glGetProgramEnvParameterdvARB: procedure(target: GLenum; index: GLuint; params: PGLdouble); stdcall;
  glGetProgramLocalParameterfvARB: procedure(target: GLenum; index: GLuint; params: PGLfloat); stdcall;
  glGetProgramLocalParameterdvARB: procedure(target: GLenum; index: GLuint; params: PGLdouble); stdcall;
  glGetProgramivARB: procedure(target: GLenum; pname: GLenum; params: PGLint); stdcall;

  glGetProgramStringARB: procedure(target: GLenum; pname: GLenum; astring: PChar); stdcall;
  glGetVertexAttribdvARB: procedure(index: GLuint; pname: GLenum; params: PGLdouble); stdcall;
  glGetVertexAttribfvARB: procedure(index: GLuint; pname: GLenum; params: PGLfloat); stdcall;
  glGetVertexAttribivARB: procedure(index: GLuint; pname: GLenum; params: PGLint); stdcall;

  glIsProgramARB: function(aprogram: GLuint): GLboolean; stdcall;
{$ENDIF}

{$IFDEF GL_SHADER}
  // GL_ARB_shader_objects
  glDeleteObjectARB: procedure(Obj: GLHandleARB); stdcall;
  glGetHandleARB: function(pname: GlEnum):GLHandleARB; stdcall;
  glDetachObjectARB: procedure(container, attached: GLHandleARB); stdcall;
  glCreateShaderObjectARB: function(shaderType: glenum):GLHandleARB; stdcall;
  glShaderSourceARB: procedure(shaderObj :GLHandleARB; count:GLsizei; _string:PPGLCharARB; lengths:PGLint);stdcall;
  glCompileShaderARB: function(shaderObj: GLHandleARB):GLboolean; stdcall;
  glCreateProgramObjectARB: function:GLHandleARB; stdcall;
  glAttachObjectARB: procedure(programObj, shaderObj:GLhandleARB); stdcall;
  glLinkProgramARB: procedure(programObj: GLHandleARB); stdcall;
  glUseProgramObjectARB: procedure(programObj:GLHandleARB); stdcall;
  glValidateProgramARB: procedure(programObj: GLhandleARB); stdcall;
  glUniform1fARB: procedure(location:GLint;v0:GLfloat); stdcall;
  glUniform2fARB: procedure(location:GLint;v0,v1:GLfloat); stdcall;
  glUniform3fARB: procedure(location:GLint;v0,v1,v2:GLfloat); stdcall;
  glUniform4fARB: procedure(location:GLint;v0,v1,v2,v3:GLfloat); stdcall;
  glUniform1iARB: procedure(location:GLint;v0:GLint); stdcall;
  glUniform2iARB: procedure(location:GLint;v0,v1:GLint); stdcall;
  glUniform3iARB: procedure(location:GLint;v0,v1,v2:GLint); stdcall;
  glUniform4iARB: procedure(location:GLint;v0,v1,v2,v3:GLint); stdcall;
  glUniform1fvARB: procedure(location:GLint;value:PGLfloat); stdcall;
  glUniform2fvARB: procedure(location:GLint;value:PGLfloat); stdcall;
  glUniform3fvARB: procedure(location:GLint;value:PGLfloat); stdcall;
  glUniform4fvARB: procedure(location:GLint;value:PGLfloat); stdcall;
  glUniform1ivARB: procedure(location:GLint;value:PGLint); stdcall;
  glUniform2ivARB: procedure(location:GLint;value:PGLint); stdcall;
  glUniform3ivARB: procedure(location:GLint;value:PGLint); stdcall;
  glUniform4ivARB: procedure(location:GLint;value:PGLint); stdcall;
  glUniformMatrix2fvARB: procedure(location:GLint;count:GLsizei;transpose:GLboolean;value:PGLfloat); stdcall;
  glUniformMatrix3fvARB: procedure(location:GLint;count:GLsizei;transpose:GLboolean;value:PGLfloat); stdcall;
  glUniformMatrix4fvARB: procedure(location:GLint;count:GLsizei;transpose:GLboolean;value:PGLfloat); stdcall;
  glGetObjectParameterfvARB: procedure(Obj:GLHandleARB; pname:GLEnum; params:PGLFloat); stdcall;
  glGetObjectParameterivARB: procedure(Obj:GLHandleARB; pname:GLEnum; params:PGLInt); stdcall;
  glGetInfoLogARB: procedure(shaderObj:GLHandleARB; maxLength:GLsizei; var length:GLint;infoLog:PGLCharARB); stdcall;
  glGetAttachedObjectsARB: procedure(programobj:GLhandleARB; maxCount:GLsizei; var count:GLsizei;objects:PGLhandleARB); stdcall;
  glGetUniformLocationARB: function(programObj:GLhandleARB; const char:PGLCharARB):GLint; stdcall;
  glGetActiveUniformARB:procedure(programobj:GLhandleARB;index:GLuint;maxLength:GLsizei;var length:GLsizei;var size:GLint;var _type:GLenum;name:PGLcharARB); stdcall;
  glGetUniformfvARB: procedure(programObj: GLhandleARB; location:GLint; params:PGLfloat); stdcall;
  glGetUniformivARB: procedure(programObj: GLhandleARB; location:GLint; params:PGLInt); stdcall;
  glGetShaderSourceARB: procedure(shader:GLhandleARB; maxLength:GLsizei; var length:GLsizei; source:PGLcharARB); stdcall;

  // GL_ARB_vertex_shader
  glGetActiveAttribARB: procedure(programobj:GLhandleARB;index:GLuint;maxLength:GLsizei;var length:GLsizei;var size:GLint;var _type:GLenum;name:PGLcharARB);stdcall;
  glGetAttribLocationARB: function(programObj:GLhandleARB;const char:PGLCharARB):GLint;stdcall;
  glBindAttribLocationARB: procedure(programObj:GLhandleARB;index:GLuint;const name:PGLcharARB);
{$ENDIF}

  {$IFNDEF NO_GLU}
  // GL utility functions and procedures
  gluErrorString: function(errCode: GLEnum): PChar; stdcall;
  gluGetString: function(name: GLEnum): PChar; stdcall;
  gluOrtho2D: procedure(left, right, bottom, top: GLdouble); stdcall;
  gluPerspective: procedure(fovy, aspect, zNear, zFar: GLdouble); stdcall;
  gluPickMatrix: procedure(x, y, width, height: GLdouble; viewport: TVector4i); stdcall;
  gluLookAt: procedure(eyex, eyey, eyez, centerx, centery, centerz, upx, upy, upz: GLdouble); stdcall;
  gluProject: function(objx, objy, objz: GLdouble; modelMatrix, projMatrix: TMatrix4d; viewport: TVector4i;
    winx, winy, winz: PGLdouble): GLint; stdcall;
  gluUnProject: function(winx, winy, winz: GLdouble; modelMatrix, projMatrix: TMatrix4d; viewport: TVector4i;
    objx, objy, objz: PGLdouble): GLint; stdcall;
  gluScaleImage: function(format: GLEnum; widthin, heightin: GLint; typein: GLEnum; datain: Pointer; widthout,
    heightout: GLint; typeout: GLEnum; dataout: Pointer): GLint; stdcall;
  gluBuild1DMipmaps: function(target: GLEnum; components, width: GLint; format, atype: GLEnum;
    data: Pointer): GLint; stdcall;
  gluBuild2DMipmaps: function(target: GLEnum; components, width, height: GLint; format, atype: GLEnum;
    Data: Pointer): GLint; stdcall;
  gluBuild3DMipmaps: function(target: GLenum; components, width, height, depth: GLint; format, atype: GLenum;
    Data: Pointer): GLint; stdcall;
  gluNewQuadric: function: PGLUquadric; stdcall;
  gluDeleteQuadric: procedure(state: PGLUquadric); stdcall;
  gluQuadricNormals: procedure(quadObject: PGLUquadric; normals: GLEnum); stdcall;
  gluQuadricTexture: procedure(quadObject: PGLUquadric; textureCoords: GLboolean); stdcall;
  gluQuadricOrientation: procedure(quadObject: PGLUquadric; orientation: GLEnum); stdcall;
  gluQuadricDrawStyle: procedure(quadObject: PGLUquadric; drawStyle: GLEnum); stdcall;
  gluCylinder: procedure(quadObject: PGLUquadric; baseRadius, topRadius, height: GLdouble; slices,
    stacks: GLint); stdcall;
  gluDisk: procedure(quadObject: PGLUquadric; innerRadius, outerRadius: GLdouble; slices, loops: GLint); stdcall;
  gluPartialDisk: procedure(quadObject: PGLUquadric; innerRadius, outerRadius: GLdouble; slices, loops: GLint;
    startAngle, sweepAngle: GLdouble); stdcall;
  gluSphere: procedure(quadObject: PGLUquadric; radius: GLdouble; slices, stacks: GLint); stdcall;
  gluQuadricCallback: procedure(quadObject: PGLUquadric; which: GLEnum; fn: GLUQuadricErrorProc); stdcall;
  gluNewTess: function: PGLUtesselator; stdcall;
  gluDeleteTess: procedure(tess: PGLUtesselator); stdcall;
  gluTessBeginPolygon: procedure(tess: PGLUtesselator; polygon_data: Pointer); stdcall;
  gluTessBeginContour: procedure(tess: PGLUtesselator); stdcall;
  gluTessVertex: procedure(tess: PGLUtesselator; coords: TVector3d; data: Pointer); stdcall;
  gluTessEndContour: procedure(tess: PGLUtesselator); stdcall;
  gluTessEndPolygon: procedure(tess: PGLUtesselator); stdcall;
  gluTessProperty: procedure(tess: PGLUtesselator; which: GLEnum; value: GLdouble); stdcall;
  gluTessNormal: procedure(tess: PGLUtesselator; x, y, z: GLdouble); stdcall;
  gluTessCallback: procedure(tess: PGLUtesselator; which: GLEnum; fn: Pointer); stdcall;
  gluGetTessProperty: procedure(tess: PGLUtesselator; which: GLEnum; value: PGLdouble); stdcall;
  gluNewNurbsRenderer: function: PGLUnurbs; stdcall;
  gluDeleteNurbsRenderer: procedure(nobj: PGLUnurbs); stdcall;
  gluBeginSurface: procedure(nobj: PGLUnurbs); stdcall;
  gluBeginCurve: procedure(nobj: PGLUnurbs); stdcall;
  gluEndCurve: procedure(nobj: PGLUnurbs); stdcall;
  gluEndSurface: procedure(nobj: PGLUnurbs); stdcall;
  gluBeginTrim: procedure(nobj: PGLUnurbs); stdcall;
  gluEndTrim: procedure(nobj: PGLUnurbs); stdcall;
  gluPwlCurve: procedure(nobj: PGLUnurbs; count: GLint; points: PGLfloat; stride: GLint; atype: GLEnum); stdcall;
  gluNurbsCurve: procedure(nobj: PGLUnurbs; nknots: GLint; knot: PGLfloat; stride: GLint; ctlarray: PGLfloat; order: GLint; atype: GLEnum); stdcall;
  gluNurbsSurface: procedure(nobj: PGLUnurbs; sknot_count: GLint; sknot: PGLfloat; tknot_count: GLint; tknot: PGLfloat; s_stride, t_stride: GLint; ctlarray: PGLfloat; sorder, torder: GLint; atype: GLEnum); stdcall;
  gluLoadSamplingMatrices: procedure(nobj: PGLUnurbs; modelMatrix, projMatrix: TMatrix4f; viewport: TVector4i); stdcall;
  gluNurbsProperty: procedure(nobj: PGLUnurbs; aproperty: GLEnum; value: GLfloat); stdcall;
  gluGetNurbsProperty: procedure(nobj: PGLUnurbs; aproperty: GLEnum; value: PGLfloat); stdcall;
  gluNurbsCallback: procedure(nobj: PGLUnurbs; which: GLEnum; fn: GLUNurbsErrorProc); stdcall;
  gluBeginPolygon: procedure(tess: PGLUtesselator); stdcall;
  gluNextContour: procedure(tess: PGLUtesselator; atype: GLEnum); stdcall;
  gluEndPolygon: procedure(tess: PGLUtesselator); stdcall;
{$ENDIF}

  // window support functions
  wglGetProcAddress: function(ProcName: PChar): Pointer; stdcall;
  wglCopyContext: function(p1: HGLRC; p2: HGLRC; p3: Cardinal): BOOL; stdcall;
  wglCreateContext: function(DC: HDC): HGLRC; stdcall;
  wglCreateLayerContext: function(p1: HDC; p2: Integer): HGLRC; stdcall;
  wglDeleteContext: function(p1: HGLRC): BOOL; stdcall;
  wglDescribeLayerPlane:function(p1: HDC; p2, p3: Integer; p4: Cardinal; p5: Pointer): BOOL; stdcall;
  wglGetCurrentContext: function: HGLRC; stdcall;
  wglGetCurrentDC: function: HDC; stdcall;
  wglGetLayerPaletteEntries: function(p1: HDC; p2, p3, p4: Integer; var pcr): Integer; stdcall;
  wglMakeCurrent: function(DC: HDC; p2: HGLRC): BOOL; stdcall;
  wglRealizeLayerPalette: function(p1: HDC; p2: Integer; p3: BOOL): BOOL; stdcall;
  wglSetLayerPaletteEntries: function(p1: HDC; p2, p3, p4: Integer; var pcr): Integer; stdcall;
  wglShareLists: function(p1, p2: HGLRC): BOOL; stdcall;
  wglSwapLayerBuffers: function(p1: HDC; p2: Cardinal): BOOL; stdcall;
  wglSwapMultipleBuffers: function(p1: UINT; const p2: PWGLSwap): DWORD; stdcall;
  wglUseFontBitmapsA: function(DC: HDC; p2, p3, p4: DWORD): BOOL; stdcall;
  wglUseFontOutlinesA: function (p1: HDC; p2, p3, p4: DWORD; p5, p6: Single; p7: Integer; p8: PGlyphMetricsFloat): BOOL; stdcall;
  wglUseFontBitmapsW: function(DC: HDC; p2, p3, p4: DWORD): BOOL; stdcall;
  wglUseFontOutlinesW: function (p1: HDC; p2, p3, p4: DWORD; p5, p6: Single; p7: Integer; p8: PGlyphMetricsFloat): BOOL; stdcall;
  wglUseFontBitmaps: function(DC: HDC; p2, p3, p4: DWORD): BOOL; stdcall;
  wglUseFontOutlines: function(p1: HDC; p2, p3, p4: DWORD; p5, p6: Single; p7: Integer; p8: PGlyphMetricsFloat): BOOL; stdcall;

  wglChoosePixelFormat: function(DC: HDC; p2: pPixelFormatDescriptor) : integer; stdcall;
  wglGetPixelFormat: function(DC: HDC) : integer; stdcall;
  wglSetPixelFormat: function(DC: HDC; PixelFormat: Integer; FormatDef: PPixelFormatDescriptor) : LongBool; stdcall;
  wglDescribePixelFormat: function(DC: HDC; PixelFormat: Integer; nBytes: Integer; var p4: TPixelFormatDescriptor) : LongBool; stdcall;
  wglSwapBuffers: function(DC: HDC): LongBool; stdcall;

  // maybe supported by OpenGL ICD, else reset to windows standard
  ChoosePixelFormat: function(DC: HDC; p2: pPixelFormatDescriptor) : integer; stdcall;
  GetPixelFormat: function(DC: HDC) : integer; stdcall;
  SetPixelFormat: function(DC: HDC; PixelFormat: Integer; FormatDef: PPixelFormatDescriptor) : LongBool; stdcall;
  DescribePixelFormat: function(DC: HDC; PixelFormat: Integer; nBytes: Integer; var p4: TPixelFormatDescriptor) : LongBool; stdcall;
  SwapBuffers: function(DC: HDC): LongBool; stdcall;

//----------------------------------------------------------------------------------------------------------------------

procedure CloseOpenGL;
function InitOpenGL: Boolean;
{$IFDEF NO_GLU}
function InitOpenGLFromLibrary(GLName: AnsiString): Boolean;
procedure gluLookAt(eyex, eyey, eyez,  centerx, centery, centerz,  upx, upy, upz: GLDouble);
procedure gluPerspective(fovy, aspect, zNear, zFar: GLdouble);
procedure gluPickMatrix(x, y, width, height: GLdouble; viewport: TVector4i);
function gluProject(objx, objy, objz: GLdouble; modelMatrix, projMatrix: TMatrix4d; viewport: TVector4i;
                    winx, winy, winz: PGLdouble): GLint;
function gluUnProject(winx, winy, winz: GLdouble; modelMatrix, projMatrix: TMatrix4d; viewport: TVector4i;
                      objx, objy, objz: PGLdouble): GLint;
{$ELSE}
function InitOpenGLFromLibrary(GLName, GLUName: AnsiString): Boolean;
{$ENDIF}
function IsOpenGLInitialized: Boolean;

procedure ActivateRenderingContext(DC: HDC; RC: HGLRC);
function CreateRenderingContext(DC: HDC; PixelFormat: Integer; var PFDescriptor: TPixelFormatDescriptor): HGLRC;
procedure ReadImplementationProperties;
procedure ShowError(MsgString: string);

//----------------------------------------------------------------------------------------------------------------------

var
  GLHandle: HINST;
  SGL  : String[127] = 'OpenGL32.dll';
{$IFNDEF NO_GLU}
  GLUHandle: HINST;
  SGLU : String[127] = 'GLU32.dll';
{$ENDIF}
  SWGL : Boolean = FALSE;

implementation

uses
  SysUtils, Classes;


const
  INVALID_MODULEHANDLE = 0;

//----------------------------------------------------------------------------------------------------------------------

procedure ShowError(MsgString: string);

begin
  MessageBox(0, PAnsiChar(MsgString), 'Error', MB_OK);
  Halt(1);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure ClearProcAddresses;

begin
  glAccum := nil;
  glAlphaFunc := nil;
  glAreTexturesResident := nil;
  glArrayElement := nil;
  glBegin := nil;
  glBindTexture := nil;
  glBitmap := nil; 
  glBlendFunc := nil;
  glCallList := nil;
  glCallLists := nil;
  glClear := nil;
  glClearAccum := nil;
  glClearColor := nil;
  glClearDepth := nil;
  glClearIndex := nil;
  glClearStencil := nil;
  glClipPlane := nil;
  glColor3b := nil;
  glColor3bv := nil;
  glColor3d := nil;
  glColor3dv := nil;
  glColor3f := nil;
  glColor3fv := nil;
  glColor3i := nil;
  glColor3iv := nil;
  glColor3s := nil;
  glColor3sv := nil;
  glColor3ub := nil;
  glColor3ubv := nil;
  glColor3ui := nil;
  glColor3uiv := nil;
  glColor3us := nil; 
  glColor3usv := nil;
  glColor4b := nil;
  glColor4bv := nil;
  glColor4d := nil; 
  glColor4dv := nil;
  glColor4f := nil;
  glColor4fv := nil;
  glColor4i := nil;
  glColor4iv := nil;
  glColor4s := nil;
  glColor4sv := nil;
  glColor4ub := nil;
  glColor4ubv := nil;
  glColor4ui := nil;
  glColor4uiv := nil;
  glColor4us := nil;
  glColor4usv := nil;
  glColorMask := nil;
  glColorMaterial := nil;
  glColorPointer := nil;
  glCopyPixels := nil;
  glCopyTexImage1D := nil;
  glCopyTexImage2D := nil; 
  glCopyTexSubImage1D := nil;
  glCopyTexSubImage2D := nil;
  glCullFace := nil;
  glDeleteLists := nil;
  glDeleteTextures := nil;
  glDepthFunc := nil;
  glDepthMask := nil;
  glDepthRange := nil;
  glDisable := nil;
  glDisableClientState := nil;
  glDrawArrays := nil;
  glDrawBuffer := nil;
  glDrawElements := nil;
  glDrawPixels := nil;
  glEdgeFlag := nil;
  glEdgeFlagPointer := nil;
  glEdgeFlagv := nil;
  glEnable := nil;
  glEnableClientState := nil;
  glEnd := nil;
  glEndList := nil;
  glEvalCoord1d := nil;
  glEvalCoord1dv := nil;
  glEvalCoord1f := nil;
  glEvalCoord1fv := nil;
  glEvalCoord2d := nil;
  glEvalCoord2dv := nil;
  glEvalCoord2f := nil;
  glEvalCoord2fv := nil;
  glEvalMesh1 := nil;
  glEvalMesh2 := nil;
  glEvalPoint1 := nil;
  glEvalPoint2 := nil;
  glFeedbackBuffer := nil;
  glFinish := nil;
  glFlush := nil;
  glFogf := nil;
  glFogfv := nil;
  glFogi := nil;
  glFogiv := nil;
  glFrontFace := nil;
  glFrustum := nil;
  glGenLists := nil;
  glGenTextures := nil;
  glGetBooleanv := nil;
  glGetClipPlane := nil;
  glGetDoublev := nil;
  glGetError := nil;
  glGetFloatv := nil;
  glGetIntegerv := nil;
  glGetLightfv := nil;
  glGetLightiv := nil;
  glGetMapdv := nil;
  glGetMapfv := nil;
  glGetMapiv := nil;
  glGetMaterialfv := nil;
  glGetMaterialiv := nil;
  glGetPixelMapfv := nil;
  glGetPixelMapuiv := nil;
  glGetPixelMapusv := nil;
  glGetPointerv := nil;
  glGetPolygonStipple := nil;
  glGetString := nil;
  glGetTexEnvfv := nil;
  glGetTexEnviv := nil; 
  glGetTexGendv := nil;
  glGetTexGenfv := nil;
  glGetTexGeniv := nil;
  glGetTexImage := nil;
  glGetTexLevelParameterfv := nil; 
  glGetTexLevelParameteriv := nil;
  glGetTexParameterfv := nil;
  glGetTexParameteriv := nil;
  glHint := nil;
  glIndexMask := nil;
  glIndexPointer := nil;
  glIndexd := nil;
  glIndexdv := nil;
  glIndexf := nil;
  glIndexfv := nil;
  glIndexi := nil;
  glIndexiv := nil;
  glIndexs := nil;
  glIndexsv := nil;
  glIndexub := nil;
  glIndexubv := nil;
  glInitNames := nil;
  glInterleavedArrays := nil;
  glIsEnabled := nil;
  glIsList := nil;
  glIsTexture := nil;
  glLightModelf := nil;
  glLightModelfv := nil;
  glLightModeli := nil;
  glLightModeliv := nil;
  glLightf := nil;
  glLightfv := nil;
  glLighti := nil;
  glLightiv := nil;
  glLineStipple := nil; 
  glLineWidth := nil;
  glListBase := nil;
  glLoadIdentity := nil;
  glLoadMatrixd := nil;
  glLoadMatrixf := nil; 
  glLoadName := nil;
  glLogicOp := nil;
  glMap1d := nil;
  glMap1f := nil;
  glMap2d := nil;
  glMap2f := nil;
  glMapGrid1d := nil;
  glMapGrid1f := nil;
  glMapGrid2d := nil;
  glMapGrid2f := nil;
  glMaterialf := nil;
  glMaterialfv := nil;
  glMateriali := nil;
  glMaterialiv := nil;
  glMatrixMode := nil;
  glMultMatrixd := nil;
  glMultMatrixf := nil;
  glNewList := nil;
  glNormal3b := nil;
  glNormal3bv := nil;
  glNormal3d := nil;
  glNormal3dv := nil;
  glNormal3f := nil;
  glNormal3fv := nil;
  glNormal3i := nil;
  glNormal3iv := nil;
  glNormal3s := nil;
  glNormal3sv := nil;
  glNormalPointer := nil;
  glOrtho := nil;
  glPassThrough := nil;
  glPixelMapfv := nil;
  glPixelMapuiv := nil;
  glPixelMapusv := nil;
  glPixelStoref := nil;
  glPixelStorei := nil;
  glPixelTransferf := nil;
  glPixelTransferi := nil;
  glPixelZoom := nil;
  glPointSize := nil;
  glPolygonMode := nil;
  glPolygonOffset := nil;
  glPolygonStipple := nil;
  glPopAttrib := nil;
  glPopClientAttrib := nil;
  glPopMatrix := nil;
  glPopName := nil;
  glPrioritizeTextures := nil;
  glPushAttrib := nil;
  glPushClientAttrib := nil;
  glPushMatrix := nil;
  glPushName := nil;
  glRasterPos2d := nil;
  glRasterPos2dv := nil;
  glRasterPos2f := nil;
  glRasterPos2fv := nil;
  glRasterPos2i := nil;
  glRasterPos2iv := nil;
  glRasterPos2s := nil;
  glRasterPos2sv := nil;
  glRasterPos3d := nil;
  glRasterPos3dv := nil;
  glRasterPos3f := nil;
  glRasterPos3fv := nil;
  glRasterPos3i := nil;
  glRasterPos3iv := nil;
  glRasterPos3s := nil;
  glRasterPos3sv := nil;
  glRasterPos4d := nil;
  glRasterPos4dv := nil;
  glRasterPos4f := nil;
  glRasterPos4fv := nil;
  glRasterPos4i := nil;
  glRasterPos4iv := nil;
  glRasterPos4s := nil;
  glRasterPos4sv := nil;
  glReadBuffer := nil;
  glReadPixels := nil;
  glRectd := nil;
  glRectdv := nil;
  glRectf := nil;
  glRectfv := nil;
  glRecti := nil;
  glRectiv := nil;
  glRects := nil;
  glRectsv := nil;
  glRenderMode := nil;
  glRotated := nil;
  glRotatef := nil; 
  glScaled := nil;
  glScalef := nil;
  glScissor := nil;
  glSelectBuffer := nil;
  glShadeModel := nil;
  glStencilFunc := nil;
  glStencilMask := nil;
  glStencilOp := nil;
  glTexCoord1d := nil;
  glTexCoord1dv := nil;
  glTexCoord1f := nil;
  glTexCoord1fv := nil;
  glTexCoord1i := nil;
  glTexCoord1iv := nil;
  glTexCoord1s := nil;
  glTexCoord1sv := nil;
  glTexCoord2d := nil;
  glTexCoord2dv := nil;
  glTexCoord2f := nil;
  glTexCoord2fv := nil;
  glTexCoord2i := nil;
  glTexCoord2iv := nil;
  glTexCoord2s := nil;
  glTexCoord2sv := nil;
  glTexCoord3d := nil;
  glTexCoord3dv := nil;
  glTexCoord3f := nil;
  glTexCoord3fv := nil;
  glTexCoord3i := nil; 
  glTexCoord3iv := nil;
  glTexCoord3s := nil; 
  glTexCoord3sv := nil;
  glTexCoord4d := nil;
  glTexCoord4dv := nil;
  glTexCoord4f := nil; 
  glTexCoord4fv := nil;
  glTexCoord4i := nil;
  glTexCoord4iv := nil;
  glTexCoord4s := nil;
  glTexCoord4sv := nil;
  glTexCoordPointer := nil;
  glTexEnvf := nil;
  glTexEnvfv := nil;
  glTexEnvi := nil;
  glTexEnviv := nil;
  glTexGend := nil;
  glTexGendv := nil;
  glTexGenf := nil;
  glTexGenfv := nil;
  glTexGeni := nil;
  glTexGeniv := nil;
  glTexImage1D := nil;
  glTexImage2D := nil;
  glTexParameterf := nil;
  glTexParameterfv := nil;
  glTexParameteri := nil;
  glTexParameteriv := nil;
  glTexSubImage1D := nil;
  glTexSubImage2D := nil;
  glTranslated := nil;
  glTranslatef := nil;
  glVertex2d := nil;
  glVertex2dv := nil;
  glVertex2f := nil;
  glVertex2fv := nil;
  glVertex2i := nil;
  glVertex2iv := nil;
  glVertex2s := nil;
  glVertex2sv := nil;
  glVertex3d := nil;
  glVertex3dv := nil;
  glVertex3f := nil;
  glVertex3fv := nil;
  glVertex3i := nil;
  glVertex3iv := nil;
  glVertex3s := nil;
  glVertex3sv := nil;
  glVertex4d := nil;
  glVertex4dv := nil;
  glVertex4f := nil;
  glVertex4fv := nil;
  glVertex4i := nil;
  glVertex4iv := nil;
  glVertex4s := nil;
  glVertex4sv := nil;
  glVertexPointer := nil;
  glViewport := nil;

{$IFDEF GL_12}
  // GL 1.2
  glDrawRangeElements := nil;
  glTexImage3D := nil;
  glTexSubImage3D := nil;
  glCopyTexSubImage3D := nil;
{$ENDIF}

{$IFDEF GL_13}
  // GL 1.3
  glActiveTexture := nil;
  glClientActiveTexture := nil;
  glCompressedTexImage1D := nil;
  glCompressedTexImage2D := nil;
  glCompressedTexImage3D := nil;
  glCompressedTexSubImage1D := nil;
  glCompressedTexSubImage2D := nil;
  glCompressedTexSubImage3D := nil;
  glGetCompressedTexImage := nil;
  glMultiTexCoord1d := nil;
  glMultiTexCoord1dv := nil;
  glMultiTexCoord1f := nil;
  glMultiTexCoord1fv := nil;
  glMultiTexCoord1i := nil;
  glMultiTexCoord1iv := nil;
  glMultiTexCoord1s := nil;
  glMultiTexCoord1sv := nil;
  glMultiTexCoord2d := nil;
  glMultiTexCoord2dv := nil;
  glMultiTexCoord2f := nil;
  glMultiTexCoord2fv := nil;
  glMultiTexCoord2i := nil;
  glMultiTexCoord2iv := nil;
  glMultiTexCoord2s := nil;
  glMultiTexCoord2sv := nil;
  glMultiTexCoord3d := nil;
  glMultiTexCoord3dv := nil;
  glMultiTexCoord3f := nil;
  glMultiTexCoord3fv := nil;
  glMultiTexCoord3i := nil;
  glMultiTexCoord3iv := nil;
  glMultiTexCoord3s := nil;
  glMultiTexCoord3sv := nil;
  glMultiTexCoord4d := nil;
  glMultiTexCoord4dv := nil;
  glMultiTexCoord4f := nil;
  glMultiTexCoord4fv := nil;
  glMultiTexCoord4i := nil;
  glMultiTexCoord4iv := nil;
  glMultiTexCoord4s := nil;
  glMultiTexCoord4sv := nil;
  glLoadTransposeMatrixd := nil;
  glLoadTransposeMatrixf := nil;
  glMultTransposeMatrixd := nil;
  glMultTransposeMatrixf := nil;
  glSampleCoverage := nil;
{$ENDIF}

{$IFDEF GL_14}
  // GL 1.4
  glFogCoordf := nil;
  glFogCoordfv := nil;
  glFogCoordd := nil;
  glFogCoorddv := nil;
  glFogCoordPointer := nil;
  glMultiDrawArrays := nil;
  glMultiDrawElements := nil;
  glPointParameterf := nil;
  glPointParameterfv := nil;
  glSecondaryColor3b := nil;
  glSecondaryColor3bv := nil;
  glSecondaryColor3d := nil;
  glSecondaryColor3dv := nil;
  glSecondaryColor3f := nil;
  glSecondaryColor3fv := nil;
  glSecondaryColor3i := nil;
  glSecondaryColor3iv := nil;
  glSecondaryColor3s := nil;
  glSecondaryColor3sv := nil;
  glSecondaryColor3ub := nil;
  glSecondaryColor3ubv := nil;
  glSecondaryColor3ui := nil;
  glSecondaryColor3uiv := nil;
  glSecondaryColor3us := nil;
  glSecondaryColor3usv := nil;
  glSecondaryColorPointer := nil;
  glBlendFuncSeparate := nil;
  glWindowPos2d := nil;
  glWindowPos2f := nil;
  glWindowPos2i := nil;
  glWindowPos2s := nil;
  glWindowPos2dv := nil;
  glWindowPos2fv := nil;
  glWindowPos2iv := nil;
  glWindowPos2sv := nil;
  glWindowPos3d := nil;
  glWindowPos3f := nil;
  glWindowPos3i := nil;
  glWindowPos3s := nil;
  glWindowPos3dv := nil;
  glWindowPos3fv := nil;
  glWindowPos3iv := nil;
  glWindowPos3sv := nil;
{$ENDIF}

{$IFDEF GL_15}
  // GL 1.5
  glBindBuffer := nil;
  glDeleteBuffers := nil;
  glGenBuffers := nil;
  glIsBuffer := nil;
  glBufferData := nil;
  glBufferSubData := nil;
  glGetBufferSubData := nil;
  glMapBuffer := nil;
  glUnmapBuffer := nil;
  glGetBufferParameteriv := nil;
  glGetBufferPointerv := nil;

  glGenQueries := nil;
  glDeleteQueries := nil;
  glIsQuery := nil;
  glBeginQuery := nil;
  glEndQuery := nil;
  glGetQueryiv := nil;
  glGetQueryObjectiv := nil;
  glGetQueryObjectuiv := nil;
{$ENDIF}

{$IFDEF GL_PROGRAM}
  // GL_ARB_vertex_program, GL_ARB_fragment_program
  glVertexAttrib1sARB := nil;
  glVertexAttrib1fARB := nil;
  glVertexAttrib1dARB := nil;
  glVertexAttrib2sARB := nil;
  glVertexAttrib2fARB := nil;
  glVertexAttrib2dARB := nil;
  glVertexAttrib3sARB := nil;
  glVertexAttrib3fARB := nil;
  glVertexAttrib3dARB := nil;
  glVertexAttrib4sARB := nil;
  glVertexAttrib4fARB := nil;
  glVertexAttrib4dARB := nil;
  glVertexAttrib4NubARB := nil;
  glVertexAttrib1svARB := nil;
  glVertexAttrib1fvARB := nil;
  glVertexAttrib1dvARB := nil;
  glVertexAttrib2svARB := nil;
  glVertexAttrib2fvARB := nil;
  glVertexAttrib2dvARB := nil;
  glVertexAttrib3svARB := nil;
  glVertexAttrib3fvARB := nil;
  glVertexAttrib3dvARB := nil;
  glVertexAttrib4bvARB := nil;
  glVertexAttrib4svARB := nil;
  glVertexAttrib4ivARB := nil;
  glVertexAttrib4ubvARB := nil;
  glVertexAttrib4usvARB := nil;
  glVertexAttrib4uivARB := nil;
  glVertexAttrib4fvARB := nil;
  glVertexAttrib4dvARB := nil;
  glVertexAttrib4NbvARB := nil;
  glVertexAttrib4NsvARB := nil;
  glVertexAttrib4NivARB := nil;
  glVertexAttrib4NubvARB := nil;
  glVertexAttrib4NusvARB := nil;
  glVertexAttrib4NuivARB := nil;
  glVertexAttribPointerARB := nil;
  glEnableVertexAttribArrayARB := nil;
  glDisableVertexAttribArrayARB := nil;
  glProgramStringARB := nil;
  glBindProgramARB := nil;
  glDeleteProgramsARB := nil;
  glGenProgramsARB := nil;
  glProgramEnvParameter4fARB := nil;
  glProgramEnvParameter4dARB := nil;
  glProgramEnvParameter4fvARB := nil;
  glProgramEnvParameter4dvARB := nil;
  glProgramLocalParameter4fARB := nil;
  glProgramLocalParameter4dARB := nil;
  glProgramLocalParameter4fvARB := nil;
  glProgramLocalParameter4dvARB := nil;
  glGetProgramEnvParameterfvARB := nil;
  glGetProgramEnvParameterdvARB := nil;
  glGetProgramLocalParameterfvARB := nil;
  glGetProgramLocalParameterdvARB := nil;
  glGetProgramivARB := nil;
  glGetProgramStringARB := nil;
  glGetVertexAttribdvARB := nil;
  glGetVertexAttribfvARB := nil;
  glGetVertexAttribivARB := nil;
  glIsProgramARB := nil;
{$ENDIF}

{$IFDEF GL_SHADER}
  // GL_ARB_shader_objects
  glDeleteObjectARB := nil;
  glGetHandleARB := nil;
  glDetachObjectARB := nil;
  glCreateShaderObjectARB := nil;
  glShaderSourceARB := nil;
  glCompileShaderARB := nil;
  glCreateProgramObjectARB := nil;
  glAttachObjectARB := nil;
  glLinkProgramARB := nil;
  glUseProgramObjectARB := nil;
  glValidateProgramARB := nil;
  glUniform1fARB := nil;
  glUniform2fARB := nil;
  glUniform3fARB := nil;
  glUniform4fARB := nil;
  glUniform1iARB := nil;
  glUniform2iARB := nil;
  glUniform3iARB := nil;
  glUniform4iARB := nil;
  glUniform1fvARB := nil;
  glUniform2fvARB := nil;
  glUniform3fvARB := nil;
  glUniform4fvARB := nil;
  glUniform1ivARB := nil;
  glUniform2ivARB := nil;
  glUniform3ivARB := nil;
  glUniform4ivARB := nil;
  glUniformMatrix2fvARB := nil;
  glUniformMatrix3fvARB := nil;
  glUniformMatrix4fvARB := nil;
  glGetObjectParameterfvARB := nil;
  glGetObjectParameterivARB := nil;
  glGetInfoLogARB := nil;
  glGetAttachedObjectsARB := nil;
  glGetUniformLocationARB := nil;
  glGetActiveUniformARB := nil;
  glGetUniformfvARB := nil;
  glGetUniformivARB := nil;
  glGetShaderSourceARB := nil;

  // GL_ARB_vertex_shader
  glGetActiveAttribARB := nil;
  glGetAttribLocationARB := nil;
  glBindAttribLocationARB := nil;
{$ENDIF}

  wglGetProcAddress := nil;
  wglCopyContext := nil;
  wglCreateContext := nil;
  wglCreateLayerContext := nil;
  wglDeleteContext := nil;
  wglDescribeLayerPlane := nil;
  wglGetCurrentContext := nil;
  wglGetCurrentDC := nil;
  wglGetLayerPaletteEntries := nil;
  wglMakeCurrent := nil;
  wglRealizeLayerPalette := nil;
  wglSetLayerPaletteEntries := nil;
  wglShareLists := nil;
  wglSwapLayerBuffers := nil;
  wglSwapMultipleBuffers := nil;
  wglUseFontBitmapsA := nil;
  wglUseFontOutlinesA := nil;
  wglUseFontBitmapsW := nil;
  wglUseFontOutlinesW := nil;
  wglUseFontBitmaps := nil;
  wglUseFontOutlines := nil;

  wglChoosePixelFormat := nil;
  wglGetPixelFormat := nil;
  wglSetPixelFormat := nil;
  wglDescribePixelFormat := nil;
  wglSwapBuffers := nil;

  ChoosePixelFormat := nil;
  GetPixelFormat := nil;
  SetPixelFormat := nil;
  DescribePixelFormat := nil;
  SwapBuffers := nil;

// GLU routines

{$IFNDEF NO_GLU}
  gluBeginCurve := nil;
  gluBeginPolygon := nil;
  gluBeginSurface := nil;
  gluBeginTrim := nil;
  gluBuild1DMipmaps := nil;
  gluBuild2DMipmaps := nil;
  gluBuild3DMipmaps := nil;
  gluCylinder := nil;
  gluDeleteNurbsRenderer := nil;
  gluDeleteQuadric := nil;
  gluDeleteTess := nil;
  gluDisk := nil;
  gluEndCurve := nil;
  gluEndPolygon := nil;
  gluEndSurface := nil;
  gluEndTrim := nil;
  gluErrorString := nil;
  gluGetNurbsProperty := nil;
  gluGetString := nil;
  gluGetTessProperty := nil;
  gluLoadSamplingMatrices := nil;
  gluLookAt := nil;
  gluNewNurbsRenderer := nil;
  gluNewQuadric := nil;
  gluNewTess := nil;
  gluNextContour := nil;
  gluNurbsCallback := nil;
  gluNurbsCurve := nil;
  gluNurbsProperty := nil;
  gluNurbsSurface := nil;
  gluOrtho2D := nil;
  gluPartialDisk := nil;
  gluPerspective := nil;
  gluPickMatrix := nil;
  gluProject := nil;
  gluPwlCurve := nil;
  gluQuadricCallback := nil;
  gluQuadricDrawStyle := nil;
  gluQuadricNormals := nil;
  gluQuadricOrientation := nil;
  gluQuadricTexture := nil;
  gluScaleImage := nil;
  gluSphere := nil;
  gluTessBeginContour := nil;
  gluTessBeginPolygon := nil;
  gluTessCallback := nil;
  gluTessEndContour := nil;
  gluTessEndPolygon := nil;
  gluTessNormal := nil;
  gluTessProperty := nil;
  gluTessVertex := nil;
  gluUnProject := nil;
{$ENDIF}

end;

//----------------------------------------------------------------------------------------------------------------------

function glProcedure(ProcName : PChar) : Pointer;
begin
  result := NIL;
  if Addr(wglGetProcAddress)<>NIL then
    result := wglGetProcAddress(ProcName);
  if result<>NIL then Exit;
  result := GetProcAddress(GLHandle, ProcName);
end;

procedure LoadProcAddresses;
begin
  if GLHandle <> INVALID_MODULEHANDLE then
  begin
    // window support routines
    wglGetProcAddress := GetProcAddress(GLHandle, 'wglGetProcAddress');
    wglCopyContext := glProcedure('wglCopyContext');
    wglCreateContext := glProcedure('wglCreateContext');
    wglCreateLayerContext := glProcedure('wglCreateLayerContext');
    wglDeleteContext := glProcedure('wglDeleteContext');
    wglDescribeLayerPlane := glProcedure('wglDescribeLayerPlane');
    wglGetCurrentContext := glProcedure('wglGetCurrentContext');
    wglGetCurrentDC := glProcedure('wglGetCurrentDC');
    wglGetLayerPaletteEntries := glProcedure('wglGetLayerPaletteEntries');
    wglMakeCurrent := glProcedure('wglMakeCurrent');
    wglRealizeLayerPalette := glProcedure('wglRealizeLayerPalette');
    wglSetLayerPaletteEntries := glProcedure('wglSetLayerPaletteEntries');
    wglShareLists := glProcedure('wglShareLists');
    wglSwapLayerBuffers := glProcedure('wglSwapLayerBuffers');
    wglSwapMultipleBuffers := glProcedure('wglSwapMultipleBuffers');
    wglUseFontBitmapsA := glProcedure('wglUseFontBitmapsA');
    wglUseFontOutlinesA := glProcedure('wglUseFontOutlinesA');
    wglUseFontBitmapsW := glProcedure('wglUseFontBitmapsW');
    wglUseFontOutlinesW := glProcedure('wglUseFontOutlinesW');
    wglUseFontBitmaps := glProcedure('wglUseFontBitmapsA');
    wglUseFontOutlines := glProcedure('wglUseFontOutlinesA');

    wglChoosePixelFormat := glProcedure('wglChoosePixelFormat');
    wglGetPixelFormat := glProcedure('wglGetPixelFormat');
    wglSetPixelFormat := glProcedure('wglSetPixelFormat');
    wglDescribePixelFormat := glProcedure('wglDescribePixelFormat');
    wglSwapBuffers := glProcedure('wglSwapBuffers');

    // maybe changed by OpenGL ICD, else set to windows standard functions
    ChoosePixelFormat := glProcedure('ChoosePixelFormat');
    if not assigned(ChoosePixelFormat) then ChoosePixelFormat := Windows.ChoosePixelFormat;
    GetPixelFormat := glProcedure('GetPixelFormat');
    if not assigned(GetPixelFormat) then GetPixelFormat := Windows.GetPixelFormat;
    SetPixelFormat := glProcedure('SetPixelFormat');
    if not assigned(SetPixelFormat) then SetPixelFormat := Windows.SetPixelFormat;
    DescribePixelFormat := glProcedure('DescribePixelFormat');
    if not assigned(DescribePixelFormat) then DescribePixelFormat := Addr(Windows.DescribePixelFormat);
    SwapBuffers := glProcedure('SwapBuffers');
    if not assigned(SwapBuffers) then SwapBuffers := Windows.SwapBuffers;

    // OpenGL support routines
    glAccum := glProcedure('glAccum');
    glAlphaFunc := glProcedure('glAlphaFunc');
    glAreTexturesResident := glProcedure('glAreTexturesResident');
    glArrayElement := glProcedure('glArrayElement');
    glBegin := glProcedure('glBegin');
    glBindTexture := glProcedure('glBindTexture');
    glBitmap := glProcedure('glBitmap');
    glBlendFunc := glProcedure('glBlendFunc');
    glCallList := glProcedure('glCallList');
    glCallLists := glProcedure('glCallLists');
    glClear := glProcedure('glClear');
    glClearAccum := glProcedure('glClearAccum');
    glClearColor := glProcedure('glClearColor');
    glClearDepth := glProcedure('glClearDepth');
    glClearIndex := glProcedure('glClearIndex');
    glClearStencil := glProcedure('glClearStencil');
    glClipPlane := glProcedure('glClipPlane');
    glColor3b := glProcedure('glColor3b');
    glColor3bv := glProcedure('glColor3bv');
    glColor3d := glProcedure('glColor3d');
    glColor3dv := glProcedure('glColor3dv');
    glColor3f := glProcedure('glColor3f');
    glColor3fv := glProcedure('glColor3fv');
    glColor3i := glProcedure('glColor3i');
    glColor3iv := glProcedure('glColor3iv');
    glColor3s := glProcedure('glColor3s');
    glColor3sv := glProcedure('glColor3sv');
    glColor3ub := glProcedure('glColor3ub');
    glColor3ubv := glProcedure('glColor3ubv');
    glColor3ui := glProcedure('glColor3ui');
    glColor3uiv := glProcedure('glColor3uiv');
    glColor3us := glProcedure('glColor3us');
    glColor3usv := glProcedure('glColor3usv');
    glColor4b := glProcedure('glColor4b');
    glColor4bv := glProcedure('glColor4bv');
    glColor4d := glProcedure('glColor4d');
    glColor4dv := glProcedure('glColor4dv');
    glColor4f := glProcedure('glColor4f');
    glColor4fv := glProcedure('glColor4fv');
    glColor4i := glProcedure('glColor4i');
    glColor4iv := glProcedure('glColor4iv');
    glColor4s := glProcedure('glColor4s');
    glColor4sv := glProcedure('glColor4sv');
    glColor4ub := glProcedure('glColor4ub');
    glColor4ubv := glProcedure('glColor4ubv');
    glColor4ui := glProcedure('glColor4ui');
    glColor4uiv := glProcedure('glColor4uiv');
    glColor4us := glProcedure('glColor4us');
    glColor4usv := glProcedure('glColor4usv');
    glColorMask := glProcedure('glColorMask');
    glColorMaterial := glProcedure('glColorMaterial');
    glColorPointer := glProcedure('glColorPointer');
    glCopyPixels := glProcedure('glCopyPixels');
    glCopyTexImage1D := glProcedure('glCopyTexImage1D');
    glCopyTexImage2D := glProcedure('glCopyTexImage2D');
    glCopyTexSubImage1D := glProcedure('glCopyTexSubImage1D');
    glCopyTexSubImage2D := glProcedure('glCopyTexSubImage2D');
    glCullFace := glProcedure('glCullFace');
    glDeleteLists := glProcedure('glDeleteLists');
    glDeleteTextures := glProcedure('glDeleteTextures');
    glDepthFunc := glProcedure('glDepthFunc');
    glDepthMask := glProcedure('glDepthMask');
    glDepthRange := glProcedure('glDepthRange');
    glDisable := glProcedure('glDisable');
    glDisableClientState := glProcedure('glDisableClientState');
    glDrawArrays := glProcedure('glDrawArrays');
    glDrawBuffer := glProcedure('glDrawBuffer');
    glDrawElements := glProcedure('glDrawElements');
    glDrawPixels := glProcedure('glDrawPixels');
    glEdgeFlag := glProcedure('glEdgeFlag');
    glEdgeFlagPointer := glProcedure('glEdgeFlagPointer');
    glEdgeFlagv := glProcedure('glEdgeFlagv');
    glEnable := glProcedure('glEnable');
    glEnableClientState := glProcedure('glEnableClientState');
    glEnd := glProcedure('glEnd');
    glEndList := glProcedure('glEndList');
    glEvalCoord1d := glProcedure('glEvalCoord1d');
    glEvalCoord1dv := glProcedure('glEvalCoord1dv');
    glEvalCoord1f := glProcedure('glEvalCoord1f');
    glEvalCoord1fv := glProcedure('glEvalCoord1fv');
    glEvalCoord2d := glProcedure('glEvalCoord2d');
    glEvalCoord2dv := glProcedure('glEvalCoord2dv');
    glEvalCoord2f := glProcedure('glEvalCoord2f');
    glEvalCoord2fv := glProcedure('glEvalCoord2fv');
    glEvalMesh1 := glProcedure('glEvalMesh1');
    glEvalMesh2 := glProcedure('glEvalMesh2');
    glEvalPoint1 := glProcedure('glEvalPoint1');
    glEvalPoint2 := glProcedure('glEvalPoint2');
    glFeedbackBuffer := glProcedure('glFeedbackBuffer');
    glFinish := glProcedure('glFinish');
    glFlush := glProcedure('glFlush');
    glFogf := glProcedure('glFogf');
    glFogfv := glProcedure('glFogfv');
    glFogi := glProcedure('glFogi');
    glFogiv := glProcedure('glFogiv');
    glFrontFace := glProcedure('glFrontFace');
    glFrustum := glProcedure('glFrustum');
    glGenLists := glProcedure('glGenLists');
    glGenTextures := glProcedure('glGenTextures');
    glGetBooleanv := glProcedure('glGetBooleanv');
    glGetClipPlane := glProcedure('glGetClipPlane');
    glGetDoublev := glProcedure('glGetDoublev');
    glGetError := glProcedure('glGetError');
    glGetFloatv := glProcedure('glGetFloatv');
    glGetIntegerv := glProcedure('glGetIntegerv');
    glGetLightfv := glProcedure('glGetLightfv');
    glGetLightiv := glProcedure('glGetLightiv');
    glGetMapdv := glProcedure('glGetMapdv');
    glGetMapfv := glProcedure('glGetMapfv');
    glGetMapiv := glProcedure('glGetMapiv');
    glGetMaterialfv := glProcedure('glGetMaterialfv');
    glGetMaterialiv := glProcedure('glGetMaterialiv');
    glGetPixelMapfv := glProcedure('glGetPixelMapfv');
    glGetPixelMapuiv := glProcedure('glGetPixelMapuiv');
    glGetPixelMapusv := glProcedure('glGetPixelMapusv');
    glGetPointerv := glProcedure('glGetPointerv');
    glGetPolygonStipple := glProcedure('glGetPolygonStipple');
    glGetString := glProcedure('glGetString');
    glGetTexEnvfv := glProcedure('glGetTexEnvfv');
    glGetTexEnviv := glProcedure('glGetTexEnviv');
    glGetTexGendv := glProcedure('glGetTexGendv');
    glGetTexGenfv := glProcedure('glGetTexGenfv');
    glGetTexGeniv := glProcedure('glGetTexGeniv');
    glGetTexImage := glProcedure('glGetTexImage');
    glGetTexLevelParameterfv := glProcedure('glGetTexLevelParameterfv');
    glGetTexLevelParameteriv := glProcedure('glGetTexLevelParameteriv');
    glGetTexParameterfv := glProcedure('glGetTexParameterfv');
    glGetTexParameteriv := glProcedure('glGetTexParameteriv');
    glHint := glProcedure('glHint');
    glIndexMask := glProcedure('glIndexMask');
    glIndexPointer := glProcedure('glIndexPointer');
    glIndexd := glProcedure('glIndexd');
    glIndexdv := glProcedure('glIndexdv');
    glIndexf := glProcedure('glIndexf');
    glIndexfv := glProcedure('glIndexfv');
    glIndexi := glProcedure('glIndexi');
    glIndexiv := glProcedure('glIndexiv');
    glIndexs := glProcedure('glIndexs');
    glIndexsv := glProcedure('glIndexsv');
    glIndexub := glProcedure('glIndexub');
    glIndexubv := glProcedure('glIndexubv');
    glInitNames := glProcedure('glInitNames');
    glInterleavedArrays := glProcedure('glInterleavedArrays');
    glIsEnabled := glProcedure('glIsEnabled');
    glIsList := glProcedure('glIsList');
    glIsTexture := glProcedure('glIsTexture');
    glLightModelf := glProcedure('glLightModelf');
    glLightModelfv := glProcedure('glLightModelfv');
    glLightModeli := glProcedure('glLightModeli');
    glLightModeliv := glProcedure('glLightModeliv');
    glLightf := glProcedure('glLightf');
    glLightfv := glProcedure('glLightfv');
    glLighti := glProcedure('glLighti');
    glLightiv := glProcedure('glLightiv');
    glLineStipple := glProcedure('glLineStipple');
    glLineWidth := glProcedure('glLineWidth');
    glListBase := glProcedure('glListBase');
    glLoadIdentity := glProcedure('glLoadIdentity');
    glLoadMatrixd := glProcedure('glLoadMatrixd');
    glLoadMatrixf := glProcedure('glLoadMatrixf');
    glLoadName := glProcedure('glLoadName');
    glLogicOp := glProcedure('glLogicOp');
    glMap1d := glProcedure('glMap1d');
    glMap1f := glProcedure('glMap1f');
    glMap2d := glProcedure('glMap2d');
    glMap2f := glProcedure('glMap2f');
    glMapGrid1d := glProcedure('glMapGrid1d');
    glMapGrid1f := glProcedure('glMapGrid1f');
    glMapGrid2d := glProcedure('glMapGrid2d');
    glMapGrid2f := glProcedure('glMapGrid2f');
    glMaterialf := glProcedure('glMaterialf');
    glMaterialfv := glProcedure('glMaterialfv');
    glMateriali := glProcedure('glMateriali');
    glMaterialiv := glProcedure('glMaterialiv');
    glMatrixMode := glProcedure('glMatrixMode');
    glMultMatrixd := glProcedure('glMultMatrixd');
    glMultMatrixf := glProcedure('glMultMatrixf');
    glNewList := glProcedure('glNewList');
    glNormal3b := glProcedure('glNormal3b');
    glNormal3bv := glProcedure('glNormal3bv');
    glNormal3d := glProcedure('glNormal3d');
    glNormal3dv := glProcedure('glNormal3dv');
    glNormal3f := glProcedure('glNormal3f');
    glNormal3fv := glProcedure('glNormal3fv');
    glNormal3i := glProcedure('glNormal3i');
    glNormal3iv := glProcedure('glNormal3iv');
    glNormal3s := glProcedure('glNormal3s');
    glNormal3sv := glProcedure('glNormal3sv');
    glNormalPointer := glProcedure('glNormalPointer');
    glOrtho := glProcedure('glOrtho');
    glPassThrough := glProcedure('glPassThrough');
    glPixelMapfv := glProcedure('glPixelMapfv');
    glPixelMapuiv := glProcedure('glPixelMapuiv');
    glPixelMapusv := glProcedure('glPixelMapusv');
    glPixelStoref := glProcedure('glPixelStoref');
    glPixelStorei := glProcedure('glPixelStorei');
    glPixelTransferf := glProcedure('glPixelTransferf');
    glPixelTransferi := glProcedure('glPixelTransferi');
    glPixelZoom := glProcedure('glPixelZoom');
    glPointSize := glProcedure('glPointSize');
    glPolygonMode := glProcedure('glPolygonMode');
    glPolygonOffset := glProcedure('glPolygonOffset');
    glPolygonStipple := glProcedure('glPolygonStipple');
    glPopAttrib := glProcedure('glPopAttrib');
    glPopClientAttrib := glProcedure('glPopClientAttrib');
    glPopMatrix := glProcedure('glPopMatrix');
    glPopName := glProcedure('glPopName');
    glPrioritizeTextures := glProcedure('glPrioritizeTextures');
    glPushAttrib := glProcedure('glPushAttrib');
    glPushClientAttrib := glProcedure('glPushClientAttrib');
    glPushMatrix := glProcedure('glPushMatrix');
    glPushName := glProcedure('glPushName');
    glRasterPos2d := glProcedure('glRasterPos2d');
    glRasterPos2dv := glProcedure('glRasterPos2dv');
    glRasterPos2f := glProcedure('glRasterPos2f');
    glRasterPos2fv := glProcedure('glRasterPos2fv');
    glRasterPos2i := glProcedure('glRasterPos2i');
    glRasterPos2iv := glProcedure('glRasterPos2iv');
    glRasterPos2s := glProcedure('glRasterPos2s');
    glRasterPos2sv := glProcedure('glRasterPos2sv');
    glRasterPos3d := glProcedure('glRasterPos3d');
    glRasterPos3dv := glProcedure('glRasterPos3dv');
    glRasterPos3f := glProcedure('glRasterPos3f');
    glRasterPos3fv := glProcedure('glRasterPos3fv');
    glRasterPos3i := glProcedure('glRasterPos3i');
    glRasterPos3iv := glProcedure('glRasterPos3iv');
    glRasterPos3s := glProcedure('glRasterPos3s');
    glRasterPos3sv := glProcedure('glRasterPos3sv');
    glRasterPos4d := glProcedure('glRasterPos4d');
    glRasterPos4dv := glProcedure('glRasterPos4dv');
    glRasterPos4f := glProcedure('glRasterPos4f');
    glRasterPos4fv := glProcedure('glRasterPos4fv');
    glRasterPos4i := glProcedure('glRasterPos4i');
    glRasterPos4iv := glProcedure('glRasterPos4iv');
    glRasterPos4s := glProcedure('glRasterPos4s');
    glRasterPos4sv := glProcedure('glRasterPos4sv');
    glReadBuffer := glProcedure('glReadBuffer');
    glReadPixels := glProcedure('glReadPixels');
    glRectd := glProcedure('glRectd');
    glRectdv := glProcedure('glRectdv');
    glRectf := glProcedure('glRectf');
    glRectfv := glProcedure('glRectfv');
    glRecti := glProcedure('glRecti');
    glRectiv := glProcedure('glRectiv');
    glRects := glProcedure('glRects');
    glRectsv := glProcedure('glRectsv');
    glRenderMode := glProcedure('glRenderMode');
    glRotated := glProcedure('glRotated');
    glRotatef := glProcedure('glRotatef');
    glScaled := glProcedure('glScaled');
    glScalef := glProcedure('glScalef');
    glScissor := glProcedure('glScissor');
    glSelectBuffer := glProcedure('glSelectBuffer');
    glShadeModel := glProcedure('glShadeModel');
    glStencilFunc := glProcedure('glStencilFunc');
    glStencilMask := glProcedure('glStencilMask');
    glStencilOp := glProcedure('glStencilOp');
    glTexCoord1d := glProcedure('glTexCoord1d');
    glTexCoord1dv := glProcedure('glTexCoord1dv');
    glTexCoord1f := glProcedure('glTexCoord1f');
    glTexCoord1fv := glProcedure('glTexCoord1fv');
    glTexCoord1i := glProcedure('glTexCoord1i');
    glTexCoord1iv := glProcedure('glTexCoord1iv');
    glTexCoord1s := glProcedure('glTexCoord1s');
    glTexCoord1sv := glProcedure('glTexCoord1sv');
    glTexCoord2d := glProcedure('glTexCoord2d');
    glTexCoord2dv := glProcedure('glTexCoord2dv');
    glTexCoord2f := glProcedure('glTexCoord2f');
    glTexCoord2fv := glProcedure('glTexCoord2fv');
    glTexCoord2i := glProcedure('glTexCoord2i');
    glTexCoord2iv := glProcedure('glTexCoord2iv');
    glTexCoord2s := glProcedure('glTexCoord2s');
    glTexCoord2sv := glProcedure('glTexCoord2sv');
    glTexCoord3d := glProcedure('glTexCoord3d');
    glTexCoord3dv := glProcedure('glTexCoord3dv');
    glTexCoord3f := glProcedure('glTexCoord3f');
    glTexCoord3fv := glProcedure('glTexCoord3fv');
    glTexCoord3i := glProcedure('glTexCoord3i');
    glTexCoord3iv := glProcedure('glTexCoord3iv');
    glTexCoord3s := glProcedure('glTexCoord3s');
    glTexCoord3sv := glProcedure('glTexCoord3sv');
    glTexCoord4d := glProcedure('glTexCoord4d');
    glTexCoord4dv := glProcedure('glTexCoord4dv');
    glTexCoord4f := glProcedure('glTexCoord4f');
    glTexCoord4fv := glProcedure('glTexCoord4fv');
    glTexCoord4i := glProcedure('glTexCoord4i');
    glTexCoord4iv := glProcedure('glTexCoord4iv');
    glTexCoord4s := glProcedure('glTexCoord4s');
    glTexCoord4sv := glProcedure('glTexCoord4sv');
    glTexCoordPointer := glProcedure('glTexCoordPointer');
    glTexEnvf := glProcedure('glTexEnvf');
    glTexEnvfv := glProcedure('glTexEnvfv');
    glTexEnvi := glProcedure('glTexEnvi');
    glTexEnviv := glProcedure('glTexEnviv');
    glTexGend := glProcedure('glTexGend');
    glTexGendv := glProcedure('glTexGendv');
    glTexGenf := glProcedure('glTexGenf');
    glTexGenfv := glProcedure('glTexGenfv');
    glTexGeni := glProcedure('glTexGeni');
    glTexGeniv := glProcedure('glTexGeniv');
    glTexImage1D := glProcedure('glTexImage1D');
    glTexImage2D := glProcedure('glTexImage2D');
    glTexParameterf := glProcedure('glTexParameterf');
    glTexParameterfv := glProcedure('glTexParameterfv');
    glTexParameteri := glProcedure('glTexParameteri');
    glTexParameteriv := glProcedure('glTexParameteriv');
    glTexSubImage1D := glProcedure('glTexSubImage1D');
    glTexSubImage2D := glProcedure('glTexSubImage2D');
    glTranslated := glProcedure('glTranslated');
    glTranslatef := glProcedure('glTranslatef');
    glVertex2d := glProcedure('glVertex2d');
    glVertex2dv := glProcedure('glVertex2dv');
    glVertex2f := glProcedure('glVertex2f');
    glVertex2fv := glProcedure('glVertex2fv');
    glVertex2i := glProcedure('glVertex2i');
    glVertex2iv := glProcedure('glVertex2iv');
    glVertex2s := glProcedure('glVertex2s');
    glVertex2sv := glProcedure('glVertex2sv');
    glVertex3d := glProcedure('glVertex3d');
    glVertex3dv := glProcedure('glVertex3dv');
    glVertex3f := glProcedure('glVertex3f');
    glVertex3fv := glProcedure('glVertex3fv');
    glVertex3i := glProcedure('glVertex3i');
    glVertex3iv := glProcedure('glVertex3iv');
    glVertex3s := glProcedure('glVertex3s');
    glVertex3sv := glProcedure('glVertex3sv');
    glVertex4d := glProcedure('glVertex4d');
    glVertex4dv := glProcedure('glVertex4dv');
    glVertex4f := glProcedure('glVertex4f');
    glVertex4fv := glProcedure('glVertex4fv');
    glVertex4i := glProcedure('glVertex4i');
    glVertex4iv := glProcedure('glVertex4iv');
    glVertex4s := glProcedure('glVertex4s');
    glVertex4sv := glProcedure('glVertex4sv');
    glVertexPointer := glProcedure('glVertexPointer');
    glViewport := glProcedure('glViewport');

{$IFDEF GL_12}
    // GL 1.2
    glDrawRangeElements := glProcedure('glDrawRangeElementsEXT');
    glTexImage3D := glProcedure('glTexImage3DEXT');
    glTexSubImage3D := glProcedure('glTexSubImage3DEXT');
    glCopyTexSubImage3D := glProcedure('glCopyTexSubImage3DEXT');
{$ENDIF}

{$IFDEF GL_13}
    // GL 1.3
    glActiveTexture := glProcedure('glActiveTextureARB');
    glClientActiveTexture := glProcedure('glClientActiveTextureARB');
    glCompressedTexImage1D := glProcedure('glCompressedTexImage1DARB');
    glCompressedTexImage2D := glProcedure('glCompressedTexImage2DARB');
    glCompressedTexImage3D := glProcedure('glCompressedTexImage3DARB');
    glCompressedTexSubImage1D := glProcedure('glCompressedTexSubImage1DARB');
    glCompressedTexSubImage2D := glProcedure('glCompressedTexSubImage2DARB');
    glCompressedTexSubImage3D := glProcedure('glCompressedTexSubImage3DARB');
    glGetCompressedTexImage := glProcedure('glGetCompressedTexImageARB');
    glMultiTexCoord1d := glProcedure('glMultiTexCoord1dARB');
    glMultiTexCoord1dv := glProcedure('glMultiTexCoord1dvARB');
    glMultiTexCoord1f := glProcedure('glMultiTexCoord1fARB');
    glMultiTexCoord1fv := glProcedure('glMultiTexCoord1fvARB');
    glMultiTexCoord1i := glProcedure('glMultiTexCoord1iARB');
    glMultiTexCoord1iv := glProcedure('glMultiTexCoord1ivARB');
    glMultiTexCoord1s := glProcedure('glMultiTexCoord1sARB');
    glMultiTexCoord1sv := glProcedure('glMultiTexCoord1svARB');
    glMultiTexCoord2d := glProcedure('glMultiTexCoord2dARB');
    glMultiTexCoord2dv := glProcedure('glMultiTexCoord2dvARB');
    glMultiTexCoord2f := glProcedure('glMultiTexCoord2fARB');
    glMultiTexCoord2fv := glProcedure('glMultiTexCoord2fvARB');
    glMultiTexCoord2i := glProcedure('glMultiTexCoord2iARB');
    glMultiTexCoord2iv := glProcedure('glMultiTexCoord2ivARB');
    glMultiTexCoord2s := glProcedure('glMultiTexCoord2sARB');
    glMultiTexCoord2sv := glProcedure('glMultiTexCoord2svARB');
    glMultiTexCoord3d := glProcedure('glMultiTexCoord3dARB');
    glMultiTexCoord3dv := glProcedure('glMultiTexCoord3dvARB');
    glMultiTexCoord3f := glProcedure('glMultiTexCoord3fARB');
    glMultiTexCoord3fv := glProcedure('glMultiTexCoord3fvARB');
    glMultiTexCoord3i := glProcedure('glMultiTexCoord3iARB');
    glMultiTexCoord3iv := glProcedure('glMultiTexCoord3ivARB');
    glMultiTexCoord3s := glProcedure('glMultiTexCoord3sARB');
    glMultiTexCoord3sv := glProcedure('glMultiTexCoord3svARB');
    glMultiTexCoord4d := glProcedure('glMultiTexCoord4dARB');
    glMultiTexCoord4dv := glProcedure('glMultiTexCoord4dvARB');
    glMultiTexCoord4f := glProcedure('glMultiTexCoord4fARB');
    glMultiTexCoord4fv := glProcedure('glMultiTexCoord4fvARB');
    glMultiTexCoord4i := glProcedure('glMultiTexCoord4iARB');
    glMultiTexCoord4iv := glProcedure('glMultiTexCoord4ivARB');
    glMultiTexCoord4s := glProcedure('glMultiTexCoord4sARB');
    glMultiTexCoord4sv := glProcedure('glMultiTexCoord4svARB');
    glLoadTransposeMatrixd := glProcedure('glLoadTransposeMatrixdARB');
    glLoadTransposeMatrixf := glProcedure('glLoadTransposeMatrixfARB');
    glMultTransposeMatrixd := glProcedure('glMultTransposeMatrixdARB');
    glMultTransposeMatrixf := glProcedure('glMultTransposeMatrixfARB');
    glSampleCoverage := glProcedure('glSampleCoverageARB');
{$ENDIF}

{$IFDEF GL_14}
    // GL 1.4
    glFogCoordf := glProcedure('glFogCoordf');
    glFogCoordfv := glProcedure('glFogCoordfv');
    glFogCoordd := glProcedure('glFogCoordd');
    glFogCoorddv := glProcedure('glFogCoorddv');
    glFogCoordPointer := glProcedure('glFogCoordPointer');
    glMultiDrawArrays := glProcedure('glMultiDrawArrays');
    glMultiDrawElements := glProcedure('glMultiDrawElements');
    glPointParameterf := glProcedure('glPointParameterfARB');
    glPointParameterfv := glProcedure('glPointParameterfvARB');
    glSecondaryColor3b := glProcedure('glSecondaryColor3b');
    glSecondaryColor3bv := glProcedure('glSecondaryColor3bv');
    glSecondaryColor3d := glProcedure('glSecondaryColor3d');
    glSecondaryColor3dv := glProcedure('glSecondaryColor3dv');
    glSecondaryColor3f := glProcedure('glSecondaryColor3f');
    glSecondaryColor3fv := glProcedure('glSecondaryColor3fv');
    glSecondaryColor3i := glProcedure('glSecondaryColor3i');
    glSecondaryColor3iv := glProcedure('glSecondaryColor3iv');
    glSecondaryColor3s := glProcedure('glSecondaryColor3s');
    glSecondaryColor3sv := glProcedure('glSecondaryColor3sv');
    glSecondaryColor3ub := glProcedure('glSecondaryColor3ub');
    glSecondaryColor3ubv := glProcedure('glSecondaryColor3ubv');
    glSecondaryColor3ui := glProcedure('glSecondaryColor3ui');
    glSecondaryColor3uiv := glProcedure('glSecondaryColor3uiv');
    glSecondaryColor3us := glProcedure('glSecondaryColor3us');
    glSecondaryColor3usv := glProcedure('glSecondaryColor3usv');
    glSecondaryColorPointer := glProcedure('glSecondaryColorPointer');
    glBlendFuncSeparate := glProcedure('glBlendFuncSeparate');
    glWindowPos2d := glProcedure('glWindowPos2dARB');
    glWindowPos2f := glProcedure('glWindowPos2fARB');
    glWindowPos2i := glProcedure('glWindowPos2iARB');
    glWindowPos2s := glProcedure('glWindowPos2sARB');
    glWindowPos2dv := glProcedure('glWindowPos2dvARB');
    glWindowPos2fv := glProcedure('glWindowPos2fvARB');
    glWindowPos2iv := glProcedure('glWindowPos2ivARB');
    glWindowPos2sv := glProcedure('glWindowPos2svARB');
    glWindowPos3d := glProcedure('glWindowPos3dARB');
    glWindowPos3f := glProcedure('glWindowPos3fARB');
    glWindowPos3i := glProcedure('glWindowPos3iARB');
    glWindowPos3s := glProcedure('glWindowPos3sARB');
    glWindowPos3dv := glProcedure('glWindowPos3dvARB');
    glWindowPos3fv := glProcedure('glWindowPos3fvARB');
    glWindowPos3iv := glProcedure('glWindowPos3ivARB');
    glWindowPos3sv := glProcedure('glWindowPos3svARB');
{$ENDIF}

{$IFDEF GL_15}
    // GL 1.5
    glBindBuffer := glProcedure('glBindBufferARB');
    glDeleteBuffers := glProcedure('glDeleteBuffersARB');
    glGenBuffers := glProcedure('glGenBuffersARB');
    glIsBuffer := glProcedure('glIsBufferARB');
    glBufferData := glProcedure('glBufferDataARB');
    glBufferSubData := glProcedure('glBufferSubDataARB');
    glGetBufferSubData := glProcedure('glGetBufferSubDataARB');
    glMapBuffer := glProcedure('glMapBufferARB');
    glUnmapBuffer := glProcedure('glUnmapBufferARB');
    glGetBufferParameteriv := glProcedure('glGetBufferParameterivARB');
    glGetBufferPointerv := glProcedure('glGetBufferPointervARB');

    glGenQueries := glProcedure('glGenQueriesARB');
    glDeleteQueries := glProcedure('glDeleteQueriesARB');
    glIsQuery := glProcedure('glIsQueryARB');
    glBeginQuery := glProcedure('glBeginQueryARB');
    glEndQuery := glProcedure('glEndQueryARB');
    glGetQueryiv := glProcedure('glGetQueryivARB');
    glGetQueryObjectiv := glProcedure('glGetQueryObjectivARB');
    glGetQueryObjectuiv := glProcedure('glGetQueryObjectuivARB');
{$ENDIF}

{$IFDEF GL_PROGRAM}
    // GL_ARB_vertex_program, GL_ARB_fragment_program
    glVertexAttrib1sARB := glProcedure('glVertexAttrib1sARB');
    glVertexAttrib1fARB := glProcedure('glVertexAttrib1fARB');
    glVertexAttrib1dARB := glProcedure('glVertexAttrib1dARB');
    glVertexAttrib2sARB := glProcedure('glVertexAttrib2sARB');
    glVertexAttrib2fARB := glProcedure('glVertexAttrib2fARB');
    glVertexAttrib2dARB := glProcedure('glVertexAttrib2dARB');
    glVertexAttrib3sARB := glProcedure('glVertexAttrib3sARB');
    glVertexAttrib3fARB := glProcedure('glVertexAttrib3fARB');
    glVertexAttrib3dARB := glProcedure('glVertexAttrib3dARB');
    glVertexAttrib4sARB := glProcedure('glVertexAttrib4sARB');
    glVertexAttrib4fARB := glProcedure('glVertexAttrib4fARB');
    glVertexAttrib4dARB := glProcedure('glVertexAttrib4dARB');
    glVertexAttrib4NubARB := glProcedure('glVertexAttrib4NubARB');
    glVertexAttrib1svARB := glProcedure('glVertexAttrib1svARB');
    glVertexAttrib1fvARB := glProcedure('glVertexAttrib1fvARB');
    glVertexAttrib1dvARB := glProcedure('glVertexAttrib1dvARB');
    glVertexAttrib2svARB := glProcedure('glVertexAttrib2svARB');
    glVertexAttrib2fvARB := glProcedure('glVertexAttrib2fvARB');
    glVertexAttrib2dvARB := glProcedure('glVertexAttrib2dvARB');
    glVertexAttrib3svARB := glProcedure('glVertexAttrib3svARB');
    glVertexAttrib3fvARB := glProcedure('glVertexAttrib3fvARB');
    glVertexAttrib3dvARB := glProcedure('glVertexAttrib3dvARB');
    glVertexAttrib4bvARB := glProcedure('glVertexAttrib4bvARB');
    glVertexAttrib4svARB := glProcedure('glVertexAttrib4svARB');
    glVertexAttrib4ivARB := glProcedure('glVertexAttrib4ivARB');
    glVertexAttrib4ubvARB := glProcedure('glVertexAttrib4ubvARB');
    glVertexAttrib4usvARB := glProcedure('glVertexAttrib4usvARB');
    glVertexAttrib4uivARB := glProcedure('glVertexAttrib4uivARB');
    glVertexAttrib4fvARB := glProcedure('glVertexAttrib4fvARB');
    glVertexAttrib4dvARB := glProcedure('glVertexAttrib4dvARB');
    glVertexAttrib4NbvARB := glProcedure('glVertexAttrib4NbvARB');
    glVertexAttrib4NsvARB := glProcedure('glVertexAttrib4NsvARB');
    glVertexAttrib4NivARB := glProcedure('glVertexAttrib4NivARB');
    glVertexAttrib4NubvARB := glProcedure('glVertexAttrib4NubvARB');
    glVertexAttrib4NusvARB := glProcedure('glVertexAttrib4NusvARB');
    glVertexAttrib4NuivARB := glProcedure('glVertexAttrib4NuivARB');
    glVertexAttribPointerARB := glProcedure('glVertexAttribPointerARB');
    glEnableVertexAttribArrayARB := glProcedure('glEnableVertexAttribArrayARB');
    glDisableVertexAttribArrayARB := glProcedure('glDisableVertexAttribArrayARB');
    glProgramStringARB := glProcedure('glProgramStringARB');
    glBindProgramARB := glProcedure('glBindProgramARB');
    glDeleteProgramsARB := glProcedure('glDeleteProgramsARB');
    glGenProgramsARB := glProcedure('glGenProgramsARB');
    glProgramEnvParameter4fARB := glProcedure('glProgramEnvParameter4fARB');
    glProgramEnvParameter4dARB := glProcedure('glProgramEnvParameter4dARB');
    glProgramEnvParameter4fvARB := glProcedure('glProgramEnvParameter4fvARB');
    glProgramEnvParameter4dvARB := glProcedure('glProgramEnvParameter4dvARB');
    glProgramLocalParameter4fARB := glProcedure('glProgramLocalParameter4fARB');
    glProgramLocalParameter4dARB := glProcedure('glProgramLocalParameter4dARB');
    glProgramLocalParameter4fvARB := glProcedure('glProgramLocalParameter4fvARB');
    glProgramLocalParameter4dvARB := glProcedure('glProgramLocalParameter4dvARB');
    glGetProgramEnvParameterfvARB := glProcedure('glGetProgramEnvParameterfvARB');
    glGetProgramEnvParameterdvARB := glProcedure('glGetProgramEnvParameterdvARB');
    glGetProgramLocalParameterfvARB := glProcedure('glGetProgramLocalParameterfvARB');
    glGetProgramLocalParameterdvARB := glProcedure('glGetProgramLocalParameterdvARB');
    glGetProgramivARB := glProcedure('glGetProgramivARB');
    glGetProgramStringARB := glProcedure('glGetProgramStringARB');
    glGetVertexAttribdvARB := glProcedure('glGetVertexAttribdvARB');
    glGetVertexAttribfvARB := glProcedure('glGetVertexAttribfvARB');
    glGetVertexAttribivARB := glProcedure('glGetVertexAttribivARB');
    glIsProgramARB := glProcedure('glIsProgramARB');
{$ENDIF}

{$IFDEF GL_SHADER}
    // GL_ARB_shader_objects
    glDeleteObjectARB := glProcedure('glDeleteObjectARB');
    glGetHandleARB := glProcedure('glGetHandleARB');
    glDetachObjectARB := glProcedure('glDetachObjectARB');
    glCreateShaderObjectARB := glProcedure('glCreateShaderObjectARB');
    glShaderSourceARB := glProcedure('glShaderSourceARB');
    glCompileShaderARB := glProcedure('glCompileShaderARB');
    glCreateProgramObjectARB := glProcedure('glCreateProgramObjectARB');
    glAttachObjectARB := glProcedure('glAttachObjectARB');
    glLinkProgramARB := glProcedure('glLinkProgramARB');
    glUseProgramObjectARB := glProcedure('glUseProgramObjectARB');
    glValidateProgramARB := glProcedure('glValidateProgramARB');
    glUniform1fARB := glProcedure('glUniform1fARB');
    glUniform2fARB := glProcedure('glUniform2fARB');
    glUniform3fARB := glProcedure('glUniform3fARB');
    glUniform4fARB := glProcedure('glUniform4fARB');
    glUniform1iARB := glProcedure('glUniform1iARB');
    glUniform2iARB := glProcedure('glUniform2iARB');
    glUniform3iARB := glProcedure('glUniform3iARB');
    glUniform4iARB := glProcedure('glUniform4iARB');
    glUniform1fvARB := glProcedure('glUniform1fvARB');
    glUniform2fvARB := glProcedure('glUniform2fvARB');
    glUniform3fvARB := glProcedure('glUniform3fvARB');
    glUniform4fvARB := glProcedure('glUniform4fvARB');
    glUniform1ivARB := glProcedure('glUniform1ivARB');
    glUniform2ivARB := glProcedure('glUniform2ivARB');
    glUniform3ivARB := glProcedure('glUniform3ivARB');
    glUniform4ivARB := glProcedure('glUniform4ivARB');
    glUniformMatrix2fvARB := glProcedure('glUniformMatrix2fvARB');
    glUniformMatrix3fvARB := glProcedure('glUniformMatrix3fvARB');
    glUniformMatrix4fvARB := glProcedure('glUniformMatrix4fvARB');
    glGetObjectParameterfvARB := glProcedure('glGetObjectParameterfvARB');
    glGetObjectParameterivARB := glProcedure('glGetObjectParameterivARB');
    glGetInfoLogARB := glProcedure('glGetInfoLogARB');
    glGetAttachedObjectsARB := glProcedure('glGetAttachedObjectsARB');
    glGetUniformLocationARB := glProcedure('glGetUniformLocationARB');
    glGetActiveUniformARB := glProcedure('glGetActiveUniformARB');
    glGetUniformfvARB := glProcedure('glGetUniformfvARB');
    glGetUniformivARB := glProcedure('glGetUniformivARB');
    glGetShaderSourceARB := glProcedure('glGetShaderSourceARB');

    // GL_ARB_vertex_shader
    glGetActiveAttribARB := glProcedure('glGetActiveAttribARB');
    glGetAttribLocationARB := glProcedure('glGetAttribLocationARB');
    glBindAttribLocationARB := glProcedure('glBindAttribLocationARB');
{$ENDIF}
  end;

{$IFNDEF NO_GLU}
  if GLUHandle <> INVALID_MODULEHANDLE then
  begin
    gluBeginCurve := GetProcAddress(GLUHandle, 'gluBeginCurve');
    gluBeginPolygon := GetProcAddress(GLUHandle, 'gluBeginPolygon');
    gluBeginSurface := GetProcAddress(GLUHandle, 'gluBeginSurface');
    gluBeginTrim := GetProcAddress(GLUHandle, 'gluBeginTrim');
    gluBuild1DMipmaps := GetProcAddress(GLUHandle, 'gluBuild1DMipmaps');
    gluBuild2DMipmaps := GetProcAddress(GLUHandle, 'gluBuild2DMipmaps');
    gluBuild3DMipmaps := GetProcAddress(GLUHandle, 'gluBuild3DMipmaps');
    gluCylinder := GetProcAddress(GLUHandle, 'gluCylinder');
    gluDeleteNurbsRenderer := GetProcAddress(GLUHandle, 'gluDeleteNurbsRenderer');
    gluDeleteQuadric := GetProcAddress(GLUHandle, 'gluDeleteQuadric');
    gluDeleteTess := GetProcAddress(GLUHandle, 'gluDeleteTess');
    gluDisk := GetProcAddress(GLUHandle, 'gluDisk');
    gluEndCurve := GetProcAddress(GLUHandle, 'gluEndCurve');
    gluEndPolygon := GetProcAddress(GLUHandle, 'gluEndPolygon');
    gluEndSurface := GetProcAddress(GLUHandle, 'gluEndSurface');
    gluEndTrim := GetProcAddress(GLUHandle, 'gluEndTrim');
    gluErrorString := GetProcAddress(GLUHandle, 'gluErrorString');
    gluGetNurbsProperty := GetProcAddress(GLUHandle, 'gluGetNurbsProperty');
    gluGetString := GetProcAddress(GLUHandle, 'gluGetString');
    gluGetTessProperty := GetProcAddress(GLUHandle, 'gluGetTessProperty');
    gluLoadSamplingMatrices := GetProcAddress(GLUHandle, 'gluLoadSamplingMatrices');
    gluLookAt := GetProcAddress(GLUHandle, 'gluLookAt');
    gluNewNurbsRenderer := GetProcAddress(GLUHandle, 'gluNewNurbsRenderer');
    gluNewQuadric := GetProcAddress(GLUHandle, 'gluNewQuadric');
    gluNewTess := GetProcAddress(GLUHandle, 'gluNewTess');
    gluNextContour := GetProcAddress(GLUHandle, 'gluNextContour');
    gluNurbsCallback := GetProcAddress(GLUHandle, 'gluNurbsCallback');
    gluNurbsCurve := GetProcAddress(GLUHandle, 'gluNurbsCurve');
    gluNurbsProperty := GetProcAddress(GLUHandle, 'gluNurbsProperty');
    gluNurbsSurface := GetProcAddress(GLUHandle, 'gluNurbsSurface');
    gluOrtho2D := GetProcAddress(GLUHandle, 'gluOrtho2D');
    gluPartialDisk := GetProcAddress(GLUHandle, 'gluPartialDisk');
    gluPerspective := GetProcAddress(GLUHandle, 'gluPerspective');
    gluPickMatrix := GetProcAddress(GLUHandle, 'gluPickMatrix');
    gluProject := GetProcAddress(GLUHandle, 'gluProject');
    gluPwlCurve := GetProcAddress(GLUHandle, 'gluPwlCurve');
    gluQuadricCallback := GetProcAddress(GLUHandle, 'gluQuadricCallback');
    gluQuadricDrawStyle := GetProcAddress(GLUHandle, 'gluQuadricDrawStyle');
    gluQuadricNormals := GetProcAddress(GLUHandle, 'gluQuadricNormals');
    gluQuadricOrientation := GetProcAddress(GLUHandle, 'gluQuadricOrientation');
    gluQuadricTexture := GetProcAddress(GLUHandle, 'gluQuadricTexture');
    gluScaleImage := GetProcAddress(GLUHandle, 'gluScaleImage');
    gluSphere := GetProcAddress(GLUHandle, 'gluSphere');
    gluTessBeginContour := GetProcAddress(GLUHandle, 'gluTessBeginContour');
    gluTessBeginPolygon := GetProcAddress(GLUHandle, 'gluTessBeginPolygon');
    gluTessCallback := GetProcAddress(GLUHandle, 'gluTessCallback');
    gluTessEndContour := GetProcAddress(GLUHandle, 'gluTessEndContour');
    gluTessEndPolygon := GetProcAddress(GLUHandle, 'gluTessEndPolygon');
    gluTessNormal := GetProcAddress(GLUHandle, 'gluTessNormal');
    gluTessProperty := GetProcAddress(GLUHandle, 'gluTessProperty');
    gluTessVertex := GetProcAddress(GLUHandle, 'gluTessVertex');
    gluUnProject := GetProcAddress(GLUHandle, 'gluUnProject');
  end;
{$ENDIF}

  if SWGL then begin
    if not assigned(wglChoosePixelFormat) then wglChoosePixelFormat := ChoosePixelFormat else ChoosePixelFormat := wglChoosePixelFormat;
    if not assigned(wglGetPixelFormat) then wglGetPixelFormat := GetPixelFormat else GetPixelFormat := wglGetPixelFormat;
    if not assigned(wglSetPixelFormat) then wglSetPixelFormat := SetPixelFormat else SetPixelFormat := wglSetPixelFormat;
    if not assigned(wglDescribePixelFormat) then wglDescribePixelFormat := DescribePixelFormat else DescribePixelFormat := wglDescribePixelFormat;
    if not assigned(wglSwapBuffers) then wglSwapBuffers := SwapBuffers else SwapBuffers := wglSwapBuffers;
  end;

end;

//----------------------------------------------------------------------------------------------------------------------

procedure TrimAndSplitVersionString(Buffer: AnsiString; var Max, Min: Integer);

// Peels out the X.Y form from the given Buffer which must contain a version string like "text Minor.Major.Build text"
// at least however "Major.Minor".

var
  Separator: Integer;

begin
  try
    // There must be at least one dot to separate major and minor version number.
    Separator := Pos('.', Buffer);
    // At least one number must be before and one after the dot.
    if (Separator > 1) and (Separator < Length(Buffer)) and (Buffer[Separator - 1] in ['0'..'9']) and
      (Buffer[Separator + 1] in ['0'..'9']) then
    begin
      // OK, it's a valid version string. Now remove unnecessary parts.
      Dec(Separator);
      // Find last non-numeric character before version number.
      while (Separator > 0) and (Buffer[Separator] in ['0'..'9']) do
        Dec(Separator);
      // Delete leading characters which do not belong to the version string.
      Delete(Buffer, 1, Separator);
      Separator := Pos('.', Buffer) + 1;
      // Find first non-numeric character after version number
      while (Separator <= Length(Buffer)) and (Buffer[Separator] in ['0'..'9']) do
        Inc(Separator);
      // delete trailing characters not belonging to the version string
      Delete(Buffer, Separator, 255);
      // Now translate the numbers.
      Separator := Pos('.', Buffer); // This is necessary because the buffer length might have changed.
      Max := StrToInt(Copy(Buffer, 1, Separator - 1));
      Min := StrToInt(Copy(Buffer, Separator + 1, 255));
    end
    else
      Abort;
  except
    Min := 0;
    Max := 0;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure ReadImplementationProperties;

var
  Buffer: AnsiString;
  MajorVersion,
  MinorVersion: Integer;

  // Checks if the given Extension string is in Buffer.
  function CheckExtension(Extension: string): Boolean;
  var ExtPos: Integer;
  begin
    // First find the position of the extension string as substring in Buffer.
    ExtPos := Pos(Extension, Buffer);
    Result := ExtPos > 0;
    // Now check that it isn't only a substring of another extension.
    if Result then
      Result := ((ExtPos + Length(Extension) - 1) = Length(Buffer)) or
        not (Buffer[ExtPos + Length(Extension)] in ['_', 'A'..'Z', 'a'..'z']);
  end;

begin
  // determine version of implementation
  // GL
  Buffer := glGetString(GL_VERSION);
  TrimAndSplitVersionString(Buffer, Majorversion, MinorVersion);
  GL_10 := True;
  GL_11 := False;
  GL_12 := False;
  GL_13 := False;
  GL_14 := False;
  GL_15 := False;
  if MajorVersion > 0 then
  begin
    if MinorVersion > 0 then
    begin
      GL_11 := True;
      if MinorVersion > 1 then begin
        GL_12 := True;
        if MinorVersion > 2 then begin
          GL_13 := True;
          if MinorVersion > 3 then begin
            GL_14 := True;
            if MinorVersion > 4 then
              GL_15 := True;
          end;
        end;
      end;
    end;
  end;

  // Extensions
  Buffer := glGetString(GL_EXTENSIONS);
  GL_PROGRAM := CheckExtension('GL_ARB_vertex_program');
  GL_SHADER := CheckExtension('GL_ARB_shader_objects') and CheckExtension('GL_ARB_vertex_shader');

  // GLU
  GLU_11 := False;
  GLU_12 := False;
  GLU_13 := False;
  GLU_14 := False;
{$IFNDEF NO_GLU}
  // gluGetString is valid for version 1.1 or later
  if Assigned(gluGetString) then
  begin
    Buffer := gluGetString(GLU_VERSION);
    TrimAndSplitVersionString(Buffer, Majorversion, MinorVersion);
    GLU_11 := True;
    if MinorVersion > 1 then
    begin
      GLU_12 := True;
      if MinorVersion > 2 then begin
        GLU_13 := True;
        if MinorVersion > 3 then
          GLU_14 := True;
      end;
    end;
  end;
{$ENDIF}
end;

//----------------------------------------------------------------------------------------------------------------------

function CreateRenderingContext(DC: HDC; PixelFormat: Integer; var PFDescriptor: TPixelFormatDescriptor): HGLRC;

// Set the OpenGL properties required to draw to the given canvas and create a rendering context for it.

begin
  // Just in case it didn't happen already.
  if not InitOpenGL then
    ShowError('OpenGL:CreateRenderingContext:InitOpenGL failed');

  // choose pixel format if none is given
  if PixelFormat = 0 then
    PixelFormat := ChoosePixelFormat(DC, @PFDescriptor);

  // throw error if no pixel format could be obtained
  if PixelFormat = 0 then
    ShowError('OpenGL:CreateRenderingContext:ChoosePixelFormat failed');

  // NOTE: It is not allowed to change a pixel format of a device context once it has been set.
  //       Hence you may create more than one rendering context for one single device only if it
  //       uses the same pixel format as the first created RC.
  if GetPixelFormat(DC) <> PixelFormat then
  begin
    if not SetPixelFormat(DC, PixelFormat, @PFDescriptor) then
      ShowError('OpenGL:CreateRenderingContext:SetPixelFormat failed');
  end;

  // Check the properties we just set.
  DescribePixelFormat(DC, PixelFormat, SizeOf(PFDescriptor), PFDescriptor);
  Result := wglCreateContext(DC);
  if Result = 0 then
    ShowError('OpenGL:CreateRenderingContext:wglCreateContext failed');
end;

//----------------------------------------------------------------------------------------------------------------------

procedure ActivateRenderingContext(DC: HDC; RC: HGLRC);

begin
  if not wglMakeCurrent(DC, RC) then ShowError('wglMakeCurrent failed');
  LoadProcAddresses;
  ReadImplementationProperties;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure CloseOpenGL;

begin
  if GLHandle <> INVALID_MODULEHANDLE then
  begin
    FreeLibrary(Cardinal(GLHandle));
    GLHandle := INVALID_MODULEHANDLE;
  end;

{$IFNDEF NO_GLU}
  if GLUHandle <> INVALID_MODULEHANDLE then
  begin
    FreeLibrary(Cardinal(GLUHandle));
    GLUHandle := INVALID_MODULEHANDLE;
  end;
{$ENDIF}

  ClearProcAddresses;
end;

//----------------------------------------------------------------------------------------------------------------------

function InitOpenGL: Boolean;

begin
  if SGL[1]='"' then delete(SGL, 1, 1);
  if SGL[Length(SGL)] = '"' then SetLength(SGL, Length(SGL)-1);
{$IFNDEF NO_GLU}
  if SGLU[1]='"' then delete(SGLU, 1, 1);
  if SGLU[Length(SGLU)] = '"' then SetLength(SGLU, Length(SGLU)-1);
{$ENDIF}
  if (GLHandle = INVALID_MODULEHANDLE) {$IFNDEF NO_GLU} or (GLUHandle = INVALID_MODULEHANDLE) {$ENDIF} then
    Result := InitOpenGLFromLibrary(SGL {$IFNDEF NO_GLU}, SGLU {$ENDIF})
  else
    Result := True;
end;

//----------------------------------------------------------------------------------------------------------------------

function InitOpenGLFromLibrary(GLName {$IFNDEF NO_GLU}, GLUName{$ENDIF}: AnsiString): Boolean;

begin
  Result := False;
  CloseOpenGL;

  GLHandle := LoadLibrary(PChar(GLName));
{$IFNDEF NO_GLU}
  GLUHandle := LoadLibrary(PChar(GLUName));
{$ENDIF}
  if (GLHandle <> INVALID_MODULEHANDLE) {$IFNDEF NO_GLU} and (GLUHandle <> INVALID_MODULEHANDLE) {$ENDIF} then
  begin
    LoadProcAddresses;
    Result := True;
  end
  else
  begin
    if GLHandle <>  INVALID_MODULEHANDLE then
      FreeLibrary(Cardinal(GLHandle));

{$IFNDEF NO_GLU}
      if GLUHandle <>  INVALID_MODULEHANDLE then
      FreeLibrary(Cardinal(GLUHandle));
{$ENDIF}
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function IsOpenGLInitialized: Boolean;

begin
  Result := GLHandle <> INVALID_MODULEHANDLE;
end;

{$IFDEF NO_GLU}
// _no_glu code start --------------------------------------------------------------------------------------------------
// GLU matrix routines -------------------------------------------------------------------------------------------------


// compute tangens
function Tan(const X: Extended): Extended;
{  Tan := Sin(X) / Cos(X) }
asm
        FLD    X
        FPTAN
        FSTP   ST(0)      { FPTAN pushes 1.0 after result }
        FWAIT
end;


// multiplies two 4x4 matrices
function MatrixMultiply(M1, M2: TMatrix4d): TMatrix4d;
var
  I, J: Integer;
  TM: TMatrix4d;
begin
  for I := 0 to 3 do
    for J := 0 to 3 do
      TM[I, J] := M1[I, 0] * M2[0, J] +
                  M1[I, 1] * M2[1, J] +
                  M1[I, 2] * M2[2, J] +
                  M1[I, 3] * M2[3, J];
  Result := TM;
end;


// Determinant of a 3x3 matrix
function MatrixDet3x3(a1, a2, a3, b1, b2, b3, c1, c2, c3: GLDouble): GLDouble;
begin
  Result := a1 * (b2 * c3 - b3 * c2) -
            b1 * (a2 * c3 - a3 * c2) +
            c1 * (a2 * b3 - a3 * b2);
end;



// Determinant of a 4x4 matrix
function MatrixDeterminant(M: TMatrix4d): GLDouble;
var
  a1, a2, a3, a4,
  b1, b2, b3, b4,
  c1, c2, c3, c4,
  d1, d2, d3, d4  : GLDouble;
begin
  a1 := M[0,0];  b1 := M[0,1];  c1 := M[0,2];  d1 := M[0,3];
  a2 := M[1,0];  b2 := M[1,1];  c2 := M[1,2];  d2 := M[1,3];
  a3 := M[2,0];  b3 := M[2,1];  c3 := M[2,2];  d3 := M[2,3];
  a4 := M[3,0];  b4 := M[3,1];  c4 := M[3,2];  d4 := M[3,3];

  Result := a1 * MatrixDet3x3(b2, b3, b4, c2, c3, c4, d2, d3, d4) -
            b1 * MatrixDet3x3(a2, a3, a4, c2, c3, c4, d2, d3, d4) +
            c1 * MatrixDet3x3(a2, a3, a4, b2, b3, b4, d2, d3, d4) -
            d1 * MatrixDet3x3(a2, a3, a4, b2, b3, b4, c2, c3, c4);
end;


// Adjoint of a 4x4 matrix - used in the computation of the inverse of a 4x4 matrix
procedure MatrixAdjoint(var M: TMatrix4d);
var
  a1, a2, a3, a4,
  b1, b2, b3, b4,
  c1, c2, c3, c4,
  d1, d2, d3, d4: GLDouble;
begin
    a1 :=  M[0,0]; b1 :=  M[0,1];
    c1 :=  M[0,2]; d1 :=  M[0,3];
    a2 :=  M[1,0]; b2 :=  M[1,1];
    c2 :=  M[1,2]; d2 :=  M[1,3];
    a3 :=  M[2,0]; b3 :=  M[2,1];
    c3 :=  M[2,2]; d3 :=  M[2,3];
    a4 :=  M[3,0]; b4 :=  M[3,1];
    c4 :=  M[3,2]; d4 :=  M[3,3];

    // row column labeling reversed since we transpose rows & columns
    M[0,0] :=  MatrixDet3x3(b2, b3, b4, c2, c3, c4, d2, d3, d4);
    M[1,0] := -MatrixDet3x3(a2, a3, a4, c2, c3, c4, d2, d3, d4);
    M[2,0] :=  MatrixDet3x3(a2, a3, a4, b2, b3, b4, d2, d3, d4);
    M[3,0] := -MatrixDet3x3(a2, a3, a4, b2, b3, b4, c2, c3, c4);

    M[0,1] := -MatrixDet3x3(b1, b3, b4, c1, c3, c4, d1, d3, d4);
    M[1,1] :=  MatrixDet3x3(a1, a3, a4, c1, c3, c4, d1, d3, d4);
    M[2,1] := -MatrixDet3x3(a1, a3, a4, b1, b3, b4, d1, d3, d4);
    M[3,1] :=  MatrixDet3x3(a1, a3, a4, b1, b3, b4, c1, c3, c4);

    M[0,2] :=  MatrixDet3x3(b1, b2, b4, c1, c2, c4, d1, d2, d4);
    M[1,2] := -MatrixDet3x3(a1, a2, a4, c1, c2, c4, d1, d2, d4);
    M[2,2] :=  MatrixDet3x3(a1, a2, a4, b1, b2, b4, d1, d2, d4);
    M[3,2] := -MatrixDet3x3(a1, a2, a4, b1, b2, b4, c1, c2, c4);

    M[0,3] := -MatrixDet3x3(b1, b2, b3, c1, c2, c3, d1, d2, d3);
    M[1,3] :=  MatrixDet3x3(a1, a2, a3, c1, c2, c3, d1, d2, d3);
    M[2,3] := -MatrixDet3x3(a1, a2, a3, b1, b2, b3, d1, d2, d3);
    M[3,3] :=  MatrixDet3x3(a1, a2, a3, b1, b2, b3, c1, c2, c3);
end;


// inverts a 4x4 matrix
procedure MatrixInvert(var M: TMatrix4d);
var
  Det : GLdouble;
  i, j : Integer;
begin
  Det := MatrixDeterminant(M);
  if Abs(Det) < 0.000001 then begin
    M[0][0]:=1.0; M[1][0]:=0.0; M[2][0]:=0.0; M[3][0]:=0.0;
    M[0][1]:=0.0; M[1][1]:=1.0; M[2][1]:=0.0; M[3][1]:=0.0;
    M[0][2]:=0.0; M[1][2]:=0.0; M[2][2]:=1.0; M[3][2]:=0.0;
    M[0][3]:=0.0; M[1][3]:=0.0; M[2][3]:=0.0; M[3][3]:=1.0;
  end
  else
    begin
      MatrixAdjoint(M);
      for i := 0 to 3 do
        for j := 0 to 3 do
          M[i, j] := M[i, j] / Det;
    end;
end;


// transforms 4d vector by 4x4 matrix
procedure Vector4dTransform(var v : TVector4d; m : TMatrix4d);

  var
    x, y, z, w : GLdouble;

  begin
    x := v[0]*m[0,0] + v[1]*m[1,0] + v[2]*m[2,0] + v[3]*m[3,0];
    y := v[0]*m[0,1] + v[1]*m[1,1] + v[2]*m[2,1] + v[3]*m[3,1];
    z := v[0]*m[0,2] + v[1]*m[1,2] + v[2]*m[2,2] + v[3]*m[3,2];
    w := v[0]*m[0,3] + v[1]*m[1,3] + v[2]*m[2,3] + v[3]*m[3,3];
    v[0] := x;
    v[1] := y;
    v[2] := z;
    v[3] := w;
  end;



procedure gluLookAt(eyex, eyey, eyez,  centerx, centery, centerz,  upx, upy, upz: GLdouble);
var
  m: Array[0..3] of Array[0..3] of GLdouble;
  x,y,z: Array[0..2] of GLdouble;
  mag: GLdouble;
begin
  // Make rotation matrix

  // Z vector
  z[0]:=eyex-centerx; z[1]:=eyey-centery; z[2]:=eyez-centerz;
  mag := sqrt(z[0]*z[0] + z[1]*z[1] + z[2]*z[2]);
  if mag>0 then begin
    z[0] := z[0] / mag;
    z[1] := z[1] / mag;
    z[2] := z[2] / mag;
  end;

  // Y vector
  y[0]:=upx; y[1]:=upy; y[2]:=upz;

  // X vector = Y cross Z
  x[0] := y[1] * z[2] - y[2] * z[1];
  x[1] := -y[0] * z[2] + y[2] * z[0];
  x[2] := y[0] * z[1] - y[1] * z[0];

  // Recompute Y = Z cross X
  y[0] := z[1] * x[2] - z[2] * x[1];
  y[1] := -z[0] * x[2] + z[2] * x[0];
  y[2] := z[0] * x[1] - z[1] * x[0];

  // cross product gives area of parallelogram, which is < 1.0 for
  // non-perpendicular unit-length vectors; so normalize x, y here

  mag := sqrt(x[0]*x[0] + x[1]*x[1] + x[2]*x[2]);
  if mag>0 then begin
    x[0] := x[0] / mag;
    x[1] := x[1] / mag;
    x[2] := x[2] / mag;
  end;
  mag := sqrt(y[0]*y[0] + y[1]*y[1] + y[2]*y[2]);
  if mag>0 then begin
    y[0] := y[0] / mag;
    y[1] := y[1] / mag;
    y[2] := y[2] / mag;
  end;

  m[0][0]:=x[0]; m[1][0]:=x[1]; m[2][0]:=x[2]; m[3][0]:=0.0;
  m[0][1]:=y[0]; m[1][1]:=y[1]; m[2][1]:=y[2]; m[3][1]:=0.0;
  m[0][2]:=z[0]; m[1][2]:=z[1]; m[2][2]:=z[2]; m[3][2]:=0.0;
  m[0][3]:=0.0; m[1][3]:=0.0; m[2][3]:=0.0; m[3][3]:=1.0;

  glMultMatrixd(Addr(m));

  // Translate Eye to Origin
  glTranslated(-eyex, -eyey, -eyez);
end;


procedure gluPerspective(fovy, aspect, zNear, zFar: GLdouble);
var
  xmin, xmax, ymin, ymax : GLdouble;
begin
  ymax := zNear * tan(fovy * PI / 360.0);
  ymin := -ymax;
  xmin := ymin * aspect;
  xmax := ymax * aspect;
  glFrustum(xmin, xmax, ymin, ymax, zNear, zFar);
end;


procedure gluPickMatrix(x, y, width, height: GLdouble; viewport: TVector4i);
var
  m: Array[0..3] of Array[0..3] of GLdouble;
  sx, sy, tx, ty: GLdouble;
begin
  sx := viewport[2] / width;
  sy := viewport[3] / height;
  tx := (viewport[2] + 2.0 * (viewport[0] - x)) / width;
  ty := (viewport[3] + 2.0 * (viewport[1] - y)) / height;

  m[0][0]:=sx;   m[1][0]:=0.0; m[2][0]:=0.0; m[3][0]:=tx;
  m[0][1]:=0.0; m[1][1]:=sy;   m[2][1]:=0.0; m[3][1]:=ty;
  m[0][2]:=0.0; m[1][2]:=0.0; m[2][2]:=1.0; m[3][2]:=0.0;
  m[0][3]:=0.0; m[1][3]:=0.0; m[2][3]:=0.0; m[3][3]:=1.0;

  glMultMatrixd(Addr(m));
end;


function gluProject(objx, objy, objz: GLdouble; modelMatrix, projMatrix: TMatrix4d; viewport: TVector4i;
                    winx, winy, winz: PGLdouble): GLint;
var v : TVector4d;
begin
   // initialize and transform vector
   v[0] := objx;
   v[1] := objy;
   v[2] := objz;
   v[3] := 1.0;
   Vector4dTransform(v, modelMatrix);
   Vector4dTransform(v, projMatrix);

   // normalize vector
   if (v[3] = 0.0) then
     result := GL_FALSE
   else begin
     v[0] := v[0] / v[3];
     v[1] := v[1] / v[3];
     v[2] := v[2] / v[3];

     // return window coordinates
     winx^ := viewport[0] + (1 + v[0]) * viewport[2] / 2;
     winy^ := viewport[1] + (1 + v[1]) * viewport[3] / 2;
     winz^ := (1 + v[2]) / 2;
     result := GL_TRUE;
   end;
end;


function gluUnProject(winx, winy, winz: GLdouble; modelMatrix, projMatrix: TMatrix4d; viewport: TVector4i;
                      objx, objy, objz: PGLdouble): GLint;
var m : TMatrix4d;  // transformation matrices
    v : TVector4d;
begin
   // normalize transformation coordinates
   v[0] := (winx - viewport[0]) * 2 / viewport[2] - 1.0;
   v[1] := (winy - viewport[1]) * 2 / viewport[3] - 1.0;
   v[2] := 2 * winz - 1.0;
   v[3] := 1.0;

   // calculate inveraw transformation
   m := MatrixMultiply(modelMatrix, projMatrix);
   MatrixInvert(m);

   // compute object coordinates
   Vector4dTransform(v, m);
   if (v[3] = 0.0) then
      result := GL_FALSE
   else begin
     objx^ := v[0] / v[3];
     objy^ := v[1] / v[3];
     objz^ := v[2] / v[3];
     result := GL_TRUE;
   end;
end;

// _no_glu code end ----------------------------------------------------------------------------------------------------
{$ENDIF}

initialization
  Set8087CW($133F);
finalization
  CloseOpenGL;
  // We don't need to reset the FPU control word as the previous set call is process specific.
end.
